      module jumpCom;
      save;
!***    global variables used in jump detection

!***    modifications:
!***      12/9/05  B. Gibbs - changed sign definition of dxdp (now partial of true - estimated state)

!**************************************
! The author grants the user a non-exclusive, worldwide, royalty-free copyright license to

! 1. reproduce and modify the software for your own purposes, and
! 2. to distribute the software provided that you give credit to the author,
!    do not make claims against the author or Wiley-Interscience,
!    and mark any modifications as your own.  If the software is incorporated in
!    a commercial product, you agree to defend and indemnify this author and
!    Wiley-Interscience against any losses, damages and costs arising from claims,
!    lawsuits and other legal actions brought by a third party.

! The software is provided on an as is basis, without warranties or conditions
! of any kind, either express or implied (including any warranties or conditions
! of title, non-infringement, merchantability or fitness for a particular purpose).
! The user is solely responsible for determining the appropriateness of using and
! distributing the program and assumes all risks associated with its exercise of
! rights under this agreement.
!**************************************

      !** dimension variables
      integer(4) :: ndim;        !max # filter states (e.g., 27)
      integer(4) :: n2dim;       !size of filter covariance matrix =(ndim*(ndim+1))/2
      integer(4) :: mdim;        !max # measurements (e.g., 3)
      integer(4) :: maxjtim;     !max # times in moving window  (e.g., 25)
      integer(4) :: maxjmp;      !max number of jump states at each time point (e.g., 15)
      integer(4) :: nparm;       !=maxjtim*maxjmp
!      integer(4) :: npar2        !=nparm+2
      integer(4) :: n2par;       !max size of jump covariance matrix =(nparm*(nparm+1))/2
      integer(4) :: npmax;       !max allowable number of "accepted" jump parm (<= nparm)
      integer(4) :: np2max;      !=(npmax*(npmax+1))/2

      !** jump variables
      integer(4) nxjmp;          !actual number of jump states in filter
      integer(4) np;             !total number of candidate jump parameters (=maxjtim*nxjmp <=nparm)
      integer(4) ifile;          !input file unit # (50)

      real(8) f2rem;             !value of f-statistic used as a minimum
                                 ! threshold when removing a parameter from the
                                 ! stepwise regression. (suggest 0.5)
      real(8) f2add;             !value of f-statistic used as a maximum
                                 ! threshold when adding a parameter to the
                                 ! stepwise regression. (suggest 9.0)
      real(8) tolmin;            !minimum value of tolerance test (variance ratio) when deciding
                                 ! whether to add a parameter to the regression.
                                 ! the tolerance is a measure of the independence
                                 ! of the parameter from other parameters in the
                                 ! regression.  (suggest tolmin=0.01).

      integer(4),allocatable :: jmpstat(:);  !index of filter states allowed to jump
!      real(8),allocatable :: tjump(:)
      real(8),allocatable :: sigxj(:); !1-sigma prior uncertainty on initial value
                                       ! of states allowed to jump. a scaled amount
                                       ! of sigxj is used as the prior uncertainty
                                       ! on the jumps in routine estjmp.
      real(8),allocatable :: qmin(:);  !minimum (constant) value of process noise
                                       ! on states allowed to jump.  This is only
                                       ! used in jumpest1 for output purposes in routine
                                       ! 'jumpsum' when generating a 'noisin' file.
      !** buffer variables
      integer(4) ibuf2;                !current # time entries in jbuff (ibuf2 <= maxjtim)
      integer(4) itim;                 !index of current time sample (should be limited ??)
      integer(4) nlag;                 !number of time lags before adding parameter (e.g., 5)
      integer(4) :: nxp;               !number of accepted "accepted" jump parameters
      integer(4),allocatable :: ilab(:,:);  !jump parameter labels
!      logical(4),allocatable :: isel(:)   !jump parameter selection flag: 1=selected  not used
      integer(4),allocatable :: ilxp(:,:);!"accepted" jump parameter labels
      real(8) sigfac;                  !factor multiplying sigxj used as prior
                                       ! information on regression states
      real(8) :: rssum;                !corrected residual sum-of-squares
      real(8),allocatable :: dxdp(:,:);!partial of filter states to possible jump parameters
      real(8),allocatable :: a12c(:,:);!cross-information matrix used by corresr
      real(8),allocatable :: xpt(:);   !"accepted" jump parameter values

      type jbuffs;
        !** buffer window variables
        integer(4) m;                     !# of scalar meas at each time
        integer(4) np;                    !# of possible jump parameters at time
        real(8) :: time;                  !sample time (min from J2000.0)
        real(8),pointer :: dznorm(:); !normalized filter residuals (1:m)
        real(8),pointer :: zvar(:);   !filter residual variances (1:m)
        real(8),pointer :: dzdp(:,:); !normalized dzdp(1:m,1:np)
      end type;
      type (jbuffs),allocatable :: jbuff(:);

      !** truth variables (for evaluation of TAI jump detection)
!      real(8) xt,at,f
!      common /truth/ xt(9),at(3),f(3,ndim)

      contains;
!**********************************************************************
      subroutine init_jump;
!***    allocates and initializes jump parameters and arrays.

!***    Note: Before calling init_jump, the following variables must be set:
!***       f2rem,f2add,tolmin,nlag,njstat,jmpstat,sigxj,qmin,
!***       ndim,mdim,maxjtim,maxjmp,sigfac,nxjmp,xt,at,f.

!***       maxjmp should be set to nxjmp so that nparm =np

!***       Also jmpstat(njstat), sigxj(njstat), qmin(njstat),
!***       xt(9),at(3),f(3,ndim) must be allocated.

      implicit none;
      integer(4) i,ier;
      !____________________________________________________

      n2dim =(ndim*(ndim+1))/2;  !size of covariance matrix
      nparm =maxjtim*maxjmp;     !max possible # of total jump states
      n2par =(nparm*(nparm+1))/2;!max size of jump state covariance matrix
!      npar2 =nparm+2
      npmax =0.5*nparm;          !or smaller ??
      np2max =(npmax*(npmax+1))/2;
      np =maxjtim*nxjmp;

      !** jump variables
      allocate (jbuff(maxjtim),dxdp(ndim,nparm),ilab(3,nparm),          &
     &          ilxp(2,nparm),a12c(npmax,nparm),                        & !isel(nparm),
     &          xpt(nparm),stat=ier);
      if (ier /= 0) call errhand ('allocate jbuff,... error');

      do i=1,maxjtim;
        allocate (jbuff(i)%dznorm(mdim), jbuff(i)%zvar(mdim),           &
     &            jbuff(i)%dzdp(mdim,nparm), stat=ier);
        if (ier /= 0) call errhand ('allocate jbuff%dznorm,... error');
        jbuff(i)%dznorm(:) =0.d0;
        jbuff(i)%zvar(:) =0.d0;
        jbuff(i)%dzdp(:,:) =0.d0;
        jbuff(i)%m =0;
        jbuff(i)%np =0;
        jbuff(i)%time =0.d0;
      enddo;

      dxdp(:,:) =0.d0;
      itim =0;
      nxp =0;
      rssum =0.d0;
      ibuf2 =0;
      ilab(:,:) =0;
!      isel(:) =.false.

      return;
      end subroutine init_jump;
! ******************************************************************
      subroutine jumptime (phi,time,nt,nx,nxdim,zCount);
!         subroutine jumptime performs the filter time update on the jump
!         parameter sensitivity matrix (dxdg). it also adds jump
!         parameters at the appropriate times defined by tjump and
!         states defined by jmpstat  (read from file jumptim.dat).

!         outputs - jbuff structure (used for stepwise regression)

      implicit none;
      save;

      integer(4),intent(in) :: nt;           !total number of active states (nx+nbias)
      integer(4),intent(in) :: nx;           !number of dynamic states
!      integer(4),intent(in) :: m             !number of scalar measurements at current time
      integer(4),intent(in) :: nxdim;        !row dimension of phi in calling program
      integer(4),intent(in) :: zCount;       !measurement time Z-Count
      real(8),intent(in) :: time;            !current time
      real(8),intent(in) :: phi(nxdim,nt);   !upper partition of state transition matrix

      !** local
      real(8) d1(nx);
      integer(4) i,ip,it,idx,npm1;
      !____________________________________________________

!***     update dxdp
      do ip=1,np;
        call smult (d1,  phi,dxdp(:,ip), nx,nx,1, nxdim,ndim,nx);
        dxdp(1:nx,ip) =d1(1:nx);
      enddo;
      write (6,'(/5("**********"))');
!      call prtmx (dxdp,ndim,nx,np,'dxdp(-) ');

!***      check for time to add parameters
!      if (time+.01.ge.tjump(itim)) then
        ibuf2 =ibuf2+1;
        if (ibuf2 > maxjtim) ibuf2 =1;
        itim =itim+1;
        idx =(ibuf2-1)*nxjmp;

        !**  shift all parameters by one time point
        npm1 =np-nxjmp;
        do ip=1,npm1;
          ilab(:,ip) =ilab(:,ip+nxjmp);
!          isel(ip) =isel(ip+nxjmp)
          dxdp(1:nx,ip) =dxdp(1:nx,ip+nxjmp);
          do it =1,maxjtim;
            jbuff(it)%dzdp(:,ip) =jbuff(it)%dzdp(:,ip+nxjmp);
            jbuff(it)%dzdp(:,ip+nxjmp) =0.d0;
          enddo;
        enddo;

        !**   add new parameters at current time (ibuf2 is location in buffer)
        dxdp(:,npm1+1:npm1+nxjmp) =0.d0;
        do i=1,nxjmp;
          ilab(1,npm1+i) =itim;     !use kCount   ??
          ilab(2,npm1+i) =jmpstat(i);
          ilab(3,npm1+i) =zCount;
!          isel(npm1+i) =.false.
!d         write (6,'("i,jmpstat(i),nx =",3i8)') i,jmpstat(i),nx
          if (jmpstat(i) > 0 .and. jmpstat(i) <= nx)                    &
     &                 dxdp(jmpstat(i),npm1+i) =1.d0;
        enddo;
        write (6,2000) time,np;
 2000   format(' candidate jump parameters added at ',f14.2,', np =',i4)
!d        write (6,'(30(/1x,5(2i7,i10,2x))') ilab
!      call prtmx (dxdp,ndim,nx,np,'dxdp(-s)');
!      endif

      jbuff(ibuf2)%time =time;
      jbuff(ibuf2)%m =0;            !# meas will be reset later in jumpmes
      jbuff(ibuf2)%np =np;
!      write (6,*) time,m,np

      return;
      end subroutine jumptime;
! *****************************************************************
      subroutine jumpmeas (h,g,zvar,dz,im,nx,lEdit);
!***      jumpmes performs the measurement update on the jump
!***      parameter sensitivity matrix (dxdg).

      implicit none;

      real(8),intent(in) :: h(ndim);         !meas partial vector
      real(8),intent(in) :: g(ndim);         !Kalman gain matrix
      real(8),intent(in) :: zvar;            !filter residual variance
      real(8),intent(in) :: dz;              !filter residual
      integer(4),intent(in) :: im;           !measurement index
      integer(4),intent(in) :: nx;           !number of dynamic states
      integer(4),intent(in) :: lEdit;        !>0 if meas was edited in filter

      !** local
      real(8) :: dzdp(np),zsigi,dznorm;
      integer(4) ip;
      !_______________________________________________

      call smult (dzdp,  h,dxdp, 1,nx,np, 1,ndim,1);
!      dzdp(1:np) =-dzdp(1:np)
!      call prtmx (h,1,1,nt,'h       ')
!      call prtmx (dzdp,1,1,m,'dzdp    ')

!***      save output residuals
      zsigi =1.d0/sqrt(zvar);
      dznorm =dz*zsigi;

      jbuff(ibuf2)%dznorm(im) =dznorm;
      jbuff(ibuf2)%zvar(im) =zvar;
      jbuff(ibuf2)%dzdp(im,1:np) =dzdp(1:np)*zsigi;
!      write (6,'(1x,5g13.5)') dznorm, jbuff(ibuf2)%dzdp(im,1:np)

!***     update partial of filter state estimate error, if meas was not edited
      if (lEdit == 0) then;
        do ip=1,np;
          dxdp(1:nx,ip) =dxdp(1:nx,ip) -g(1:nx)*dzdp(ip);
        enddo;
!d      call prtmx (dxdp,ndim,nx,np,'dxdp2   ')
      endif;

      jbuff(ibuf2)%m =max(jbuff(ibuf2)%m,im);

      return;
      end subroutine jumpmeas;
! *****************************************************************
      subroutine glrjmp (x1,  x,p,  nx);
!         subroutine glrjmp estimates the jump states and times using
!         the kalman filter residuals as data for the glr algorithm.
!         note: prior information (sigxj) is used

!           inputs:
!             sigxj(nxjmp) - 1-sigma uncertainty on jump states
!             np - maximum number of parameters estimated
!             nxp -
!             nxjmp - number of jump states
!             f2add - f-statistic to add a parameter (e.g., 7.0)

      implicit none;

      integer(4),intent(in) :: nx;
      real(8),intent(inout) :: x(nx);           !filter state
      real(8),intent(inout) :: p(nx*(nx+1)/2);  !filter covariance
      real(8),intent(out) :: x1(nx);            !filter state vector corrected for accepted jumps

      !** local
      integer(4) iter,it,mt,i,j,k,l,ibuf,m,np1,im,ip,imax;
      real(8) a(np), b(np), dzdp(np);
      real(8) xp(np),dz(mdim),sig0(nxjmp);
      real(8) rss,time,t,pz,fmax,dzc,xpe,f,xpm,xsig,rss1,dxp;
      !____________________________________________________

      iter=0;
!      maxit=5
                                                call timer ('covarsum');
      it=0;
      rss=0.;
      mt=0;
      a(:) =0.d0;
      b(:) =0.d0;

!         add prior on jump magnitudes
      l=0;
      do i=1,np;
        l=l+i;
        k=mod(i-1,nxjmp)+1;
        sig0(k)=sigfac*sigxj(k);
        a(i)=1.d0/sig0(k)**2;
      enddo;
!      write(6,'(1x,10g12.5)') (a(i),i=1,np)
!      call prtsmx(a,j,'a0      ')

      time =jbuff(ibuf2)%time;   !tbuf(1,ibuf2)

!         read data file
!         assume that dz and dzdp are normalized by sqrt(c)
      do ibuf=1,maxjtim;
        t =jbuff(ibuf)%time;        !tbuf(1,ibuf)
        m =jbuff(ibuf)%m;           !tbuf(2,ibuf)
        np1 =jbuff(ibuf)%np;        !tbuf(3,ibuf)
!        write (6,'(1x,f10.2,2i5)') t,m,np1
        if (np1 > np) call errhand ('glrjmp np1 > np');
!        it =it+1
        do im=1,m;
          dz(im) =jbuff(ibuf)%dznorm(im);
          pz =jbuff(ibuf)%zvar(im);
          dzdp(1:np1) =jbuff(ibuf)%dzdp(im,1:np1);
          l =0;
          mt =mt+1;
          rss =rss+dz(im)**2;
          !**  compute information arrays (assumes non-zero intercept)
          do ip=1,np1;
            dzc=dzdp(ip);   !=dzdp(ip)/pz
            b(ip)=b(ip)+dz(im)*dzc;
            a(ip)=a(ip)+dzc**2;
          enddo;
        enddo;

      enddo;

      rssum =rssum+rss;
!      write (6,*) 'completed information accumulation. np1,np= ',np1,np
      write (6,'(" glrjmp sos, rssum = ",2g12.5)') rss,rssum;
                                                call timer ('covarsum');

!***     test for jumps (glr exceeding threshold)
      fmax =f2add;
      imax =0;
      l=0;
      do ip=1,np;
        l =l+ip;
        xpe =b(ip)/a(ip);
        f =xpe**2*a(ip);
        it =ilab(1,ip);
        if (f > fmax .and. (itim-it) >= nlag) then;
          imax =ip;
          fmax =f;
          xpm =xpe;
          xsig =sqrt(1.d0/a(ip));
        endif;
      enddo;

      if (imax > 0) then;
        !**  jump parameter estimated - correct state and residuals
        write (6,1000) time,imax,ilab(:,imax),fmax,xpm,xsig;
 1000   format(/'at time ',f10.2,', parameter ',i4,' (label =',i9,'/',  &
     &        i4,'/',i10,'), fmax =',g12.5/' estimate =',g12.5,         &
     &        ', sig =',g12.5)

        nxp =nxp+1;
        xp(nxp) =xpm;
        ilxp(1,nxp) =ilab(3,imax);
        ilxp(2,nxp) =ilab(2,imax);

        rss1 =0.;
        do it=1,maxjtim;
          m =jbuff(it)%m;
          do im=1,m;
            dzc =jbuff(it)%dzdp(im,imax);
            jbuff(it)%dznorm(im) =0.d0;
            rss1 =rss1 +jbuff(it)%dznorm(im)**2;
          enddo;
        enddo;

        write (6,'(" original and corrected sos =",2g13.5)') rss,rss1;
        write (6,'(/1x,10g12.5)') (p((l*(l+1))/2),l=1,nx);
        call prtmx (dxdp(1,imax),1,1,nx,'dxdpimax');

        l=0;
        do i=1,nx;
          x(i) =x(i)+dxdp(i,imax)*xpm;
          dxp =dxdp(i,imax)*xsig**2;
          do j=1,i;
            l =l+1;
            p(l) =p(l)+dxp*dxdp(j,imax);
          enddo;
        enddo;
        write (6,'(/1x,10g12.5)') (p((l*(l+1))/2),l=1,nx);

        x1(:) =x(:);

        write (6,'(/"ilxp = "/50(1x,6(i8,i6)/))') (ilxp(:,i),i=1,nxp);
        write (6,'("xp   = "/50(1x,6g14.6/))') (xp(i),i=1,nxp);

      endif;

      return;
      end subroutine glrjmp;
! *****************************************************************
      subroutine estjmpj (xcor,  x,p,  nx);
!         subroutine estjmp estimates the jump states and times using
!         the kalman filter residuals as data for least squares.
!         forward stepwise regression is used to determine significant jump states
!         note: prior information (sigxj) is used

!         this version (estjmpr) is used for real-time jump detection

!           inputs:
!             jbuff(npar+2,mdim,maxjtim) - buffer containing
!                time, # meas, # parm at each meas, information partials
!                wrt the jump parameters
!             ilab(3,np) - labels of jump parameters
!             sigxj(nxjmp) - 1-sigma uncertainty on jump states
!             np - maximum number of parameters estimated
!             nxjmp - number of jump states
!             f2rem - f-statistic to remove a parameter (e.g., 0.5)
!             f2add - f-statistic to add a parameter (e.g., 16.0 for 4-sigma)
!             tolmin - minimum tolerance required to include a parameter
!                      (e.g., 0.01)
!             sigfac - factor multiplying sigxj used as
!                     prior information on regression states

!         modified 11/20/90 to include biases

      implicit none;

      integer(4),intent(in) :: nx;              !number of filter states
      real(8),intent(inout) :: x(nx);           !filter state vector corrected for accepted jumps leaving window
      real(8),intent(inout) :: p(nx*(nx+1)/2);  !filter error covariance corrected for accepted jumps leaving window
      real(8),intent(out) :: xcor(nx);            !filter state vector corrected for candidate accepted jumps

      !** local
      logical(4),allocatable :: isel(:);
      integer(4),allocatable :: mapx(:);
      integer(4) iter,it,i,j,k,m,njAdd,l,l1,mt,ibuf,np1,np2,ip,im,jp;
      integer(4) ladd,ier;
      integer(4) :: nbias =0;    !no bias states in regression
      real(8) rss,time,t,dzc,eps,epsm,xpm,rss1,rsos;
      real(8),allocatable :: a(:), b(:), d1(:);
      real(8),allocatable :: xp(:),dz(:),p0i(:),sig0(:);
      real(8),allocatable :: a1(:),b1(:),a12(:),dzdp(:);
      !____________________________________________________

      allocate (a(n2par), b(np), d1(np), dzdp(np),                      &
     &          xp(np),dz(mdim),p0i(npmax),sig0(nxjmp),                 &
     &          a1(np2max),b1(npmax),a12(npmax),                        &
     &          isel(np),mapx(npmax),stat=ier);
      if (ier /= 0) call errhand ('estjmpj allocate error');

      iter =0;
!      maxit =5    !max iterations
                                                call timer ('covarsum');
      it =0;
      rss =0.;
      mt =0;
      a(:) =0.d0;
      b(:) =0.d0;
      isel(:) =.false.;

!***     add prior on jump magnitudes
      l=0;
      do i=1,np;
        l =l+i;
        k =mod(i-1,nxjmp)+1;
        sig0(k) =sigfac*sigxj(k);
        b(i) =0.d0;        !prior nominal is 0
        a(l) =1.d0/sig0(k)**2;
      enddo;
!      call prtsmx(a,j,'a0      ')

      time =jbuff(ibuf2)%time;   !tbuf(1,ibuf2)

!***     read data buffer and form information arrays
!***     assume that dz and dzdp are normalized by sqrt(c)
      do ibuf=1,maxjtim;
        t =jbuff(ibuf)%time;        !tbuf(1,ibuf)
        m =jbuff(ibuf)%m;           !tbuf(2,ibuf)
        np1 =jbuff(ibuf)%np;        !tbuf(3,ibuf)
        if (np1 > np) call errhand ('estjmpj error: np1 > np');
!        np2 =np1
!        it =it+1

        do im=1,m;
          dz(im) =jbuff(ibuf)%dznorm(im);
          dzdp(1:np1) =jbuff(ibuf)%dzdp(im,1:np1);
          l =0;
          mt =mt+1;
          rss =rss +dz(im)**2;
!          if (t == time) rssum =rssum +dz(im)**2   !cumulative uncorrected filter residual RSS
          !**   compute information arrays (assumes non-zero intercept)
!         write (6,'("ibuf ",i3,", t=",f12.2,", im=",i4," dz =",g13.5)')&
!    &              ibuf,t,im,dz(im);
          do ip=1,np1;
            it =ilab(1,ip);
            if ((itim-it) > nlag) then;   !this test is important ??
              l =(ip*(ip-1))/2;
              dzc =dzdp(ip);
              b(ip) =b(ip)+dz(im)*dzc;
!             write (6,'("ip ",i3," dzdp,b =",2g13.5)') ip,dzc,b(ip);
              do jp=1,ip;
                l =l+1;
!                if ((itim-ilab(1,jp)) > nlag) then      !test has no effect
                  a(l) =a(l)+dzdp(jp)*dzc;
!                endif
              enddo;
            endif;
          enddo;     !___ end ip loop
        enddo;    !___ end im loop
      enddo;   !___ end ibuf loop

!      rssum =rssum+rss     !??
!      write (6,*) 'completed information accumulation. np1,np= ',np1,np
      write (6,'(" estjmpj: time =",f14.3,", for ",i6,                  &
     &    " meas in window, filter residual sos =",g12.5)') time, mt,rss;
!     &       " cumulative uncorrected residual rss =",g12.5)') time, mt,rss,rssum
      np2 =np;
                                                call timer ('covarsum');
                                                call timer ('stepregr');

!***    start of stepwise regression loop:
!***    form information matrix & vector for estimated states
      ladd =1;
      eps =1.d-5;

      do while (ladd /= 0);
        iter =iter+1;
        l1=0;
        njAdd =0;
        do i=1,np2;
          if (isel(i)) then;
            njAdd =njAdd+1;
            if (njAdd > npmax)                                          &
     &               call errhand ('estjmpj error: njAdd > npmax');
            mapx(njAdd) =i;
!            j =mod(i-1,nxjmp)+1
!            p0i(njAdd) =1.d0/sig0(j)**2     !not used
            b1(njAdd) =b(i);
            l =(i*(i-1))/2;
            do j=1,i;
              if (isel(j)) then;
                l1 =l1+1;
                a1(l1) =a(l+j);
              endif;
            enddo;
          endif;
        enddo;

        rsos =rss;

        if (njAdd > 0) then;
                                                call timer ('step1   ');
!          call prtsmx (a1,njAdd,'a info  ')
          !**    compute least squares estimate
          call sinv (ier,epsm,  a1,  njAdd,eps,.true.);
!          call prtsmx (a1,njAdd,'a inv   ')
          if (ier < 0) call errhand ('matrix inversion error');
          call symult(xp,  a1,b1,njAdd);

          l=0;
          do i=1,njAdd;
            l =l+i;
            d1(i) =sqrt(a1(l));
          enddo;
          if (iter == 1) then;
            call prtmx (xp,1,1,njAdd,'xp      ');
            call prtmx (d1,1,1,njAdd,'xp sig  ');
          endif;

          rsos =rsos -dot_product(b1(1:njAdd),xp(1:njAdd));
          write(6,2200) iter,rss,rsos,mt,njAdd;
 2200     format(/'**************'/                                     &
     &     ' at iteration ',i3,', data sum-of-squares =',g12.5,         &
     &     ', resid. sos =',g12.5,', # meas=',i5,', # added parm =',i3/)

                                                call timer ('step1   ');
        endif;

        !**   check f-statistic to add or remove states from regression.
        !**   Output is ladd and isel
        call addremr (ladd,  isel, a,b,a1,xp,np2,njAdd,mt,f2add,f2rem,  &
     &                tolmin,rsos,ilab,mapx,nbias,itim,nlag);

        !** if states added or removed, recompute regression
      enddo;   !___ end do while loop

                                                call timer ('stepregr');
      xcor(1:nx) =x(1:nx);

      if (njAdd > 0) then;
        !**  Note: individual rss1 below may be bigger or smaller than rsos above,
        !**  but sum over all jump parameters matches

        do jp=1,njAdd;
          xpm =xp(jp);
          ip =mapx(jp);
!          xvar =a1((jp*(jp+1))/2)
          !**  jump parameter estimated - correct state and residuals
          k =mod(ip-1,nxjmp)+1;
          rss1 =(xpm/sig0(k))**2;
          do it=1,maxjtim;
            m =jbuff(it)%m;      !tbuf(2,it)
            do im=1,m;
              dzc =jbuff(it)%dzdp(im,ip)*xpm;
              rss1 =rss1 +(jbuff(it)%dznorm(im)-dzc)**2;
            enddo;
          enddo;
          write (6,'(" Parameter ",i8,"/",i4,                           &
     &          ": original and corrected sos =",2g13.5)')              &
     &          ilab(1:2,ip),rss,rss1;

          xcor(1:nx) =xcor(1:nx) +dxdp(1:nx,ip)*xpm;

          !** If jump parameter will be dropped on next time step
          !** (ip <= nxjmp), then update filter state and covariance

          if (ip <= nxjmp) then;
            nxp =nxp+1;    !don't want to store jump states indefinitely: nxp is reset in filtGPS
            xpt(nxp) =xpm;
            ilxp(1,nxp) =ilab(3,ip);
            ilxp(2,nxp) =ilab(2,ip);
            write (6,1000) time,ip,ilab(:,ip),xpm,d1(jp);
 1000       format(/'at time ',f10.2,', parameter ',i4,' (label =',     &
     &        i9,'/',i4,'/',i10,')'/' estimate =',g12.5,', sig =',g12.5)
            x(1:nx) =x(1:nx) +dxdp(1:nx,ip)*xpm;

            write (6,'(/"Jump parameter leaving sliding window:",       &
     &               " update state and covariance")');
            write (6,'("ilxp = "/30(1x,6(i9,"/",i4)/))') ilxp(:,1:nxp);
            write (6,'("xp   = "/30(1x,6g14.6/))') xpt(1:nxp);
!          else
!            write (6,'(/"jump parameter not finalized: ip =",i6)')ip
          endif;
        enddo;    !___ end jp loop

        !** calculate the corrected residuals and
        !** residual partials given the estimated bias parameters (uses jbuff(i)%dzdp)
        call corresr (dzdp,  a,a1,xp,isel,njAdd,np2,nxjmp);

        !** update the filter covariance matrix to account for the modified filter state
        !** (only if jump parameter will be dropped on next time step)
        l=0;
        do i=1,nx;
          do jp=1,njAdd;
            ip =mapx(jp);
            b1(jp) =0.d0;
            if (ip <= nxjmp) b1(jp) =dxdp(i,ip);
          enddo;
          call symult (a12,  a1,b1,njAdd);
          do j=1,i;
            l=l+1;
            do jp=1,njAdd;
              ip =mapx(jp);
              if (ip <= nxjmp) p(l) =p(l) +a12(jp)*dxdp(j,ip);
            enddo;
          enddo;
        enddo;
      endif;   !___ end if (njAdd > 0)

      deallocate (a, b, d1, xp,dz,p0i,sig0,a1,b1,a12,isel,mapx);

      return;
      end subroutine estjmpj;
! **********************************************************************
      subroutine addremr (ladd,  isel,  a,b,a1,xp,np2,njAdd,mt,f2add,   &
     &                    f2rem,tolmin,rsos,ilab,mapx,nbias,itim,nlag);
!***     subroutine addrem tests the f-statistic to determine whether
!***     states should be added or removed from the regression

      implicit none;

      integer(4),intent(in) :: np2;       !number of jump parameters in a and b information arrays
      integer(4),intent(in) :: njAdd;     !number of current "accepted" jump parm
      integer(4),intent(in) :: mt;        !total number of (scalar) filter meas in window
      integer(4),intent(in) :: nbias;     !number of bias states in regression
      integer(4),intent(in) :: itim;      !index of current time sample (should be limited ??)
      integer(4),intent(in) :: nlag;      !number of time lags before adding parameter (e.g., 5)
      integer(4),intent(in) :: ilab(3,np2);!labels of parameters
      integer(4),intent(in) :: mapx(npmax);!mapping from accepted states to np2 index
      real(8),intent(in) :: a(np2*(np2+1)/2); !information matrix for candadate jump parameters
      real(8),intent(in) :: b(np2);       !information vector for candadate jump parameters
      real(8),intent(in) :: a1(np2max);   !information array for "accepted" jump parameters
      real(8),intent(in) :: xp(npmax);    !estimates of jump parameters
      real(8),intent(in) :: f2add;        !regression f-test threshold to add state (e.g., 16.0)
      real(8),intent(in) :: f2rem;        !regression f-test threshold to remove state (e.g., 0.5)
      real(8),intent(in) :: tolmin;       !minimum variance ratio required to add a state (e.g. 0.01)
      real(8),intent(in) :: rsos;         !residual sum-of-squares before regression
      logical(4),intent(inout) :: isel(np2); !jump parameter selection flag: 1=selected
      integer(4),intent(out) :: ladd;     !number of parameters added to regression

      !** local
      integer(4) i,j,j1,l,l1,imin,imax,iflag,it;
      real(8) a12(np2),c12(np2),xsig,f,sms,fmin,fmax;
      real(8) a22,b2,c22,tol,ri,fi,df,tolmax,rmax;
      !____________________________________________________

      ladd =0;
      df =mt -njAdd +njAdd;
      sms =rsos/df;

      if (njAdd > 0) then;
        !**   check f-to-remove to remove states from regression
        write (6,'("** F-to-remove check")');
        l =0;
        iflag =0;
        fmin =f2rem;
        imin =0;
        do i=1,njAdd;
          l =l+i;
          xsig =sqrt(a1(l));
          f =xp(i)**2/(a1(l)*sms);
          write(6,2300) mapx(i),xp(i),xsig,f;
 2300     format(' parameter ',i4,', xp =',g12.5,', xsig =',g12.5,      &
     &           ', ftest =',g12.5)
          if (f < fmin) then;
            fmin =f;
            imin =i;
          endif;
        enddo;

        if (imin > 0) then;
          isel(mapx(imin)) =.false.;
          l =(imin*(imin+1))/2;
          write(6,2000) imin,mapx(imin),xp(imin),a1(l);
 2000     format(' **** estimated state ',i4,'(',i4,')',                &
     &         ' removed from regression. xp = ',g12.5,', cov =',g12.5)
          ladd =-1;
          return;
        endif;
      endif;

!***     check f-to-add to add states
                                                call timer ('step-add');
      imax=0;
      fmax =f2add;
      write (6,'("** F-to-add check: mt =",i5,", rsos =",g13.5,         &
     &       ", njadd =",i3)')mt,rsos,njadd;
      do i=1,np2;
        it =ilab(1,i);     !it =time index
!        write (6,*) it,itim,nlag
        if (.not.isel(i) .and. (itim-it) > nlag) then;
          l =(i*(i+1))/2;
          a22 =a(l);
          b2 =b(i);
          j1 =0;
          do j=1,np2;
            if (isel(j)) then;
              j1=j1+1;
              if (j <= i) then;
                l1 =l-i+j;
              else;
                l1 =(j*(j-1))/2+i;
              endif;
              a12(j1) =a(l1);
            endif;
          enddo;

          c22 =a22;
          if (njAdd > 0) then;
            call symult(c12,  a1,a12,njAdd);
            do j=1,njAdd;
              b2 =b2-xp(j)*a12(j);
              c22 =c22-c12(j)*a12(j);
            enddo;
            if (c22 <= 0.d0) c22 =1.d-20;
          endif;

!         if (b2 /= 0.d0) write(6,'(i3,", a22,c22,b2 =",                &
!    &    3g13.5,", b2**2/c22 =",g13.5)') i, a22,c22, b2, b2**2/c22;
          if (a22 > 0.d0) then;
            tol =c22/a22;
            df =mt+(njAdd+1)-(njAdd+1);   !inclusion of prior increases df to offset reduction
            ri =b2/sqrt(c22*rsos);
            fi =(b2**2*df)/(c22*rsos-b2**2);

            if (tol > tolmin .and. fi > fmax) then;
              imax =i;
              fmax =fi;
              tolmax =tol;
              rmax =ri;
            endif;
          endif;

        endif;    !___ end test for not selected and (itim-it) > nlag
      enddo;   !___ end i-loop
                                                call timer ('step-add');

      if (imax > 0) then;
        isel(imax) =.true.;
        write(6,2100) imax, fmax, tolmax,rmax, df;
 2100   format(/' ** tested parameter ',i4,' added to regression.'/     &
     &         ' fmax =',g13.5,', tolmax =',g13.5,', rmax =',g13.5,     &
     &         ', deg-of-freedom =',f6.0)
        ladd=1;
        return;
      endif;

      return;
      end subroutine addremr;
! ****************************************************************
      subroutine corresr (dzdp, a,a1,xp,isel,njAdd,np2,nxjmp);
!***      subroutine corres computes the corrected residuals and
!***      residual partials given the estimated bias parameters
!***      Uses structure jbuff data.
      implicit none;

      integer(4),intent(in) :: njAdd;        !current number of "accepted" jump states
      integer(4),intent(in) :: np2;          !number of possible jump parameters
      integer(4),intent(in) :: nxjmp;        !actual number of jump states in filter
      logical(4),intent(in) :: isel(np2);    !jump parameter selection flag: 1=selected
      real(8),intent(in) :: a(np2*(np2+1)/2); !information matrix for all possible jump parameters
      real(8),intent(in) :: a1(njAdd*(njAdd+1)/2);!covariance matrix for "accepted" jump parameters
      real(8),intent(in) :: xp(njAdd);       !estimates of jump parameters
      real(8),intent(out) :: dzdp(np2);      !partial of meas residual wrt jump parameters

      !** local
      integer(4) i,j,j1,l,l1,ibuf,m,np1,im,ip;
      real(8) dum(njAdd),dz(mdim),h1(njAdd),t,yp,ha;
!      real(8) a12c(njAdd,np2)           !
!      common /array/ a12c(npmax,nparm)
      !____________________________________________________


!***    save cross information between adjusted and all other parameters
      do i=1,np2;
        l=(i*(i+1))/2;
        j1=0;
        do j=1,np2;
          if (isel(j)) then;
            j1 =j1+1;
            if (j <= i) then;
              l1 =l-i+j;
            else;
              l1 =(j*(j-1))/2+i;
            endif;
            a12c(j1,i) =a(l1);
          endif;
        enddo;
        if (j1 /= njAdd) call errhand ('corresr error: j1 /= njAdd');
      enddo;

!***     save inv(a1)*a12c in a12c
      do i=1,np2;
        call symult (dum,  a1,a12c(:,i),njAdd);
        a12c(1:njAdd,i) =dum(1:njAdd);
      enddo;

!***     now form corrected residuals and residual partials

      do ibuf=1,maxjtim;
        t =jbuff(ibuf)%time;
        m =jbuff(ibuf)%m;
        np1 =jbuff(ibuf)%np;
        dzdp(:) =0.d0;
        do im=1,m;
          dz(im) =jbuff(ibuf)%dznorm(im);
!          pz =jbuff(ibuf)%zvar(im)
          dzdp(1:np1) =jbuff(ibuf)%dzdp(im,1:np1);
          yp=0.d0;
          j=0;
          h1(:) =0.d0;
          do i=1,nxjmp;    !np1
            if (isel(i)) then;
              j=j+1;
              h1(j) =dzdp(i);
              yp =yp +dzdp(i)*xp(j);
            endif;
          enddo;

          jbuff(ibuf)%dznorm(im) =dz(im)-yp;

          !**  compute corrected partials
          do ip=1,np2;
!            if (np1 >= ip1) then      !np1 test used to ??
              call smult (ha,  h1,a12c(1,ip), 1,njAdd,1, 1,npmax,1);
              jbuff(ibuf)%dzdp(im,ip) =jbuff(ibuf)%dzdp(im,ip) -ha;
!            endif
          enddo;

        enddo;
      enddo;

      return;
      end subroutine corresr;
!******************************************************************
      subroutine timer(a);
      character(*) a;
      return;
      end subroutine timer;
!******************************************************************
      end module jumpCom;
