!**  simMeas.f90 generates a file of simulated measurements for testing the MLE_parameter ID program.
!**  Module modelMod contains the model-specific information

!**  Search for "###" to find parameters that are typically changed

!**************************************
! The author grants the user a non-exclusive, worldwide, royalty-free copyright license to

! 1. reproduce and modify the software for your own purposes, and
! 2. to distribute the software provided that you give credit to the author,
!    do not make claims against the author or Wiley-Interscience,
!    and mark any modifications as your own.  If the software is incorporated in
!    a commercial product, you agree to defend and indemnify this author and
!    Wiley-Interscience against any losses, damages and costs arising from claims,
!    lawsuits and other legal actions brought by a third party.

! The software is provided on an as is basis, without warranties or conditions
! of any kind, either express or implied (including any warranties or conditions
! of title, non-infringement, merchantability or fitness for a particular purpose).
! The user is solely responsible for determining the appropriateness of using and
! distributing the program and assumes all risks associated with its exercise of
! rights under this agreement.
!**************************************

!** sample compile and execute using Intel compiler:

!**     Include in ifort.cfg:
!**   /check:bounds /traceback /include:\kf_book\distribSWm\library /link /nodefaultlib:msvcrt.lib
!**     \kf_book\distribSWm\library\libkf_bnd_ifort.lib

!**     To compile and execute:
!**   set ifortcfg=\kf_book\distribSWm\ifort.cfg
!**   ifort modelMod.f90 /c
!**   ifort simMeas.f90 modelMod.obj
!**   simMeas

!*****************************************************************************
      program simMeas
      !** generates a file of simulated measurements for testing the MLE_parameter ID program.
      use modelMod, only:nEstf,ndynam,xstate,phi,qd,mmax,mmeas,nsensor,   &
     &                   defaultp,get_yh,t0;
      use filterPrec, only: real_kind1;   !precision (4 or 8)
      implicit none;

      integer(4) :: ix =191919;           !seed of random number generator
      integer(4) i,lenferr,nargs,ier,nmeas,im,isensor;
      character(80) line;
      character(80) setupFile;
      character(80) :: measFile ="sim_4state.txt"    !name of file containing measurement data
      real(8) xt(nEstf),yp(mmeas),yt,dy,h(nEstf),xp(nestf);
      real(8) t,rnor,rvar(mmeas,nsensor)
      real(8) :: dt =0.25d0;  !integration time sec (sec)   ### (also 0.1 or 0.2)
      real(8) :: tmax =100.d0;!maximum simulation time (sec)  ### (also 50)
      !______________________________________________

      if (real_kind1 /= 8) call errhand ('real_kind1 /= 8')

! ....  Get run control file name
!      lenferr =nargs()-1;         !get number of command line arguments
!      if (lenferr == 0) then;
!        write (6,'(" Input setup file:")');
!        read (5,'(a40)') setupFile;
!      else;
!        call getarg(1,line,ier);    !get command line argument
!        read (line,*) setupFile;
!        if (ier < 0) stop 'ERROR IN RETRIEVING setupFile';
!      endif;

!** read text control file to set the xstate structure.  Note that variables nEstf and xlabel
!** must have been set by module modelMod before calling setup.  Setup also sets arrays
!** mapx and mapxi that define the mappings between estimated and total states

!      call read_setup (setupFile);

      !** get default values from modelMod
      call defaultp;

      !** set simulation values
      rvar(:,1) =(/ 0.5, 0.001 /);    !true meas noise variance, sensor 1 ###
      rvar(:,2) =(/ 0.5, 0.001 /);    !true meas noise variance, sensor 2

      !** set true xstate values for this case ###
      xstate(:)%x0 =0.d0;
      xstate(:)%sig0 =(/ 2., 4., 3., 4. /);
      xstate(:)%Qs =(/0.d0, 0.8d0, 0.d0, 1.6d0 /);
      xstate(:)%const(1) =(/0.d0, 0.1d0, 0.d0, 0.2d0 /);

      xt(:) =0.d0;    !or set randomly from sig0
      xt(1) =10.;     !set x-component to +10  ###
      t0 =0.d0

      !** compute phi and q from parameter vector xparm
      call get_phiq (phi,qd,xp,   t0,dt,xt,.false.);
!      call prtsmx (qd,ndynam,"qd")
      nmeas =0;
      t =t0;
      yt =rnor(ix);   !initialize rnor seed

      open (10,file=measFile,status='unknown');

      !** time loop
      do;
        t =t+dt;
        if (t > tmax) exit;

        !** time update of state with added diagonal process noise
        xt(:ndynam) =matmul(phi,xt);
        do i=1,ndynam;
          xt(i) =xt(i) +sqrt(qd(i*(i+1)/2))*rnor(0);
        enddo;

        nmeas =nmeas+1;
        if (nmeas > mmax) stop 'sim_meas error: nmeas > mmax';

        !** compute measurements and add measurement noise
        isensor =1;         !only use 1 sensor for this test
        yt =0.d0;
        do im=1,mmeas;
          call get_yh (yp(im),dy,h,  yt,xt,isensor,im);
          yp(im) =yp(im) +sqrt(rvar(im,isensor))*rnor(0);
        enddo;
        write (10,'(g16.7,i3,5g16.7)') t,isensor,yp(:),rvar(:,isensor);
      enddo;

      close (10);
      end program;