!** tst_chebv.f tests the accuracy of Cholesky
!** inversion of normal equation matrix for Chebyshev polynomial model

!**************************************
! The author grants the user a non-exclusive, worldwide, royalty-free copyright license to

! 1. reproduce and modify the software for your own purposes, and
! 2. to distribute the software provided that you give credit to the author,
!    do not make claims against the author or Wiley-Interscience,
!    and mark any modifications as your own.  If the software is incorporated in
!    a commercial product, you  agree to defend and indemnify this author and
!    Wiley-Interscience against any losses, damages and costs arising from claims,
!    lawsuits and other legal actions brought by a third party.

! The software is provided on an as is basis, without warranties or conditions
! of any kind, either express or implied (including any warranties or conditions
! of title, non-infringement, merchantability or fitness for a particular purpose).
! The user is solely responsible for determining the appropriateness of using and
! distributing the program and assumes all risks associated with its exercise of
! rights under this agreement.
!**************************************

!** sample compile and execute using Intel compiler:

!**   ifort tst_chebv.f90 ..\..\library\libkf.lib
!**   ifort tst_chebv.f90 ..\..\library\libkf_bnd_ifort.lib    (checks array bounds)
!**     (libkf.lib includes rnoruni, l1normM, l2NormM, sinv, symult)

!**     To execute:
!**   tst_chebv >tsti_cheb.plt
!**************************************

      implicit none;
      integer(4),parameter :: nd =40, mmax=101, nsamp=100, lwork=10000;          !use nd=40 for hdiag =T
      integer(4) i,j,k,n,m,kk,ier,ix,nn2,isamp,nmax;
      logical(4) :: scaleh =.false.;    !normally all false
      real(8) dt,stat(3,6);
      real(8) a(nd,nd),al(nd*(nd+1)/2),r(nd*(nd+3)/2);
      real(8) b(nd,nd),c(nd,nd),esum,eps,x(nd);
      real(8) h(mmax,nd+1),bb(nd),dx(nd);
      real(8) cnb,xe(nd),y(mmax),emax,errm,ystat(4,2*mmax);
      real(8),external :: rnor,l2normM,l1normM;
      !____________________________________________________

      open (10,file='tstcheby.plt',status='unknown');
      write (10,1100);
 1100 format('VARIABLES ="n", "Cholesky-RMS-error", ',                  &
     &  '"Cholesky-max-error", "Cholesky-precision"')

      m =mmax;
      nmax =20;
      dt =1.d0/(mmax-1);

      write (6,1200);
 1200 format('VARIABLES ="n", "Cholesky-precision", ',                  &
     &  '"Cholesky-Frobenius-error", "Cholesky-L1-error"')

      do n=4,nmax;     !matrix order
        call get_haPoly (h,al,mmax,nd,n,scaleh,dt);      !polynomial

        kk =0;
        do j=1,n;
          do k=1,j;
            kk =kk+1;
            a(k,j) =al(kk);
            a(j,k) =al(kk);
          enddo;
        enddo;

        ix =191919;
        cnb =rnor(ix);
        stat(:,:) =0.d0;
        nn2 =(n*(n+1))/2;

        do isamp =1,nsamp;
          do i=1,n;
            x(i) =rnor(0);
          enddo;
!          if (scaleh) x(1:4) =0.01*x(1:4)    !###
          y(:) =matmul(h(:,:n),x(:n));
          h(:,n+1) =y(:);
          forall (i=1:n) bb(i) =dot_product(y,h(:,i));
!          bb(:) =matmul(y,h(:,:n))     !=trn(h)*y

          !********************* method 1: normal equations using Cholesky factor
          eps =1.d-3;
          r(:nn2) =al(:);
          call sinv (ier,errm,  r,  n,eps,.false.);
          if (ier < 0) then;
!            write (6,'("sinv error",i3)') ier
            esum =1.d20;
            emax =1.d20;
            errm =1.d20;
          else;
            call symult (xe,  r,bb,n);
            dx(:n) =xe(:n) -x(:n);
            stat(1,1) =stat(1,1) +sum(dx(:n)**2);
            stat(2,1) =max(stat(2,1),maxval(abs(dx(:n))));
            stat(3,1) =stat(3,1) +1;

            kk =0;
            do j=1,n;
              do k=1,j;
                kk =kk+1;
                b(k,j) =r(kk);
                b(j,k) =r(kk);
              enddo;
            enddo;
            c(:n,:n) =matmul(a(:n,:n),b(:n,:n));
            forall (i=1:n) c(i,i) =c(i,i)-1.d0;
            esum =l2normM(c,nd,n);     !sqrt(sum(c(:n,:n)**2)/n**2)
            emax =l1normM(c,nd,n);  !maxval(abs(c(:n,:n)))
          endif;

          if (isamp == 1) then;
            write (6,1000) n,log10(errm),log10(esum),log10(emax);
1000        format (i3,16g13.5)
          endif;

          if (n == nmax) call predict (ystat, xe,x,n,m,dt,.false.);

        enddo;    !___ end isamp loop

        do i=1,1;
          if (stat(1,i) <= 0.d0) then;
            stat(1,i) =1.d40;
          else;
            stat(1,i) =stat(1,i)/stat(3,i);
          endif;
          if (stat(2,i) <= 0.d0) stat(2,i) =1.d40;
        enddo;

        write (10,1000) n,log10(sqrt(stat(1,1))),                       &
     &                 log10(stat(2,1)), log10(errm);

        if (n == nmax) then;
          open (12,file="predictErr.plt",status='unknown');
          write (12,3000);
 3000     format ('VARIABLES ="Time", "log10(Y-Ymodel)", ',             &
     &             '"log10(Y-meas)", "Y-meas"')
          do i=1,2*mmax;
            write (12,'(4g13.5)') dt*i,0.5*log10(ystat(2,i)/ystat(1,i)),&
     &               0.5*log10(ystat(3,i)/ystat(1,i)),                  &
     &               sqrt(ystat(3,i)/ystat(1,i));
          enddo;
        endif;
      enddo;   !___ end n (order) loop

      stop;
      end program;
!*******************************************************************
      subroutine get_haPoly (h,al,mmax,nd,n,scaleh,dt);
      implicit none;

      integer(4),intent(in) :: mmax,nd,n;
      logical(4),intent(in) :: scaleh;
      real(8),intent(in) :: dt;
      real(8),intent(out) :: h(mmax,nd+1),al(nd*(nd+1)/2);

      integer(4) it,jt,j,k,kk;
      real(8) t,hnorm(mmax),c(n),cheb,dfdc(n),a,b;
      !_________________________________________

      al(:) =0.d0;
      h(:,:) =0.d0;
!      dt =1.d0/(mmax-1)
      a =0.d0;
      b =(mmax-1)*dt;
      c(:) =0.d0;    !set

      do it=0,mmax-1;
        t =it*dt;
        kk =0;
        jt =it+1;
!        jt =mmax+1-(it+1)    !it+1      !test re-order ###
        call chebyPoly (cheb,dfdc,  a,b,c,n,t);
!        write (6,'("dfdc =",10g13.5)') dfdc(:n)
        h(jt,:n) =dfdc(:);
        do j=1,n;
!          if (scaleh .and. j <= 4) h(jt,j) =h(jt,j)*100.d0    !test effect of scaling: makes QR worse and SVD slightly worse
          do k=1,j;
            kk =kk+1;
            al(kk) =al(kk) +h(jt,j)*h(jt,k);
          enddo;
        enddo;
        hnorm(jt) =sqrt(sum(h(jt,1:n)**2));
      enddo;    !end time loop

!      write (*,'("n =",i3,", hnorm =",10(/10g13.5))') n,hnorm(1:mmax)

      return;
      end subroutine get_haPoly;
!******************************************
      subroutine chebyPoly (fx,dfdc,  x1,x2,c,m,x);
      !** evaluates Chebyshev polynomial at a specific point x, and also computes
      !** partial derivatives WRT the polynomial coefficients
      implicit none;
      integer(4),intent(in) :: m;   !order of model
      real(8),intent(in) :: x;      !intermediate point between x1 and x2 for evaluation
      real(8),intent(in) :: x1,x2;  !start and end points of interval
      real(8),intent(in) :: c(m);   !chebyshev coeff
      real(8),intent(out) :: fx;    !interpolated value at x
      real(8),intent(out) :: dfdc(m);!interpolated partials of fx wrt c

      integer(4) j;
      real(8) d,dd,sv,y,y2,dddc(m),ddddc(m),ddddcL(m);
      !__________________________________________

      if ((x-x1)*(x-x2) > 0.) stop 'chebev error: x not in range';
      d =0.d0;
      dd =0.d0;
      y =(2.d0*x-x1-x2)/(x2-x1);
      y2 =2.d0*y;
      dddc(:) =0.d0;
      ddddc(:) =0.d0;

      do j =m,2,-1;
        sv =d;
        ddddcL(:) =dddc(:);
        dddc(:) =y2*dddc(:) -ddddc(:);
        dddc(j) =dddc(j) +1.d0;
        d =y2*d -dd +c(j);
        dd =sv;
        ddddc(:) =ddddcL(:);
      enddo;

      fx =y*d -dd +0.5d0*c(1);
      dfdc(:) =y*dddc(:) -ddddc(:);
      dfdc(1) =dfdc(1) +0.5d0;

      return;
      end subroutine chebyPoly;
!***********************************************
      subroutine predict (ystat,  xe,x,n,m,dt,hdiag);
!***    predicts measurements for double original interval
      implicit none;
      integer(4),intent(in) :: n,m;
      logical(4),intent(in) :: hdiag;
      real(8),intent(in) :: xe(n),x(n),dt;
      real(8),intent(inout) :: ystat(4,2*m);

      integer(4) i,mmax;
      real(8) h(2*m,n+1),al(n*(n+1)/2),ye(2*m),yt(2*m);
      logical(4) :: scaleh =.false.;
      !_________________________________________________

      mmax =2*m;
      if (hdiag) then;
!        call get_haDiag (h,al,mmax,nd,n,scaleh)       !bi-diagonal h
        stop 'invalid use of haDiag';
      else;
        call get_haPoly (h,al,mmax,n,n,scaleh,dt);
      endif;

      ye(:) =matmul(h(:mmax,:n),xe(:n));
      yt(:) =matmul(h(:mmax,:n),x(:n));

!      open (12,file="predictErr.plt",status='unknown')
!      write (12,1000)
! 1000 format ('VARIABLES ="Time", "Yrres-SVD","Yrres-QR", "Y-meas"')
      do i=1,mmax;
!        write (12,'(4g13.5)') dt*i,ysvd(i)-yt(i),yqr(i)-yt(i), yt(i)
        ystat(1,i) =ystat(1,i) +1.d0;
        ystat(2,i) =ystat(2,i) +(ye(i)-yt(i))**2;
        ystat(3,i) =ystat(3,i) +yt(i)**2;
      enddo;

      return;
      end subroutine predict;
