!** tst_invert.f tests the accuracy of Cholesky and Gaussian elimination
!** inversion of normal equation matrix for polynomial model

!**************************************
! The author grants the user a non-exclusive, worldwide, royalty-free copyright license to
! 1. reproduce and modify the software for your own purposes, and
! 2. to distribute the software provided that you give credit to the author,
!    do not make claims against the author or Wiley-Interscience,
!    and mark any modifications as your own.  If the software is incorporated in
!    a commercial product, you  agree to defend and indemnify this author and
!    Wiley-Interscience against any losses, damages and costs arising from claims,
!    lawsuits and other legal actions brought by a third party.

! The software is provided on an as is basis, without warranties or conditions
! of any kind, either express or implied (including any warranties or conditions
! of title, non-infringement, merchantability or fitness for a particular purpose).
! The user is solely responsible for determining the appropriateness of using and
! distributing the program and assumes all risks associated with its exercise of
! rights under this agreement.
!**************************************

!** sample compile and execute using Intel compiler:
!** Note: The subroutine lsqr is linked in this example, but is not included with this package.
!**       lsqr.f may be obtained at http://www.stanford.edu/group/SOL/software/lsqr.html.

!**     Include in ifort.cfg:
!**   /check:bounds /traceback /include:\kf_book\distribswm\library /link /nodefaultlib:msvcrt.lib
!**     \kf_book\distribSWm\library\libkf_bnd_ifort.lib \library\lapack-3.1.1\lapack_win32.lib \library\lapack-3.1.1\blas_win32.lib

!**     To compile and execute:
!**   set ifortcfg=\kf_book\distribSWm\ifort.cfg
!**   ifort lsqr.f /c
!**   ifort tst_invert.f90 lsqr.obj   (libkf_bnd_ifort.lib (rnoruni, l1normM, l2NormM, sinv, symult, condest) and lapack should be linked by default)
!**   tst_invert 2>lsqr.lis
!**************************************

      implicit none;
      integer(4),parameter :: nd =40, mmax=101, nsamp=100, lwork=10000;          !use nd=40 for hdiag =T
      integer(4) i,j,k,n,m,m1,it,jt,kk,ier,nstrt,ix,nn2,isamp,nmax;
      integer(4) ITNLIM,NOUT,LENIW,LENRW,IW,RW,ITN,ISTOP;
      integer(4) IWORK(8*min(nd,mmax)),llwork,liwork;
      logical(4) :: qrd =.false., svdR =.false., scaleh =.false.;    !normally all false   ###
      logical(4) :: hdiag =.false., scaleSVD =.false., checkEig =.false.;
      character(30) rDesc;
      real(8) t,dt,dtn,stat(3,6);
      real(8) a(nd,nd),al(nd*(nd+1)/2),at(nd*(nd+1)/2),r(nd*(nd+3)/2);
      real(8) b(nd,nd),c(nd,nd),e(nd,nd),esum,eps,d(nd),x(nd),br(nd);
      real(8) h(mmax,nd+1),h1(mmax,nd+1),h2(mmax,nd+1),bb(nd),dx(nd);
      real(8) cnb,conds,xe(nd),y(mmax),y1(mmax),conde,hnorm(mmax);
      real(8) emax,errm,esumg,emaxg,esumr,emaxr,esums,emaxs,rsos;
      real(8) xer(nd),dscal(nd), ystat(4,2*mmax);
      real(8) DAMP,Uy(mmax),V(nd),W(nd),SE(nd),XL(nd),ATOL,BTOL,CONLIM;
      real(8) ANORM,ACOND,RNORM,ARNORM,XNORM;
      real(8) U(mmax,mmax), VT(nd,nd), WORK(lwork);   !LAPACK SVD
      real(8) :: singTol =1.d-20;   !pseudo-inverse singular value tolerance  ###
!      real(8) :: singTol =1.d-11    !pseudo-inverse singular value tolerance  ###
      real(8),external :: rnor,l2normM,l1normM,condest;     !,aprod
      !____________________________________________________

      rdesc ='thhc_sdd';
!      rdesc ='thhc_sdd_scaled'
!      rdesc ='thhc_sdd_scaled_x14s'
!      rdesc ='thhc_sdd_reorder1'
!      rdesc ='thhc_sdd_psinvE11'

      open (10,file='tstMont_'//trim(rdesc)//'.plt',status='unknown');
      open (6,file='tstCov_'//trim(rdesc)//'.plt',status='unknown');
      write (10,1100);
 1100 format('VARIABLES ="n", "Cholesky-RMS-error", ',                  &
     &  '"Cholesky-max-error", "Gauss-Jordan-RMS-error", ',             &
     &  '"Gauss-Jordan-max-error", "QR-RMS-error", "QR-max-error", ',   &
     &  '"SVD-RMS-error", "SVD-max-error", "CGNR-RMS-error", ',         &
     &  '"CGNR-max-error", "LSQR-RMS-error", "LSQR-max-error", ',       &
     &  '"Cholesky precision", "QR-precision", "SVD-precision"')

      m =mmax;

      write (6,1200);
!      '("  n     errm         esum         emax        ",
!     &       "esumg        emaxg       esumr        emaxr       ",
!     &       "cnb          esums        emaxs      cond")')
 1200 format('VARIABLES ="n", "Cholesky-precision", ',                  &
     &  '"Cholesky-Frobenius-error", "Cholesky-L1-error", ',            &
     &  '"Gauss-Jordan-Frobenius-error", "Gauss-Jordan-L1-error", ',    &
     &  '"QR-Frobenius-error", "QR-L1-error", "QR-precision", ',        &
     &  '"SVD-Frobenius-error", "SVD-L1-error", "SVD-precision", ',     &
     &  '"LINPAC-precision"')

      if (hdiag) then;
        nmax =nd;
      else;
        nmax =20;
      endif;

      do n=4,nmax;     !matrix order

        !** use one of these measurement models
        if (hdiag) then;
          call get_haDiag (h,al,mmax,nd,n,scaleh);      !bi-diagonal h
        else;
          dt =1.d0/(mmax-1);
          call get_haPoly (h,al,mmax,nd,n,scaleh,dt);      !polynomial
        endif;

        kk =0;
        do j=1,n;
          do k=1,j;
            kk =kk+1;
            a(k,j) =al(kk);
            a(j,k) =al(kk);
          enddo;
        enddo;

        ix =191919;
        cnb =rnor(ix);
        stat(:,:) =0.d0;
        nn2 =(n*(n+1))/2;

        do isamp =1,nsamp;
          do i=1,n;
            x(i) =rnor(0);
          enddo;
!          if (scaleh) x(1:4) =0.01*x(1:4)    !###
          y(:) =matmul(h(:,:n),x(:n));
          h(:,n+1) =y(:);
          forall (i=1:n) bb(i) =dot_product(y,h(:,i));
!          bb(:) =matmul(y,h(:,:n))     !=trn(h)*y

          !********************* method 1: normal equations using Cholesky factor
          eps =1.d-9;
          r(:nn2) =al(:nn2);
          call sinv (ier,errm,  r,  n,eps,.false.);
          if (ier < 0) then;
!            write (6,'("sinv error",i3)') ier
            esum =1.d20;
            emax =1.d20;
            errm =1.d20;
          else;
            call symult (xe,  r,bb,n);
            dx(:n) =xe(:n) -x(:n);
            stat(1,1) =stat(1,1) +sum(dx(:n)**2);
            stat(2,1) =max(stat(2,1),maxval(abs(dx(:n))));
            stat(3,1) =stat(3,1) +1;

            kk =0;
            do j=1,n;
              do k=1,j;
                kk =kk+1;
                b(k,j) =r(kk);
                b(j,k) =r(kk);
              enddo;
            enddo;
            c(:n,:n) =matmul(a(:n,:n),b(:n,:n));
            forall (i=1:n) c(i,i) =c(i,i)-1.d0;
            esum =l2normM(c,nd,n);     !sqrt(sum(c(:n,:n)**2)/n**2)
            emax =l1normM(c,nd,n);  !maxval(abs(c(:n,:n)))

            if (isamp ==1 .and. checkEig) then;
              !** check eigenvalues of H^T*H
              llwork =lwork;
              liwork =8*min(nd,mmax);
              b(:,:) =a(:,:);
              call DSYEVD( "N", "U", n, b, nd, d, WORK, llwork,         &
     &                   IWORK, liwork, ier );
              if (ier == 0) then;
                write (6,'("H^T*H eigenvalues =",5(/10g13.5))') d(:n);
              else;
                write (6,'("DSYEVD info =",i4)') ier;
                stop 'DSYEVD feilure';
              endif;
            endif;
          endif;

          !********************* method 2: normal equations using Gauss-jackson inverse
          b(:,:) =a(:,:);
          call matInvGJ (ier,  b,br, nd,n,1); !was gaussj (ier,b,  n,nd,br,1,1)
          if (ier /= 0) then;
!            write (6,'("matInvGJ error",i3)') ier
            esumg =0.;
            emaxg =0.;
          else;
            c(:n,:n) =matmul(a(:n,:n),b(:n,:n));
            forall (i=1:n) c(i,i) =c(i,i)-1.d0;
            esumg =l2normM(c,nd,n);     !sqrt(sum(c(:n,:n)**2)/n**2)
            emaxg =l1normM(c,nd,n);  !maxval(abs(c(:n,:n)))

            xe(:n) =matmul(b(:n,:n),bb(:n));
            dx(:n) =xe(:n) -x(:n);
            stat(1,2) =stat(1,2) +sum(dx(:n)**2);
            stat(2,2) =max(stat(2,2),maxval(abs(dx(:n))));
            stat(3,2) =stat(3,2) +1;
          endif;

          !********************* method 3: invert using SRIF (QR)
          r(:) =0.d0;
          rsos =0.d0;    !computes RHS
          h1(:,:) =h(:,:);
          if (qrd) then;
            call QRdcmp (h1,rsos,   mmax,n,m);
            kk =0;
            do i=1,n;
              do j=1,i;
                kk =kk+1;
                r(kk) =h1(j,i);
              enddo;
            enddo;
            r(nn2+1:nn2+n) =h1(:n,n+1);
          else;
            nstrt =1;
            call thhc_m (r,h1,rsos,  n,mmax,m,nstrt,.false.);
          endif;
          y1(:n) =r(nn2+1:nn2+n);
          call rinz_m (xe,ier,  r,n,r(nn2+1:nn2+n));
          dx(:n) =xe(:n) -x(:n);
          xer(:) =xe(:);
          stat(1,3) =stat(1,3) +sum(dx(:n)**2);
          stat(2,3) =max(stat(2,3),maxval(abs(dx(:n))));
          stat(3,3) =stat(3,3) +1;
          conde =condest(r,n);

          !** form R-inverse
          e(:,:) =0.d0;
          j =0;
          do i=1,n;
            d(:) =0.d0;
            d(i) =1.d0;
            call rinz_m (xe,ier,  r,n,d);     !form b =R^(-1)
            b(:,i) =xe(:);
            e(:i,i) =r(j+1:j+i);
            j =j+i;
          enddo;
          !** get bound on condition number
          cnb =0.d0;
          call rincon_m (at,  cnb,  r,n);   !at =rinv

          c(:n,:n) =matmul(e(:n,:n),b(:n,:n));
          forall (i=1:n) c(i,i) =c(i,i)-1.d0;
!          esumr =sqrt(sum(c(:n,:n)**2)/n**2)
          esumr =l2normM(c,nd,n);     !sqrt(sum(c(:n,:n)**2)/n**2)
          esumr =max(esumr,1.d-17);
          emaxr =l1normM(c,nd,n);  !maxval(abs(c(:n,:n)))
          emaxr =max(emaxr,1.d-17);

          !********************* method 4: svd
          if (svdr) then;
            !** svd on R matrix from THHC
            m1 =n;
            h1(:n,:n) =e(:n,:n);
!            y1(:n) =r(nn2+1:nn2+n)   !from above
          else;
            m1 =m;
            h1(:,:) =h(:,:);
            y1(:m) =y(:m);
            if (scaleSVD) then;
              forall (i=1:n) dscal(i) =sqrt(sum(h(:m,i)**2));
              forall (i=1:n) h1(:m,i) =h1(:m,i)/dscal(i);
            else;
              dscal(:) =1.d0;
            endif;
          endif;
          h2(:,:) =h1(:,:);
          !** numerical recipes SVD
!          call svdcmp (d,b,  h2,  m1,n,mmax,nd)      !d=w, b=v, output h2 =u
          !** LAPACK SVD - SDD is divide and conquer
          llwork =lwork;
          call DGESDD( "S", m1, N, h2, mmax, d, U, mmax, VT, nd, WORK,  &
     &                  llwork, IWORK, ier );
          !** LAPACK SVD - SVD is QR iteration
!          call DGESVD( "S", "S", m1, N, h2, mmax, d, U, mmax, VT, nd,
!     &             WORK, llwork, ier )

          if (llwork < 0) write (*,'("work(1) =",g13.5)') work(1);
          if (ier /= 0) then;
            write (*,'("DGESDD ier =",i5)') ier;
            stop;
          endif;
          !** next 2 lines only for LAPACK
          h2(:m1,:n) =u(:m1,:n);          !###
          b(:n,:n) =transpose(vt(:n,:n)); !###

          e(:n,1) =matmul(transpose(h2(1:m1,1:n)),y1(:m1)); != U**T * y
          !** xe = V * S^(-1) * U**T * y
          do i=1,n;
            if (d(i) > d(1)*singTol) then;
              e(i,1) =e(i,1)/d(i);
            else;
              e(i,1) =0.d0;
            endif;
          enddo;
          xe(:n) =matmul(b(:n,:n),e(:n,1)) /dscal(:n);
          dx(:n) =xe(:n) -x(:n);
          stat(1,4) =stat(1,4) +sum(dx(:n)**2);
          stat(2,4) =max(stat(2,4),maxval(abs(dx(:n))));
          stat(3,4) =stat(3,4) +1;

          !** e =U**T * H
          e(:n,:n) =matmul(transpose(h2(1:m1,1:n)),h1(1:m1,1:n));
          !** C = V * S^(-1) * U**T * H -I
          forall (i=1:n) e(i,:) =e(i,:)/d(i);
          c(:n,:n) =matmul(b(:n,:n),e(:n,:n));
          forall (i=1:n) c(i,i) =c(i,i)-1.d0;
          esums =l2normM(c,nd,n);     !sqrt(sum(c(:n,:n)**2)/n**2)
          emaxs =l1normM(c,nd,n);  !maxval(abs(c(:n,:n)))
          conds =maxval(d(:n))/minval(d(:n));
          if (isamp == 1 .and. checkEig)                                &
     &        write (6,'("svd sing values =",5(/10g13.5))') d(:n);
          if (n == nmax .and. (.not.svdr .and. .not.hdiag))             &
     &              call predict (ystat, xe,xer,x,n,m,dt,hdiag);

          !********************* method 5: pcg - CGNR
          xe(:) =0.d0;
          h1(:,:) =h(:,:);
          call cgnr (ier,  xe,  h1,y,n,m,mmax);

          if (ier /= 0) then;
            write (6,'("*** CGNR error",i3)') ier;
!            esum? =0.
!            emax? =0.
          else;
            dx(:n) =xe(:n) -x(:n);
            stat(1,5) =stat(1,5) +sum(dx(:n)**2);
            stat(2,5) =max(stat(2,5),maxval(abs(dx(:n))));
            stat(3,5) =stat(3,5) +1;
          endif;

          !********************* method 6: pcg - LSQR
          xe(:) =0.d0;
          h1(:,:) =h(:,:);
          damp =0.d0;
          leniw =1;
          lenrw =1;
          uy(:m) =y(:m);
          atol =1.d-6;
          btol =1.d-6;
          conlim =1./(10.*1.d-8);
          itnlim =4*n;
          nout =11;
          call LSQR( M,N,APROD,DAMP,                                    &
     &               LENIW,LENRW,IW,RW,                                 &
     &               uy,V,W,xe,SE,XL,                                   &
     &               ATOL,BTOL,CONLIM,ITNLIM,NOUT,                      &
     &               ISTOP,ANORM,ACOND,RNORM,ARNORM,XNORM,ITN );

!          if (ISTOP > 1 .and. ISTOP /= 4)
!     &       write (6,'("LSQR return code",i3)') ISTOP
          if (ISTOP == 3 .or. ISTOP == 6 .or. ISTOP == 7) then;
            write (6,'("*** LSQR error",i3)') ISTOP;
!            esum? =0.
!            emax? =0.
          else;
            dx(:n) =xe(:n) -x(:n);
            stat(1,6) =stat(1,6) +sum(dx(:n)**2);
            stat(2,6) =max(stat(2,6),maxval(abs(dx(:n))));
            stat(3,6) =stat(3,6) +1;
          endif;
          !*********************

          if (isamp == 1) then;
            write (6,1000) n,log10(errm),log10(esum),log10(emax),       &
     &        log10(esumg),log10(emaxg),                                &
     &        log10(esumr),log10(emaxr),log10(cnb*1.1d-16),             &
     &        log10(esums),log10(emaxs),log10(conds*1.1d-16),           &
     &        log10(conde*1.1d-16);
1000        format (i3,16g13.5)
          endif;

        enddo;    !end isamp loop

        do i=1,6;
          if (stat(1,i) <= 0.d0) then;
            stat(1,i) =1.d40;
          else;
            stat(1,i) =stat(1,i)/stat(3,i);
          endif;
          if (stat(2,i) <= 0.d0) stat(2,i) =1.d40;
        enddo;

        write (10,1000) n,(log10(sqrt(stat(1,i))),                      &
     &        log10(stat(2,i)), i=1,6), log10(errm),                    &
     &        log10(cnb*1.1d-16), log10(conds*1.1d-16);

        if (n == nmax .and. (.not.svdr .and. .not.hdiag)) then;
          open (12,file="predictErr.plt",status='unknown');
          write (12,3000);
 3000     format ('VARIABLES ="Time", "Y-Ymodel (pseudo-inverse)", ',   &
     &            '"Y-Ymodel (QR)", "Y-meas"')
          do i=1,2*mmax;
            write (12,'(4g13.5)') dt*i, sqrt(ystat(2,i)/ystat(1,i)),    &
     &       sqrt(ystat(3,i)/ystat(1,i)), sqrt(ystat(4,i)/ystat(1,i));
          enddo;
        endif;
      enddo;   !___ end n (order) loop

      stop;
      contains;
!**************************************************************
      subroutine APROD( MODE,M,N,X,Y,LENIW,LENRW,IW,RW );
!                IF MODE = 1, COMPUTE  Y = Y + A*X.
!                IF MODE = 2, COMPUTE  X = X + A(TRANSPOSE)*Y.
      implicit none;
      integer(4),intent(in) :: LENIW,LENRW,mode,m,n;
      real(8),intent(inout) :: x(n),y(m);
      integer(4),intent(inout) :: IW(leniw);
      real(8),intent(inout) :: rw(lenrw);
      !_____________________________________

      if (mode == 1) then;
        y(:m) =y(:m) +matmul(h1(:m,:n),x(:n));
      else if (mode == 2) then;
        x(:n) =x(:n) +matmul(y(:m),h1(:m,:n));
      else;
        stop 'invalid mode';
      endif;

      return;
      end subroutine APROD;
      end program;
!*******************************************************************
      subroutine get_haPoly (h,al,mmax,nd,n,scaleh,dt);
      implicit none;

      integer(4),intent(in) :: mmax,nd,n;
      logical(4),intent(in) :: scaleh;
      real(8),intent(in) :: dt;
      real(8),intent(out) :: h(mmax,nd+1),al(nd*(nd+1)/2);

      integer(4) it,jt,j,k,kk;
      real(8) dtn,t,hnorm(mmax);
      !_________________________________________

      al(:) =0.d0;
      h(:,:) =0.d0;
!      dt =1.d0/(mmax-1)

      do it=0,mmax-1;
        t =it*dt;
        dtn =1.d0;
        kk =0;
        jt =it+1;
!        jt =mmax+1-(it+1)    !it+1      !test re-order ###
        do j=1,n;
          h(jt,j) =dtn;
          if (scaleh .and. j <= 4) h(jt,j) =h(jt,j)*100.d0;   !test effect of scaling: makes QR worse and SVD slightly worse
          dtn =dtn*t;
          do k=1,j;
            kk =kk+1;
            al(kk) =al(kk) +h(jt,j)*h(jt,k);
          enddo;
        enddo;
        hnorm(jt) =sqrt(sum(h(jt,1:n)**2));
      enddo;    !end time loop

!      write (*,'("n =",i3,", hnorm =",10(/10g13.5))') n,hnorm(1:mmax)

      return;
      end subroutine get_haPoly;
!*******************************************************************
      subroutine get_haDiag (h,al,mmax,nd,n,scaleh);
      implicit none;

      integer(4),intent(in) :: mmax,nd,n;
      logical(4),intent(in) :: scaleh;
      real(8),intent(out) :: h(mmax,nd+1),al(nd*(nd+1)/2);

      integer(4) it,jt,j,k,kk;
      real(8) dtn,t,hnorm(mmax);
      real(8) :: dt; ! =0.01d0
      !_________________________________________

      al(:) =0.d0;
      h(:,:) =0.d0;

      dt =1.d0/mmax;
      dtn =1.d0/n;
!      nn2 =(n*(n+1))/2

      do it=1,mmax;
        t =it*dt;
        j =min(max(floor(t*(n-1)),1),n-2);
        jt =it;
!        jt =mmax+1-it    !test re-order ###
        h(jt,j) =j*dtn -t;
        h(jt,j+1) =(j+1)*dtn -t;
        h(jt,n) =1.d0;

        kk =0;
        do j=1,n;
          do k=1,j;
            kk =kk+1;
            al(kk) =al(kk) +h(jt,j)*h(jt,k);
          enddo;
        enddo;
        hnorm(jt) =sqrt(sum(h(jt,1:n)**2));
      enddo;    !end time loop

!      write (*,'("n =",i3,", hnorm =",10(/10g13.5))') n,hnorm(1:mmax)

      return;
      end subroutine get_haDiag;
!***********************************************
      subroutine predict (ystat,  xes,xer,x,n,m,dt,hdiag);
!***    predicts measurements for double original interval
      implicit none;
      integer(4),intent(in) :: n,m;
      logical(4),intent(in) :: hdiag;
      real(8),intent(in) :: xes(n),xer(n),x(n),dt;
      real(8),intent(inout) :: ystat(4,2*m);

      integer(4) i,j,mmax;
      real(8) h(2*m,n+1),al(n*(n+1)/2),ysvd(2*m),yqr(2*m),yt(2*m);
      logical(4) :: scaleh =.false.;
      !_________________________________________________

      mmax =2*m;
      if (hdiag) then;
!        call get_haDiag (h,al,mmax,nd,n,scaleh)       !bi-diagonal h
        stop 'invalid use of haDiag';
      else;
        call get_haPoly (h,al,mmax,n,n,scaleh,dt);
      endif;

      ysvd(:) =matmul(h(:mmax,:n),xes(:n));
      yqr(:) =matmul(h(:mmax,:n),xer(:n));
      yt(:) =matmul(h(:mmax,:n),x(:n));

!      open (12,file="predictErr.plt",status='unknown')
!      write (12,1000)
! 1000 format ('VARIABLES ="Time", "Yrres-SVD","Yrres-QR", "Y-meas"')
      do i=1,mmax;
!        write (12,'(4g13.5)') dt*i,ysvd(i)-yt(i),yqr(i)-yt(i), yt(i)
        ystat(1,i) =ystat(1,i) +1.d0;
        ystat(2,i) =ystat(2,i) +(ysvd(i)-yt(i))**2;
        ystat(3,i) =ystat(3,i) +(yqr(i)-yt(i))**2;
        ystat(4,i) =ystat(4,i) +yt(i)**2;
      enddo;

      return;
      end subroutine predict;
