!** time_invert.f times Cholesky and Gaussian elimination
!** inversion of normal equation matrix for polynomial model

!**************************************
! The author grants the user a non-exclusive, worldwide, royalty-free copyright license to
! 1. reproduce and modify the software for your own purposes, and
! 2. to distribute the software provided that you give credit to the author,
!    do not make claims against the author or Wiley-Interscience,
!    and mark any modifications as your own.  If the software is incorporated in
!    a commercial product, you  agree to defend and indemnify this author and
!    Wiley-Interscience against any losses, damages and costs arising from claims,
!    lawsuits and other legal actions brought by a third party.

! The software is provided on an as is basis, without warranties or conditions
! of any kind, either express or implied (including any warranties or conditions
! of title, non-infringement, merchantability or fitness for a particular purpose).
! The user is solely responsible for determining the appropriateness of using and
! distributing the program and assumes all risks associated with its exercise of
! rights under this agreement.
!**************************************

!** sample compile and execute using Intel compiler: (normally compile without bound checking)

!**     Include in ifort.cfg:
!**   /check:bounds /traceback /include:\kf_book\distribswm\library /link /nodefaultlib:msvcrt.lib
!**     \kf_book\distribSWm\library\libkf_bnd_ifort.lib \library\lapack-3.1.1\lapack_win32.lib \library\lapack-3.1.1\blas_win32.lib

!**     To compile and execute:
!**   set ifortcfg=\kf_book\distribSWm\ifort.cfg
!**   ifort time_invert.f90 ..\..\library\libkf_ifort.lib /check:nobounds   (libkf.lib includes sinv, thhc_m, symult, optlib and svdcmp)
!**   time_invert
!**************************************

      implicit none;
      integer(4),parameter :: nd =20, mmax=101, nsamp=100000, lwork=5000;
      integer(4) i,j,k,n,m,m1,it,kk,ier,nstrt,ix,nn2,isamp;
      integer(4) IWORK(8*min(nd,mmax)),llwork;
      logical(4) :: qrd =.false., llsqr=.true., svdr =.false.;
      real(8) t,dt,dtn,stat(2,4);
      real(8) a(nd,nd),al(nd*(nd+1)/2),at(nd*(nd+1)/2),r(nd*(nd+3)/2);
      real(8) b(nd,nd),c(nd,nd),e(nd,nd),esum,eps,d(nd),x(nd),br(nd);
      real(8) h(mmax,nd+1),h1(mmax,nd+1),h2(mmax,nd+1),bb(nd),dx(nd);
      real(8) cnb,conds,xe(nd),y(mmax),y1(mmax),t0, t1,t2,t3,t4;
      real(8) emax,errm,esumg,emaxg,esumr,emaxr,esums,emaxs,rsos;
      real(8) U(mmax,mmax), VT(nd,nd), WORK(lwork);   !LAPACK SVD
      !____________________________________________________

      dt =0.01d0;

      do n=13,13;     !matrix order
        al(:) =0.d0;
        h(:,:) =0.d0;
        do it=0,mmax-1;
          t =it*dt;
          dtn =1.d0;
          kk =0;
          do j=1,n;
            h(it+1,j) =dtn;
            dtn =dtn*t;
            do k=1,j;
              kk =kk+1;
              al(kk) =al(kk) +h(it+1,j)*h(it+1,k);
            enddo;
          enddo;
        enddo;    !end time loop

        kk =0;
        do j=1,n;
          do k=1,j;
            kk =kk+1;
            a(k,j) =al(kk);
            a(j,k) =al(kk);
          enddo;
        enddo;

        ix =191919;
        stat(:,:) =0.d0;
        nn2 =(n*(n+1))/2;

        x(:) =0.d0;
        y(:) =matmul(h(:,:n),x(:n));
        h(:,n+1) =y(:);
        forall (i=1:n) bb(i) =dot_product(y,h(:,i));

        write (6,'("start nsamp cholesky")');
        t0 =0.d0;
        t0 =secnds(t0);
        eps =1.d-3;

        do i=1,nsamp;
          !******* method 1: normal equations using Cholesky factor
          al(:) =0.d0;
          do it=0,mmax-1;
            kk =0;
            do j=1,n;
              do k=1,j;
                kk =kk+1;
                al(kk) =al(kk) +h(it+1,j)*h(it+1,k);
             enddo;
            enddo;
          enddo;
          r(:nn2) =al(:nn2);
          call sinv1 (ier,errm,  r,  n,eps);
          if (ier < 0) stop 'sinv error';
        enddo;
        t1 =secnds(t0);
        write (6,'("end Cholesky",f10.2)') t1;

        do i=1,nsamp;
          !******* method 2: normal equations using Gauss-jackson inverse
          al(:) =0.d0;
          do it=0,mmax-1;
            kk =0;
            do j=1,n;
              do k=1,j;
                kk =kk+1;
                al(kk) =al(kk) +h(it+1,j)*h(it+1,k);
             enddo;
            enddo;
          enddo;
          kk =0;
          do j=1,n;
            do k=1,j;
              kk =kk+1;
              b(k,j) =al(kk);
              b(j,k) =al(kk);
            enddo;
          enddo;
          call matInvGJ (ier,  b,br,  nd,n,1);  !was gaussj (ier,b,  n,nd,br,1,1)
          if (ier /= 0) stop 'gaussj error';
        enddo;
        t2 =secnds(t0);
        write (6,'("end Gauss-jackson",f10.2)') t2-t1;

        do i=1,nsamp;
          !******* method 3: invert using SRIF (QR)
          r(:) =0.d0;
          rsos =0.d0;    !computes RHS
          m =mmax;
          nstrt =1;
          h1(:,:) =h(:,:);
!          call thhc (r,n,h1,mmax,m,rsos,nstrt)
          call thhc_m (r,h1,rsos,  n,mmax,m,nstrt,.false.); !was thhc (r,n,h1,mmax,m,rsos,nstrt)
!          call QRdcmp(h1,rsos,   mmax,n,m)
!          call rinz (r,n,r(nn2+1:),xe,ier)
          call rinz_m (xe,ier,  r,n,y1(:n));
        enddo;
        t3 =secnds(t0);
        write (6,'("end QR svd",f10.2)') t3-t2;

        do i=1,nsamp;
          !******* method 4: svd
          if (svdr) then;
            m1 =n;
            h1(:n,:n) =e(:n,:n);
!            y1(:n) =r(nn2+1:nn2+n)   !from above
          else;
            m1 =m;
            h1(:,:) =h(:,:);
!            y1(:m) =y(:m)
          endif;
          h2(:,:) =h1(:,:);
!          call svdcmp (d,b,  h2,  m1,n,mmax,nd)      !d=w, b=v, output h2 =u
          !** LAPACK SVD
          llwork =lwork;
          call DGESVD( "N", "S", m1, N, h2, mmax, d, U, mmax, VT, nd,   &
     &             WORK, llwork, ier );
!          call DGESDD( "O", m1, N, h2, mmax, d, U, mmax, VT, nd, WORK,
!     &                  llwork, IWORK, ier )
          if (llwork < 0) write (*,'("work(1) =",g13.5)') work(1);
          if (ier /= 0) then;
            write (*,'("DGESDD ier =",i5)') ier;
            stop;
          endif;
          h2(:m1,:n) =u(:m1,:n);
          b(:n,:n) =transpose(vt(:n,:n));
        enddo;
        t4 =secnds(t0);
        write (6,'("end svd: t1,t2,t3,t4 =",4f10.2)')                   &
     &         t1, t2-t1, t3-t2, t4-t3;

      enddo;   !end n=13 loop

!** 4.00  6.05  7.45  28.28  using QR and no bound checking
!** 4.12  6.16  4.94  28.17  using thhc and no bound checking
!** 4.17  6.48  5.09  0.80   using thhc, no bound checking, and SVDR
!** 4.06  6.08  5.02  82.31  using thhc, no bound checking, and LAPACK SVD divide and conquer SVD (full U) (A)
!** 4.12  6.11  5.08  36.49   using thhc, no bound checking, and LAPACK SVD divide and conquer SVD (n-col U) (S)
!** 4.12  6.11  5.08  36.49   using thhc, no bound checking, and LAPACK SVD divide and conquer SVD (n-col U) (O)

      stop;
      end;
! **********************************************************************
!
!       **************
!       *            *
!       *    sinv    *
!       *            *
!       **************
!
      subroutine sinv1 (ier,errm,  a,  n,eps);
!
!        symetric matrix inverse subroutine.  calculates inverse by
!        factoring matrix a using cholesky decomposition. thus, input
!        matrix a must be positve definite for good results.  note that
!        for proper use ier should always be checked.
!
!        arguments
!        ---------
!        a      =  input matrix a.  a is a column vector which contains
!                  the upper triangular elements of the symetric matrix
!                  a by columns.  nominal size of a is (n,n).  number
!                  of elements stored in a is (n*(n+1))/2.
!        n      =  input order of matrix a.
!        eps    =  input tolerance for calculations.  a value of 1.d-5
!                  is a good example.
!        ier    =  output error flag.
!                  = 0              no error.
!                  = k with k > 0.  loss of precision ocurred.  the
!                                   radicand formed during factorization
!                                   of the k+1 row was still positive
!                                   but no longer greater than
!                                   abs(eps*a(k+1,k+1).
!                  = k with k < 0.  output result is garbage. the
!                                   radicand formed during factorization
!                                   of the k+1 row went negative
!                                   indicating that matirx a is not
!                                   positive definite.

      implicit real(8) (a-h,o-z);

      integer(4),intent(in) :: n;
      real(8),intent(in) :: eps;
      real(8),intent(inout) :: a(n*(n+1)/2);
      real(8),intent(out) :: errm;   !minimum remaining precision
      integer(4),intent(out) :: ier;

      integer(4) i,j,k;
      real(8) wmax;
      !_______________________________________________

      iflg=0;
!        test on wrong input parameter n
      k=1;
      if (n.lt.1) go to 12;

!      if (ier.eq.-2) iflg=1     !iflg=1 denotes pseudo-inverse (disable)
      ier=0;
      wmax =0.d0;

!        initialize diagonal-loop
      kpiv=0;
      do k=1,n;
        kpiv=kpiv+k;
        ind=kpiv;
        lend=k-1;
        tol=dabs(eps*a(kpiv));      !calculate tolerance

!          start factorization-loop over k-th row
        do i=k,n;
          dsum=0.;

          do l=1, lend;   !start inner loop
            lanf=kpiv-l;
            lind=ind-l;
            dsum =dsum +a(lanf)*a(lind);
          enddo;          !end of inner loop

!           transform element a(ind)
          dsum =a(ind)-dsum;

          if (i.eq.k) then;     !diagonal
            !** test for negative pivot element and for loss of significance
!            wmax =wmax +(a(kpiv)/dsum)**2      !little difference between using max vs RSS
            wmax =max(wmax,abs(a(kpiv)/dsum));
            if (dsum.le.tol) then;
              if (dsum.le.0.)go to 12;

              work =log10(abs(a(kpiv)/dsum));
!              write (6,100) k,work
  100         format (/' AT ROW',i5,',',f7.1,                           &
     &          ' DIGITS WERE LOST IN MATRIX INVERSION')
              if (iflg.eq.1) dsum=1.d25;
              if (ier.eq.0) ier=k;
            endif;

!            compute pivot element
    9       dpiv =sqrt(dsum);
            a(kpiv) =dpiv;
            dpiv =1.0/dpiv;

          else;                !off-diagonal:  calculate terms in row
            a(ind)=dsum*dpiv;
          endif;

          ind=ind+i;
        enddo;
      enddo;

!      errm =1.1d-16*sqrt(wmax)      !little difference between using max vs RSS
      errm =1.1d-16*wmax;     !1.1d-16 is mantissa LSB (53) of IEEE T_floating on PC

!        invert upper triangular matrix t
!        prepare inversion loop
      ipiv=n*(n+1)/2;
      ind=ipiv;

!        initialize inversion loop
      do i=1,n;
        din=1./a(ipiv);
        a(ipiv)=din;
        min=n;
        kend=i-1;
        lanf=n-kend;

        j=ind;
        do k=1,kend;        !initialize row-loop
          work=0.;
          min=min-1;
          lhor=ipiv;
          lver=j;
          do l=lanf,min;    !start inner loop
            lver=lver+1;
            lhor=lhor+l;
            work=work+a(lver)*a(lhor);
          enddo;            !end of inner loop

          a(j)=-work*din;
          j=j-min;          !end of row loop
        enddo;

        ipiv=ipiv-min;
        ind=ind-1;            !end of inversion loop
      enddo;


!        calculate inverse. inv(a)=inv(t)*trn(inv(t))
      do i=1,n;
        ipiv=ipiv+i;
        j=ipiv;
        do k=i,n;       !initialize row loop
          work=0.;
          lhor=j;

          do l=k,n;     !start inner loop
            lver=lhor+k-i;
            work=work+a(lhor)*a(lver);
            lhor=lhor+l;
          enddo;        !end of inner loop

          a(j)=work;
          j=j+k;
        enddo;
      enddo;

      return;
!
   12 ier=-(k);
      write(6,1000)k;
 1000 format('0MATRIX IS SINGULAR AT ROW ',i5)
      if (iflg.eq.0) return;

      write(6,1010);
 1010 format(' BUT PSEUDO-INVERSE IS BEING COMPUTED BY DELETING ROW')
      dsum=1.d25;
      stop 'SINV ERROR';
!      go to 9

      end subroutine sinv1;
! *******************************************************************
      subroutine thhc (r,n,hc,mmax,m,rsos,nstrt);
!       thhcs is a general purpose routine for transforming an upper
!       triangular plus a rectangular 2-d array into an upper
!       triangular using orthogonal householder transformations.
!**     this version is BGG mod of Biernan's routine

!       calling sequence:
!         inputs:
!           r(n*(n+3)/2) - input triangular matrix to be triangularized
!                        (if rsos >= 0, n+1 column contains information vector)
!           n - number of rows in r
!           hc(mmax,n+1) - measurement matrix with measurements in
!                        last column (if rsos >= 0)
!           m - number of rows in hc  (m.le.mmax)
!           nstrt - index of first nonzero column in hc
!         outputs:
!           r(n*(n+3)/2) - output triangularized matrix
!           rsos - residual sum-of-squares

      implicit none;
      integer(4),intent(in) :: n,mmax,m;
      integer(4),intent(inout) :: nstrt;
      real(8),intent(inout) :: r((n*(n+3))/2),hc(mmax,n+1),rsos;

!       local
      integer(4) np1,ic,j,l,l1;
      real(8) a,sumh,sumr,vic,c,c1;
      real(8) :: eps =1.d-60;
      !_______________________________________________

      if (m > mmax) stop 'THHCS ERROR: M > MMAX';
      np1 =n;
      if (rsos >= 0.d0) np1=n+1;

      nstrt =min(max(nstrt,1),n);
      l =(nstrt*(nstrt-1))/2;

      !**   loop for columns
      do ic =nstrt,n;
        !**  form householder vector and coefficient
        l=l+ic;
        a =r(l);
        sumh =a**2 +sum(hc(1:m,ic)**2);

        if (sumh > eps) then;
!            print *,'THHCS IC,SUM,NMAP',ic,sum,nmap
          sumr =sign(sqrt(sumh),a);
          vic =a+sumr;
          c =1.d0/(sumh+a*sumr);
          !**  vic | hc(*,ic) is the householder vector
          r(l) =-sumr;
          l1=l+ic;

          !** loop thru columns
          do j=ic+1,np1;
            !**  form v*col
            sumh =vic*r(l1) +dot_product(hc(:m,ic),hc(:m,j));
            c1=c*sumh;
            if (c1 /= 0.) then;
              r(l1) =r(l1) -c1*vic;
              hc(:m,j) =hc(:m,j) -c1*hc(:m,ic);
            endif;
            l1=l1+j;
          enddo;
        endif;
      enddo;

!**       calculate a posteriori residual root sum-of-squares
      if (rsos >= 0.d0) then;
        rsos =rsos**2 +sum(hc(:m,np1)**2);
        rsos =sqrt(rsos);
      endif;

      return;
      end subroutine thhc;
! *******************************************************************
      subroutine QRdcmp (hc,rsos,   mmax,n,m);
!       QRdcmp is a general purpose routine for transforming
!       a rectangular 2-d array into an upper triangular array
!       using orthogonal householder transformations.

!**     QRdcmp is about an order of magnitude less accurate than THHCS,
!**     probably because it does not modify elements of hc above the diagonal
!**     at each step    !!!!!!

!       calling sequence:
!         inputs:
!           n - number of rows in r
!           hc(mmax,n+1) - measurement matrix with measurements in
!                        last column (if rsos >= 0)
!           m - number of rows in hc  (m <= mmax)
!         outputs:
!           hc(mmax,n+1)  - output triangularized matrix + RHS
!           rsos - residual sum-of-squares

      implicit none;
      integer(4),intent(in) :: n,mmax,m;
      real(8),intent(inout) :: hc(mmax,n+1),rsos;

!       local
      integer(4) np1,ic,j;
      real(8) a,sumh,sumr,vic,c,c1,scaleh;
      real(8) :: eps =1.d-60;
      !_______________________________________________

      if (m > mmax) stop 'QRdcmp ERROR: M > MMAX';
      np1 =n;
      if (rsos >= 0.d0) np1 =n+1;

!**      loop for columns
      do ic =1,n-1;
        !** form householder vector and coefficient
        scaleh =maxval(abs(hc(ic:m,ic)));

        if (scaleh > eps) then;
          hc(ic:m,ic) =hc(ic:m,ic)/scaleh;   !doesn't seem to improve result much
          a =hc(ic,ic);
          sumh =sum(hc(ic:m,ic)**2);

          if (sumh > eps) then;
!              print *,'QRdcmp IC,SUM,NMAP',ic,sum,nmap
            sumr =sign(sqrt(sumh),a);
            vic =a +sumr;
            c =1.d0/(sumh +a*sumr);
!            c =1.d0/(sumr*(a+sumr))
            hc(ic,ic) =vic;   !vic | hc(ic+1:m,ic) is the householder vector

            !** loop thru columns
            do j=ic+1,np1;
              !**  form v*col
              sumh =dot_product(hc(ic:m,ic),hc(ic:m,j));
              c1 =c*sumh;
              if (c1 /= 0.d0) then;
                hc(ic:m,j) =hc(ic:m,j) -c1*hc(ic:m,ic);
              endif;
            enddo;
            hc(ic,ic) =-sumr*scaleh;    !reset diagonal of column ic
          endif;
        else;
          hc(ic,ic) =0.d0;
        endif;
      enddo;

!**   calculate a posteriori residual root sum-of-squares
      if (rsos >= 0.d0) then;
        rsos =sqrt(sum(hc(:m,np1)**2));
      endif;

      return;
      end subroutine QRdcmp;
! ***************************************************************
      real(8) function secnds(t0);
      real(4),intent(in) :: t0;

      integer(4) idatetime(8);

!       get today'S DATE/TIME
      call date_and_time (values=idatetime);

!        idatetime(1) = year (e.g. 1997)
!        idatetime(2) = month (1-12)
!        idatetime(3) = day (1-31)
!        idatetime(4) = utc time difference
!        idatetime(5) = hour (0-23)
!        idatetime(6) = minutes (0-59)
!        idatetime(7) = seconds (0-59)
!        idatetime(8) = milliseconds (0-999)

      secnds=idatetime(7)+60.*(idatetime(6)+60*idatetime(5))+           &
     &        idatetime(8)*0.001-t0;

      return;
      end;
