!** time_invert2.f times least squares solution methods for a n=200 time-correlated model

!**************************************
! The author grants the user a non-exclusive, worldwide, royalty-free copyright license to
! 1. reproduce and modify the software for your own purposes, and
! 2. to distribute the software provided that you give credit to the author,
!    do not make claims against the author or Wiley-Interscience,
!    and mark any modifications as your own.  If the software is incorporated in
!    a commercial product, you  agree to defend and indemnify this author and
!    Wiley-Interscience against any losses, damages and costs arising from claims,
!    lawsuits and other legal actions brought by a third party.

! The software is provided on an as is basis, without warranties or conditions
! of any kind, either express or implied (including any warranties or conditions
! of title, non-infringement, merchantability or fitness for a particular purpose).
! The user is solely responsible for determining the appropriateness of using and
! distributing the program and assumes all risks associated with its exercise of
! rights under this agreement.
!**************************************

!** sample compile and execute using Intel compiler: (for timing tests, compile library without bound checking!!)

!**      Include in ifort.cfg:
!**   /check:bounds /traceback /include:\kf_book\distribSWm\library /link /nodefaultlib:msvcrt.lib
!**     \kf_book\distribSWm\library\libkf_bnd_ifort.lib \library\lapack-3.1.1\lapack_win32.lib \library\lapack-3.1.1\blas_win32.lib

!**      To compile and execute:
!**   set ifortcfg=\kf_book\distribSWm\ifort.cfg
!**   ifort time_invert2.f90 ..\..\library\libkf_ifort.lib /check:nobounds  (libkf.lib includes sinv, thhc_m, symult, optlib and svdcmp)
!**   time_invert2
!**************************************

      implicit none;
      integer(4),parameter :: nd =200, mmax=2000, nsamp=100,lwork=170000;
      integer(4) i,j,k,n,m,m1,it,kk,ier,nstrt,ix,nn2,isamp;
      integer(4) IWORK(8*min(nd,mmax)),llwork;
      logical(4) :: qrd =.false., svdr =.false., lapack =.true.; !###
      real(8) t,dt,dtn;
      real(8) a(nd,nd),al(nd*(nd+1)/2),at(nd*(nd+1)/2),r(nd*(nd+3)/2);
      real(8) b(nd,nd),c(nd,nd),e(nd,nd),esum,eps,d(nd),x(nd),br(nd);
      real(8) h(mmax,nd+1),h1(mmax,nd+1),h2(mmax,nd+1),bb(nd),dx(nd);
      real(8) cnb,conds,xe(nd),y(mmax),y1(mmax),t0, t1,t2,t3,t4;
      real(8) emax,errm,esumg,emaxg,esumr,emaxr,esums,emaxs,rsos;
      real(8) U(mmax,mmax), VT(nd,nd), WORK(lwork);   !LAPACK SVD
      !____________________________________________________

      dt =1.d0/mmax;
      dtn =1.d0/nd;
      n=nd;    !matrix order
      nn2 =(n*(n+1))/2;
      h(:,:) =0.d0;

!      do it=1,mmax
!        t =it*dt
!        j =min(max(floor(t*nd),1),n-1)
!        h(it,j) =j*dtn -t
!        h(it,j+1) =(j+1)*dtn -t
!      enddo     !end time loop
      call get_haDiag (h,al,mmax,nd,n,.false.);

      x(:) =1.d0;
      y(:) =matmul(h(:,:n),x(:n));
      h(:,n+1) =y(:);
      forall (i=1:n) bb(i) =dot_product(y,h(:,i));

      write (6,'("start nsamp cholesky")');
      t0 =0.d0;
      t0 =secnds(t0);
      eps =1.d-3;

        do i=1,nsamp;
          !******* method 1: normal equations using Cholesky factor
          al(:) =0.d0;
          d(:) =0.d0;
          do it=1,mmax;
            kk =0;
            d(:n) =d(:n) +h(it,:n)*y(it);
            do j=1,n;
              do k=1,j;
                kk =kk+1;
                al(kk) =al(kk) +h(it,j)*h(it,k);
             enddo;
            enddo;
          enddo;
          r(:nn2) =al(:nn2);
          call sinv (ier,errm,  r,  n,eps,.false.);   !was sinv1
          if (ier < 0) stop 'sinv error';
        enddo;
        t1 =secnds(t0);
        call symult (xe,  r,d,n);
        write (*,'("xe =",2(/10g13.5))') xe(:20);
        write (6,'("end Cholesky",f10.2)') t1;

        do i=1,nsamp;
          !******* method 2: normal equations using Gauss-jackson inverse
          al(:) =0.d0;
          do it=1,mmax;
            kk =0;
            do j=1,n;
              do k=1,j;
                kk =kk+1;
                al(kk) =al(kk) +h(it,j)*h(it,k);
             enddo;
            enddo;
          enddo;
          kk =0;
          do j=1,n;
            do k=1,j;
              kk =kk+1;
              b(k,j) =al(kk);
              b(j,k) =al(kk);
            enddo;
          enddo;
          call matInvGJ (ier,  b,br,  nd,n,1);  !was gaussj (ier,b,  n,nd,br,1,1)
          if (ier /= 0) stop 'matInvGJ error';
        enddo;
        t2 =secnds(t0);
        xe(:n) =matmul(b(:n,:n),bb(:n));
        write (*,'("xe =",2(/10g13.5))') xe(:20);
        write (6,'("end Gauss-jackson",f10.2)') t2-t1;

        do i=1,nsamp;
          !******* method 3: invert using SRIF (QR)
          r(:) =0.d0;
          rsos =0.d0;    !computes RHS
          m =mmax;
          nstrt =1;
          h1(:,:) =h(:,:);
          call thhc_m (r,h1,rsos,  n,mmax,m,nstrt,.false.); !was thhc (r,n,h1,mmax,m,rsos,nstrt)
!          call QRdcmp(h1,rsos,   mmax,n,m)
          y1(:n) =r(nn2+1:nn2+n);
          call rinz_m (xe,ier,  r,n,y1(:n));
        enddo;
        t3 =secnds(t0);
        !** copy R to E
        e(:,:) =0.d0;
        j =0;
        do i=1,n;
          e(:i,i) =r(j+1:j+i);
          j =j+i;
        enddo;
        write (*,'("xe =",2(/10g13.5))') xe(:20);
        write (6,'("end QR svd",f10.2)') t3-t2;

        do i=1,nsamp;
          !******* method 4: svd
          if (svdr) then;
            m1 =n;
            h1(:n,:n) =e(:n,:n);
!            y1(:n) =r(nn2+1:nn2+n)   !from above
          else;
            m1 =m;
            h1(:,:) =h(:,:);
            y1(:m) =y(:m);
          endif;
          h2(:,:) =h1(:,:);

          if (.not.lapack) then;
            call svdcmp (d,b,  h2,  m1,n,mmax,nd);     !d=w, b=v, output h2 =u
          else;
            !** LAPACK SVD: DGESVD is normal, DGESDD is divide and conquer
            llwork =lwork;
            call DGESVD( "S", "S", m1, N, h2, mmax, d, U, mmax, VT, nd, &
     &             WORK, llwork, ier );
!           call DGESDD("S", m1, N, h2, mmax, d, U, mmax, VT, nd, WORK,  !divide and conquer SVD
!     &                  llwork, IWORK, ier )
            if (i == nsamp) then;
              h2(:m1,:n) =u(:m1,:n); !do not use for "O" or "N"
              b(:n,:n) =transpose(vt(:n,:n));  !do not use for svdcmp
            endif;
            if (i == 1) write (*,'("work(1) =",g13.5)') work(1);
            if (ier /= 0) then;
              write (*,'("DGESDD ier =",i5)') ier;
              stop;
            endif;
          endif;

        enddo;
        t4 =secnds(t0);

        e(:n,1) =matmul(transpose(h2(1:m1,1:n)),y1(:m1)); != U**T * y
        !** xe = V * S^(-1) * U**T * y
        forall (i=1:n) e(i,1) =e(i,1)/d(i);
        xe(:n) =matmul(b(:n,:n),e(:n,1));
        write (*,'("xe =",2(/10g13.5))') xe(:20);
        write (6,'("end svd: t1,t2,t3,t4 =",4f10.2)')                   &
     &         t1, t2-t1, t3-t2, t4-t3;

!** all runs without bound checking: nd =200, mmax=2000, nsamp=100
!** 37.52  47.78  12.74  29.33  using thhc, LAPACK DGESVD (N,S)
!** 37.58  47.83  12.62  77.17  using thhc, LAPACK DGESVD (S,S) (n-col U)
!** 37.75  47.81  11.72  62.55  using QRdcmp, LAPACK DGESDD (S) (n-col U)  can also use "O" to overwrite
!** 37.73  48.17  11.61  504.34 using QRdcmp, LAPACK DGESDD (A) (full U)
!** 37.67  47.78  12.50  228.06 using thhc, svdcmp
!** 37.61  47.61  12.36  23.61  using thhc, svdcmp on thhc R matrix

!** with sinv, matInvGJ, thhc_m bound checking
!** 46.53  54.84  18.86  29.12 using thhc, LAPACK DGESVD (N,S)

      stop;
      end;
! *******************************************************************
      subroutine QRdcmp (hc,rsos,   mmax,n,m);
!       QRdcmp is a general purpose routine for transforming
!       a rectangular 2-d array into an upper triangular array
!       using orthogonal householder transformations.

!**     QRdcmp is about an order of magnitude less accurate than THHCS,
!**     probably because it does not modify elements of hc above the diagonal
!**     at each step    !!!!!!

!       calling sequence:
!         inputs:
!           n - number of rows in r
!           hc(mmax,n+1) - measurement matrix with measurements in
!                        last column (if rsos >= 0)
!           m - number of rows in hc  (m <= mmax)
!         outputs:
!           hc(mmax,n+1)  - output triangularized matrix + RHS
!           rsos - residual sum-of-squares

      implicit none;
      integer(4),intent(in) :: n,mmax,m;
      real(8),intent(inout) :: hc(mmax,n+1),rsos;

!       local
      integer(4) np1,ic,j;
      real(8) a,sumh,sumr,vic,c,c1,scaleh;
      real(8) :: eps =1.d-60;
      !_______________________________________________

      if (m > mmax) stop 'QRdcmp ERROR: M > MMAX';
      np1 =n;
      if (rsos >= 0.d0) np1 =n+1;

!**      loop for columns
      do ic =1,n-1;
        !** form householder vector and coefficient
        scaleh =maxval(abs(hc(ic:m,ic)));

        if (scaleh > eps) then;
          hc(ic:m,ic) =hc(ic:m,ic)/scaleh;   !doesn't seem to improve result much
          a =hc(ic,ic);
          sumh =sum(hc(ic:m,ic)**2);

          if (sumh > eps) then;
!              print *,'QRdcmp IC,SUM,NMAP',ic,sum,nmap
            sumr =sign(sqrt(sumh),a);
            vic =a +sumr;
            c =1.d0/(sumh +a*sumr);
!            c =1.d0/(sumr*(a+sumr))
            hc(ic,ic) =vic;   !vic | hc(ic+1:m,ic) is the householder vector

            !** loop thru columns
            do j=ic+1,np1;
              !**  form v*col
              sumh =dot_product(hc(ic:m,ic),hc(ic:m,j));
              c1 =c*sumh;
              if (c1 /= 0.d0) then;
                hc(ic:m,j) =hc(ic:m,j) -c1*hc(ic:m,ic);
              endif;
            enddo;
            hc(ic,ic) =-sumr*scaleh;    !reset diagonal of column ic
          endif;
        else;
          hc(ic,ic) =0.d0;
        endif;
      enddo;

!**   calculate a posteriori residual root sum-of-squares
      if (rsos >= 0.d0) then;
        rsos =sqrt(sum(hc(:m,np1)**2));
      endif;

      return;
      end subroutine QRdcmp;
! ***************************************************************
      real(8) function secnds(t0);
      real(4),intent(in) :: t0;

      integer(4) idatetime(8);
      !_________________________________________

!       get today'S DATE/TIME
      call date_and_time (values=idatetime);

!        idatetime(1) = year (e.g. 1997)
!        idatetime(2) = month (1-12)
!        idatetime(3) = day (1-31)
!        idatetime(4) = utc time difference
!        idatetime(5) = hour (0-23)
!        idatetime(6) = minutes (0-59)
!        idatetime(7) = seconds (0-59)
!        idatetime(8) = milliseconds (0-999)

      secnds=idatetime(7)+60.*(idatetime(6)+60*idatetime(5))+           &
     &        idatetime(8)*0.001-t0;

      return;
      end function secnds;
!*******************************************************************
      subroutine get_haPoly (h,al,mmax,nd,n,scaleh);
      implicit none;

      integer(4),intent(in) :: mmax,nd,n;
      logical(4),intent(in) :: scaleh;
      real(8),intent(out) :: h(mmax,nd+1),al(nd*(nd+1)/2);

      integer(4) it,jt,j,k,kk;
      real(8) dtn,t,hnorm(mmax);
      real(8) :: dt; ! =0.01d0
      !_________________________________________

      al(:) =0.d0;
      h(:,:) =0.d0;
      dt =1.d0/(mmax-1);

      do it=0,mmax-1;
        t =it*dt;
        dtn =1.d0;
        kk =0;
        jt =it+1;
!        jt =mmax+1-(it+1)    !it+1      !test re-order ###
        do j=1,n;
          h(jt,j) =dtn;
          if (scaleh .and. j <= 4) h(jt,j) =h(jt,j)*100.d0;   !test effect of scaling: makes QR worse and SVD slightly worse
          dtn =dtn*t;
          do k=1,j;
            kk =kk+1;
            al(kk) =al(kk) +h(jt,j)*h(jt,k);
          enddo;
        enddo;
        hnorm(jt) =sqrt(sum(h(jt,1:n)**2));
      enddo;    !end time loop

!      write (*,'("n =",i3,", hnorm =",10(/10g13.5))') n,hnorm(1:mmax)

      return;
      end subroutine get_haPoly;
!*******************************************************************
      subroutine get_haDiag (h,al,mmax,nd,n,scaleh);
      implicit none;

      integer(4),intent(in) :: mmax,nd,n;
      logical(4),intent(in) :: scaleh;
      real(8),intent(out) :: h(mmax,nd+1),al(nd*(nd+1)/2);

      integer(4) it,jt,j,k,kk;
      real(8) dtn,t,hnorm(mmax);
      real(8) :: dt; ! =0.01d0
      !_________________________________________

      al(:) =0.d0;
      h(:,:) =0.d0;

      dt =1.d0/mmax;
      dtn =1.d0/n;
!      nn2 =(n*(n+1))/2

      do it=1,mmax;
        t =it*dt;
        j =min(max(floor(t*(n-1)),1),n-2);
        jt =it;
!        jt =mmax+1-it    !test re-order ###
        h(jt,j) =j*dtn -t;
        h(jt,j+1) =(j+1)*dtn -t;
        h(jt,n) =1.d0;

        kk =0;
        do j=1,n;
          do k=1,j;
            kk =kk+1;
            al(kk) =al(kk) +h(jt,j)*h(jt,k);
          enddo;
        enddo;
        hnorm(jt) =sqrt(sum(h(jt,1:n)**2));
      enddo;    !end time loop

!      write (*,'("n =",i3,", hnorm =",10(/10g13.5))') n,hnorm(1:mmax)

      return;
      end subroutine get_haDiag;
