!**  Imag_lib.f includes several library routines used for the Imag_misalign.f program

!**************************************
! The author grants the user a non-exclusive, worldwide, royalty-free copyright license to

! 1. reproduce and modify the software for your own purposes, and
! 2. to distribute the software provided that you give credit to the author,
!    do not make claims against the author or Wiley-Interscience,
!    and mark any modifications as your own.  If the software is incorporated in
!    a commercial product, you  agree to defend and indemnify this author and
!    Wiley-Interscience against any losses, damages and costs arising from claims,
!    lawsuits and other legal actions brought by a third party.

! The software is provided on an as is basis, without warranties or conditions
! of any kind, either express or implied (including any warranties or conditions
! of title, non-infringement, merchantability or fitness for a particular purpose).
! The user is solely responsible for determining the appropriateness of using and
! distributing the program and assumes all risks associated with its exercise of
! rights under this agreement.

!**************************************
!** sample compile using Intel compiler:

!**        in ifort.cfg
!**   /check:bounds /traceback /include:\kf_book\distribSWm\library /link /nodefaultlib:msvcrt.lib
!**     \kf_book\distribSWm\library\libkf_bnd_ifort.lib \library\lapack-3.1.1\lapack_win32.lib \library\lapack-3.1.1\blas_win32.lib

!**   set ifortcfg=\kf_book\distribSWm\ifort.cfg
!**   ifort Imag_lib.f /c
!**********************************************************
      subroutine scanImag (sOcs, mirror,x,nx,xdet);
!***  compute input scan vector in orbital coordinates for Imager 2-mirror

      implicit none;
      integer(4),intent(in) :: nx;;    !number of misalignment parameters
      real(8),intent(in) :: x(nx);;    !misalignment vector (equivalenced with
                                       ! structure errParms)
      real(8),intent(in) :: mirror(2);;!mirror gimbal angles (EW and NS)
      real(8),intent(in) :: xdet(3);;  !detector location in instrument coord (e.g. -1,0,0)
      real(8),intent(out) :: sOcs(3);; !output scan unit vector in instrument coord

      real(8) d(3),inc(3),ne(3),nn(3),ren(3),sins(3);
      real(8) me3,mn1,ce,se,cn,sn,dot,arot(3,3),dmag;
      real(8) mem(3),me(3),mnm(3),mn(3);
      real(8),parameter :: pi4 =3.141592653589793d0/4.d0;

      type errParms;
        sequence;
        real(8) rpy(3);    !instrument roll,pitch,yaw wrt reference
        real(8) md1;       !detector array rotation about x axis
        real(8) deld(2);   !detector y,z error offset
        real(8) mem(2);    !EW mirror misalignment rotation wrt gimbal
        real(8) me(3);     !EW gimbal misalignment rotation
        real(8) mnm(2);    !NS mirror misalignment rotation wrt gimbal
        real(8) mn(3);     !NS gimbal misalignment rotation
      end type;
      type (errParms) errParm;
      !_______________________________________

       errParm =transfer(x,errParm);   !move misalignments to errParm structure

       d(1) =xdet(1);
       d(2) =xdet(2)+errParm%deld(1)-xdet(3)*errParm%md1;
       d(3) =xdet(3)+errParm%deld(2)+xdet(2)*errParm%md1;

       dmag =sqrt(sum(d(:)**2));
       inc(:) =-d(:)/dmag;

       !** EW mirror and gimbal
!       me(:) = (/ errParm%me(1), errParm%me(2,
!     &             mirror(1) +pi4 +errParm%me(3)) /)

       mem(:) =(/ errParm%mem(1), 0.d0, errParm%mem(2) /);
       me3 =mirror(1) +pi4;
       ce =cos(me3);
       se =sin(me3);

       ne(1) =-(mem(3)+errParm%me(3))*ce +se;
       ne(2) =(mem(3)+errParm%me(3))*se +ce;
       ne(3) =-errParm%me(2)*se +errParm%me(1)*ce +mem(1);

       dot =dot_product(ne,inc);
       ren(:) =inc(:) -2.d0*dot*ne(:);

       !** NS mirror and gimbal
       mnm(:) =(/ errParm%mnm(1), 0.d0, errParm%mnm(2) /);
       mn1 =mirror(2) +pi4;
       cn =cos(mn1);
       sn =sin(mn1);

       nn(1) =mnm(3) +errParm%mn(3)*cn +errParm%mn(2)*sn;
       nn(2) =-(mnm(1) +errParm%mn(1))*sn -cn;
       nn(3) =-(mnm(1) +errParm%mn(1))*cn +sn;

       dot =dot_product(nn,ren);
       sins(:) =ren(:) -2.d0*dot*nn(:);
!      write (6,'("ne =",3f10.4,", ren =",3f10.4,", nn =",3f10.4,       &
!    &           ", sins =",3f10.4)') ne,ren,nn,sins;

       !*** roll,pitch,yaw rotation
       arot(1,1) = 1.d0-0.5d0*(errParm%rpy(2)**2 +errParm%rpy(3)**2);
       arot(1,2) = -errParm%rpy(3);
       arot(1,3) = errParm%rpy(2);
       arot(2,1) = errParm%rpy(3) + errParm%rpy(1)*errParm%rpy(2);
       arot(2,2) = 1.d0-0.5d0*(errParm%rpy(1)**2 +errParm%rpy(3)**2);
       arot(2,3) = -errParm%rpy(1);
       arot(3,1) = -errParm%rpy(2) + errParm%rpy(1)*errParm%rpy(3);
       arot(3,2) = errParm%rpy(1) + errParm%rpy(2)*errParm%rpy(3);
       arot(3,3) = 1.d0-0.5d0*(errParm%rpy(1)**2 +errParm%rpy(2)**2);

       sOcs(:) =matmul(arot,sins);

       return;
       end subroutine scanImag;
!**********************************************************
      subroutine scanImag2 (sOcs, mirror,x,nx,xdet);
!****   scanImag2 is an alternate version of scanImag
!***   compute input scan vector in orbital coordinates for Imag 2-mirror

      implicit none;
      integer(4),intent(in) :: nx;  !number of misalignment parameters
      real(8),intent(in) :: x(nx),mirror(2),xdet(3);
      real(8),intent(out) :: sOcs(3);

      real(8) d(3),inc(3),ne(3),nn(3),ren(3),sins(3);
      real(8) me3,mn1,ce,se,cn,sn,dot,arot(3,3),dmag;
      real(8) mem(3),me(3),mnm(3),mn(3),xm(3),xm1(3),nor(3);
      real(8),parameter :: pi4 =3.141592653589793d0/4.d0;
      type errParms;
        sequence;
        real(8) rpy(3);  !roll,pitch,yaw
        real(8) md1,deld(2),mem(2),me(3),mnm(2),mn(3);
      end type;
      type (errParms) errParm;
      !_______________________________________

      errParm =transfer(x,errParm);

      d(1) =xdet(1);
      d(2) =xdet(2)+errParm%deld(1)-xdet(3)*errParm%md1;
      d(3) =xdet(3)+errParm%deld(2)+xdet(2)*errParm%md1;

      dmag =sqrt(sum(d(:)**2));
      inc(:) =-d(:)/dmag;

      !** EW mirror and gimbal

      !*** EW mirror misalignment about y-axis
      mem(:) =(/ errParm%mem(1), 0.d0, errParm%mem(2) /);
      me(:) =errParm%me(:);
      me3 =mirror(1) +pi4;

      nor(:) =(/0.d0, 1.d0, 0.d0 /);
      call small_rot (xm, nor, mem);
      call rotate (xm1, xm,-me3, 3);  !rotate about -z axis
      call small_rot (ne, xm1, me);

      dot =dot_product(ne,inc);
      ren(:) =inc(:) -2.d0*dot*ne(:);

      !** NS mirror and gimbal
      mnm(:) =(/ errParm%mnm(1), 0.d0, errParm%mnm(2) /);
      mn(:) =errParm%mn(:);
      mn1 =mirror(2) +pi4;

      nor(:) =(/0.d0, -1.d0, 0.d0 /);
      call small_rot (xm,   nor, mnm);
      call rotate (xm1,  xm,-mn1, 1);!rotate about -x axis
      call small_rot (nn,  xm1, mn);

      dot =dot_product(nn,ren);
      sins(:) =ren(:) -2.d0*dot*nn(:);
!     write (6,'("ne =",3f10.4,", ren =",3f10.4,", nn =",3f10.4,        &
!    &          ", sins =",3f10.4)') ne,ren,nn,sins;

      call rotate (xm,  sins,errParm%rpy(1), 1);
      call rotate (xm1,  xm,errParm%rpy(2), 2);
      call rotate (sOcs,  xm1,errParm%rpy(3), 3);

      return;
      end subroutine scanImag2;
!**********************************************************
      subroutine scanGOES (sOcs, mirror,x,nx,xdet);
!***  compute input scan vector in orbital coordinates for GOES 1-mirror
      implicit none;
      integer(4),intent(in) :: nx;  !number of misalignment parameters
      real(8),intent(in) :: x(nx),mirror(2),xdet(3);
      real(8),intent(out) :: sOcs(3);

      real(8) d(3),inc(3),ne(3),nn(3),ren(3),sins(3);
      real(8) me2,mn1,ce,se,cn,sn,dot,arot(3,3),ce2n(3,3),dmag;
      real(8) mem(3),me(3),mnm(3),mn(3),xm(3),xm1(3),nor(3);
      real(8),parameter :: pi4 =3.141592653589793d0/4.d0;
      type errParms;
        sequence;
        real(8) rpy(3);  !roll,pitch,yaw
        real(8) md1,deld(2),mem(2),me(3),mnm(2),mn(3);
      end type;
      type (errParms) errParm;
      !_______________________________________

      errParm =transfer(x,errParm);

      d(1) =-xdet(1);
      d(2) =xdet(2)+errParm%deld(1)-xdet(3)*errParm%md1;
      d(3) =xdet(3)+errParm%deld(2)+xdet(2)*errParm%md1;

      dmag =sqrt(sum(d(:)**2));
      inc(:) =-d(:)/dmag;

      !** EW mirror and gimbal
!      me(:) = (/ errParm%me(1), errParm%me(2,
!    &              mirror(1) +pi4 +errParm%me(3)) /)

      mem(:) =(/ 0.d0, errParm%mem(1), errParm%mem(2) /);
      me2 =mirror(1) +pi4;
      ce =cos(me2);
      se =sin(me2);

      ne(1) =(mem(2)+errParm%me(2))*se -ce;
      ne(2) =-errParm%me(3)*ce -errParm%me(1)*se -mem(3);
      ne(3) =(mem(2)+errParm%me(2))*ce +se;

      !** NS gimbal
!      mnm(:) = not used
      mn(:) =errParm%mn(:);
      mn1 =mirror(2);
      cn =cos(mn1);
      sn =sin(mn1);

      ce2n(1,:) = (/ 1.d0, -mn(3)*cn +mn(2)*sn, mn(3)*sn +mn(2)*cn /);
      ce2n(2,:) = (/ mn(3), cn-mn(1)*sn,  -sn-mn(1)*cn /);
      ce2n(3,:) = (/ -mn(2), sn+mn(1)*cn, cn-mn(1)*sn /);

      nn(:) =matmul(ce2n,ne);
!      write (6,'("nn mag =",g13.5)') sqrt(sum(nn(:)**2))
      dot =dot_product(nn,inc);
      sins(:) =inc(:) -2.d0*dot*nn(:);
!     write (6,'("ne =",3f10.5,", nn =",3f10.5,", sins =",3f10.5)')     &
!    &            ne,nn,sins;

      !*** roll,pitch,yaw rotation
      arot(1,1) = 1.d0-0.5d0*(errParm%rpy(2)**2 +errParm%rpy(3)**2);
      arot(1,2) = -errParm%rpy(3);
      arot(1,3) = errParm%rpy(2);
      arot(2,1) = errParm%rpy(3) + errParm%rpy(1)*errParm%rpy(2);
      arot(2,2) = 1.d0-0.5d0*(errParm%rpy(1)**2 +errParm%rpy(3)**2);
      arot(2,3) = -errParm%rpy(1);
      arot(3,1) = -errParm%rpy(2) + errParm%rpy(1)*errParm%rpy(3);
      arot(3,2) = errParm%rpy(1) + errParm%rpy(2)*errParm%rpy(3);
      arot(3,3) = 1.d0-0.5d0*(errParm%rpy(1)**2 +errParm%rpy(2)**2);

      sOcs(:) =matmul(arot,sins);

       return;
       end subroutine scanGOES;
!**********************************************************
      subroutine scanGOES2 (sOcs, mirror,x,nx,xdet);
!***  compute input scan vector in orbital coordinates for GOES 1-mirror
      implicit none;
      integer(4),intent(in) :: nx;  !number of misalignment parameters
      real(8),intent(in) :: x(nx),mirror(2),xdet(3);
      real(8),intent(out) :: sOcs(3);

      real(8) d(3),inc(3),ne(3),nn(3),ren(3),sins(3);
      real(8) me2,mn1,ce,se,cn,sn,dot,arot(3,3),ce2n(3,3),dmag;
      real(8) mem(3),me(3),mnm(3),mn(3),xm(3),xm1(3),nor(3);
      real(8),parameter :: pi4 =3.141592653589793d0/4.d0;
      type errParms;
        sequence;
        real(8) rpy(3);  !roll,pitch,yaw
        real(8) md1,deld(2),mem(2),me(3),mnm(2),mn(3);
      end type;
      type (errParms) errParm;
      !_______________________________________

      errParm =transfer(x,errParm);

      d(1) =-xdet(1);
      d(2) =xdet(2)+errParm%deld(1)-xdet(3)*errParm%md1;
      d(3) =xdet(3)+errParm%deld(2)+xdet(2)*errParm%md1;

      dmag =sqrt(sum(d(:)**2));
      inc(:) =-d(:)/dmag;

      !** EW mirror and gimbal
      mem(:) =(/ 0.d0, errParm%mem(1), errParm%mem(2) /);
      me(:) =errParm%me(:);
      me2 =mirror(1) +pi4;

      nor(:) =(/-1.d0, 0.d0, 0.d0 /);
      call small_rot (xm, nor, mem);
      call rotate (xm1, xm,me2, 2);
      call small_rot (ne, xm1, me);

      !** NS gimbal
!      mnm(:) = not used
      mn1 =mirror(2);
      mn(:) =errParm%mn(:);

      call rotate (xm1, ne,mn1, 1);  !rotation is along +x axis
      call small_rot (nn, xm1, mn);

!      write (6,'("nn mag =",g13.5)') sqrt(sum(nn(:)**2))
      dot =dot_product(nn,inc);
      sins(:) =inc(:) -2.d0*dot*nn(:);
!     write (6,'("ne =",3f10.5,", nn =",3f10.5,", sins =",3f10.5)')     &
!    &            ne,nn,sins;

      call rotate (xm,  sins,errParm%rpy(1), 1);
      call rotate (xm1,  xm,errParm%rpy(2), 2);
      call rotate (sOcs,  xm1,errParm%rpy(3), 3);

      return;
      end subroutine scanGOES2;
!*****************************************************************
      subroutine small_rot (xout,  xin,rot);
!***    transforms xin by infinitessimally small rotations, rot,
!***    about the three orthogonal axes.  Order is unimportand for
!***    infinitessimally small rotations.
      implicit none;
      real(8),intent(in) :: xin(3),rot(3);
      real(8),intent(out) :: xout(3);
      real(8) c(3,3);
      !_____________________

      c(1,:) =(/ 1.d0, -rot(3), rot(2) /);
      c(2,:) =(/ rot(3), 1.d0, -rot(1) /);
      c(3,:) =(/ -rot(2), rot(1), 1.d0 /);
      xout(:) =matmul(c,xin);
      return;
      end subroutine small_rot;
!*****************************************************************
      subroutine rotate (xout,  xin,rot,axis);
!***    transforms xin by rotation "rot" about the axis "axis".
      implicit none;
      integer(4),intent(in) :: axis;
      real(8),intent(in) :: xin(3),rot;
      real(8),intent(out) :: xout(3);
      real(8) c(3,3),cr,sr;
      !_____________________

      cr =cos(rot);
      sr =sin(rot);

      if (axis == 1) then;
        c(1,:) =(/ 1.d0, 0.d0, 0.d0 /);
        c(2,:) =(/ 0.d0,   cr, -sr /);
        c(3,:) =(/ 0.d0,   sr, cr /);
      else if (axis == 2) then;
        c(1,:) =(/   cr, 0.d0, sr /);
        c(2,:) =(/ 0.d0, 1.d0, 0.d0 /);
        c(3,:) =(/  -sr, 0.d0, cr /);
      else if (axis == 3) then;
        c(1,:) =(/   cr,  -sr, 0.d0 /);
        c(2,:) =(/   sr,   cr, 0.d0 /);
        c(3,:) =(/ 0.d0, 0.d0, 1.d0 /);
      else;
        stop 'invalid axis in subroutine rot';
      endif;

      xout(:) =matmul(c,xin);
      return;
      end subroutine rotate;
