!*** Program imag_misalign models misalignments in an optical imaging instrument,
!*** and tests the accuracy of reduced-order models for least squares estimation.
!*** Singular value decomposition of the information matrix is used to determine
!*** observable linear combinations.  Fit accuracy is evaluated using Monte Carlo
!*** simulation.  Stepwise regression is also used to fit using reduced-order models.
!*** This is used in Examples 6.2-4 and 6.3-1.

!**************************************
! The author grants the user a non-exclusive, worldwide, royalty-free copyright license to

! 1. reproduce and modify the software for your own purposes, and
! 2. to distribute the software provided that you give credit to the author,
!    do not make claims against the author or Wiley-Interscience,
!    and mark any modifications as your own.  If the software is incorporated in
!    a commercial product, you agree to defend and indemnify this author and
!    Wiley-Interscience against any losses, damages and costs arising from claims,
!    lawsuits and other legal actions brought by a third party.

! The software is provided on an as is basis, without warranties or conditions
! of any kind, either express or implied (including any warranties or conditions
! of title, non-infringement, merchantability or fitness for a particular purpose).
! The user is solely responsible for determining the appropriateness of using and
! distributing the program and assumes all risks associated with its exercise of
! rights under this agreement.
!**************************************

!** sample compile and execute using Intel compiler:

!**     Include in ifort.cfg
!**   /check:bounds /traceback /include:\kf_book\distribSWm\library /link /nodefaultlib:msvcrt.lib
!**     \kf_book\distribSWm\library\libkf_bnd_ifort.lib \library\lapack-3.1.1\lapack_win32.lib \library\lapack-3.1.1\blas_win32.lib

!**     To compile and execute:
!**   set ifortcfg=\kf_book\distribSWm\ifort.cfg
!**   ifort jumpcom.f90 Imag_lib.f90 /c
!***  ifort Imag_misalign.f90 Imag_lib.obj jumpcom.obj   (libkf_bnd_ifort.lib should be linked by default)
!                                                      !may also need svdcmp.obj
!**   Imag_misalign >miss_init3_f2r2.lis
!**************************************

      use jumpcom;
      implicit none;
      integer(4) i,j,k,l,m,n,ne,ier,ix,isamp,npest;
      integer(4),parameter :: nsamp =500;
      integer(4),parameter :: nx=16;
      integer(4),parameter :: nscan=17, nmeas =2, nrd=nmeas*nscan;
      integer(4),parameter :: nscanE=15,  nrdE=nmeas*nscanE;
      integer(4),parameter :: iError =4;  !###
      integer(4) nj,nsel(nx),jsel(nx),nset,jset(nx,50),njset(50);
      logical(4) :: estim(nx) =.true.;
      logical(4) :: GOEStype =.false.;
      logical(4) :: ROM =.false.; !.true.   !###
      logical(4) :: testH =.false.;
      logical(4) :: visible =.true.;
      logical(4) :: mNoise =.false.;!true.  ###
      logical(4) debugP,match;
      logical(4) isel(nx);
      real(8) x(nx),xref(nx),xe(nx),d(nx*(nx+1)/2),eps;      !state vector
      real(8) h(nrd,nx),ah(nrd,nx),ahn(nrd,nx),t2(nx),c(nx);
      real(8) t1(nx,nx),u(nrd,nrd);
      real(8) b(nrd),be(nrd),bn(nrd),b0(nrd);
      real(8) besum(nrd),bemax(nrd), bemaxE(nrdE,2),besumE(nrdE,2);
      real(8) xdet(3),xdet0(3),xdet1(3,nscan);   !detector coords
      real(8) mirror(2),mirrorList(2,nscan);  !E and N mirror angles -pi/4
      real(8) mirrorLstEval(2,nscanE);        !landmark E and N mirror angles for evaluation
      real(8) scan0(2), bem2(2),bem2a(2);
      real(8) sOcs(3),sOcs0(3),x0,w(nx),v(nx,nx);
      real(8) dz(nrd),dzdp(nrd,nx),xp(nx),rrmsj(50),rrms,errm;

      !*** visible
      real(8) :: delDet(2) =(/ 50d-6, 7850d-6 /);   !max y and z detector offset from mean
      real(8) :: meanDet(2) =(/ 3300d-6, 0.d0 /);   !mean y and z detector offset from origin (vis)  (0.46 & 0.64)
!      real(8) :: meanDet(2) =(/ 10000d-6, 0.d0 /)    !mean y and z detector offset from origin (vis)  (0.86 & 1.61)
      !*** IR
!      real(8) :: delDet(2) =(/ 16600d-6, 7850d-6 /) !max y and z detector offset from mean
!      real(8) :: meanDet(2) =(/ 0.d0, 0.d0 /)       !mean y and z detector offset from origin (IR)

      real(8) sigx(nx);
      real(8) :: sigm(2) =10d-6;       !meas noise
      real(8),external :: rnor, uni;
      character(12) :: xnames(nx) =                                     &
     &    (/ "roll", "pitch", "yaw", "detector-rot", "detector-y",      &
     &       "detector-z", "EW mirror-x", "EW mirror-z", "EW shaft-x",  &
     &       "EW shaft-y", "EW shaft-z", "NS mirror-x", "NS mirror-z",  &
     &       "NS shaft-x", "NS shaft-y", "NS shaft-z" /);
      !** error parameters
      type errParms;
        sequence;
        real(8) rpy(3);  !roll,pitch,yaw
        real(8) md1,deld(2),mem(2),me(3),mnm(2),mn(3);
      end type;
      type (errParms) errParm;
      !_______________________________________________________

      xref(:) =0.d0;
      errParm =transfer(xref,errParm);

      xdet(:) = (/1.d0, 550d-6, -5700d-6 /); !detector coordinates   (best fit error is 1 microrad)
!      xdet(:) = (/1.d0, 550d-6, 100d-6 /)    !detector coordinates
      xdet0(:) =(/1.d0, 0.d0, 0.d0 /);       !can use either 0 or xdet for partials
      xdet0(2:3) =meanDet(:);

      if (ierror == 1) then;
        sigx(4:nx) =100.d-6;
        sigx(4:6) =(1.d-6/3.d0)*(/ 400., 70., 30.  /);        !detector
        sigx(7:11) =(1.d-6/3.d0)*(/ 100., 10.,  50., 300., 5. /); !EW mirror and shaft
        sigx(12:16) =(1.d-6/3.d0)*(/ 10., 40.,  5., 300., 300. /); !NS mirror and shaft
      else if (ierror == 2) then;
        !*** worst case
        sigx(1:3) =(1.d-6/3.d0)*300.;                         !roll-pitch-yaw
        sigx(4:6) =(1.d-6/3.d0)*(/ 360., 70., 30.  /);        !detector
        sigx(7:11) =(1.d-6/3.d0)*(/ 100., 10.,  50., 350., 5. /); !EW mirror and shaft
        sigx(12:16) =(1.d-6/3.d0)*(/ 10., 40.,  5., 500., 800. /); !NS mirror and shaft
      else if (ierror == 3) then;
        !*** manufacturer worst case
        sigx(1:3) =(1.d-6/3.d0)*(/ 60., 60., 30. /);          !roll-pitch-yaw
        sigx(4:6) =(1.d-6/3.d0)*(/ 360., 150., 70.  /);        !detector
        sigx(7:11) =(1.d-6/3.d0)*(/ 60., 5., 220., 220.,  50. /); !EW mirror and shaft
        sigx(12:16) =(1.d-6/3.d0)*(/ 10., 20.,  240., 80., 100. /); !NS mirror and shaft
!        sigx(12:16) =(1.d-6/3.d0)*(/ 30., 30.,  5., 400., 800. /)  !NS mirror and shaft   before 6/12/03
      else if (ierror == 4) then;
        !*** revised model for KF book
        sigx(1:3) =(1.d-6/3.d0)*(/ 60., 60., 30. /);          !roll-pitch-yaw
        sigx(4:6) =(1.d-6/3.d0)*(/ 360., 120., 100.  /);        !detector
        sigx(7:11) =(1.d-6/3.d0)*(/ 60., 10., 200., 200.,  70. /); !EW mirror and shaft
        sigx(12:16) =(1.d-6/3.d0)*(/ 10., 20.,  220., 100., 120. /); !NS mirror and shaft
      else;
        stop 'invalid ierror';
      endif;

!** initialize variables mirrorList,mirrorLstEval,estim, and perform SVD analysis
!** uses global variables
!             xref,xdet0,meanDet,DelDet,sigx,nx,nrd,nscan,nscanE,nmeas,
!             xnames,GOEStype,testH,ROM)
      call init_svd (mirrorList,mirrorLstEval,estim);

!*** compute least squares estimate using b =H*x with all x =1

      ix =191919;
      x0 =rnor(ix); !also initializes uni

!***   compute nominal (no misalignment) scan angles and save as b0
!      l=0
!      do i=1,nscan
!        mirror =mirrorList(:,i)
!        x(:) =xref(:)
!        !*** randomize detector locations within +/- 500 microrad EW,  7800 microrad NS
!        xdet1(1,i) =xdet0(1)
!        xdet1(2,i) =xdet0(2)+2.d0*deldet(1)*(uni(0) -0.5)
!        xdet1(3,i) =xdet0(3)+2.d0*deldet(2)*(uni(0) -0.5)
!        if (GOEStype) then
!          call scanGOES2 (sOcs0, mirror, x,nx,xdet1(:,i))
!        else
!          call scanImag2 (sOcs0, mirror, x,nx,xdet1(:,i))
!        endif
!        scan0(:) = (/ asin(sOcs0(1)), -atan(sOcs0(2)/sOcs0(3)) /)  !GOES EW, NS
!        b0(l+1:l+nmeas) =scan0(:)
!        l=l+nmeas
!      enddo

      besum(:) =0.d0;
      bemax(:) =0.d0;
      besume(:,:) =0.d0;
      bemaxe(:,:) =0.d0;
      nsel(:) =0;
      nset =0;
      jset(:,:) =0;
      njset(:) =0;
      rrmsj(:) =0.d0;
      call init_jump (nrd,nx,xnames);

      do isamp=1,nsamp;
        do i=1,nx;
          x(i) = xref(i) +rnor(0)*sigx(i);
        enddo;

        !** generate E,N scan angles using full nonlinear model with detector offset
        l=0;
        do i=1,nscan;
          mirror(:) =mirrorList(:,i);

          !*** randomize detector locations
          xdet1(1,i) =xdet0(1);
          xdet1(2,i) =xdet0(2) +2.d0*deldet(1)*(uni(0) -0.5);
          xdet1(3,i) =xdet0(3) +2.d0*deldet(2)*(uni(0) -0.5);
!          xdet1(2,i) =xdet0(2) +meanDet(1) +2.d0*deldet(1)*(uni(0) -0.5)
!          xdet1(3,i) =xdet0(3) +meanDet(2) +2.d0*deldet(2)*(uni(0) -0.5)

          !**  compute nominal (no misalignment) and misalignment-affected scan angles
          !**  save nominal as b0
          if (GOEStype) then;
            call scanGOES2 (sOcs0, mirror, xref,nx,xdet1(:,i));
            call scanGOES2 (sOcs, mirror, x,nx,xdet1(:,i));
          else;
            call scanImag2 (sOcs0, mirror, xref,nx,xdet1(:,i));
            call scanImag2 (sOcs, mirror, x,nx,xdet1(:,i));
          endif;
          scan0(:) = (/ asin(sOcs0(1)), -atan(sOcs0(2)/sOcs0(3)) /); !GOES EW, NS
          b0(l+1:l+nmeas) =scan0(:);
          b(l+1:l+nmeas) = (/ asin(sOcs(1)), -atan(sOcs(2)/sOcs(3)) /);  !GOES EW, NS
          l =l +nmeas;
        enddo;

        !*** compute scan residual due to misalignments
        b(:) =b(:) -b0(:);

        !*** get partials at actual detector locations and no misalignments
        call hPartial (h,ah,n,                                          &
     &                 mirrorlist,xref,xdet1,xdet0,nx,nrd,nscan,nmeas,  &
     &                 GOEStype,testH,estim);

        ne =n;

        !*** scale residuals and partials by meas noise
        m=0;
        do i=1,nscan;
          bn(m+1:m+nmeas) = b(m+1:m+nmeas)/sigm(:);
          ahn(m+1,:) =ah(m+1,:)/sigm(1);
          ahn(m+nmeas,:) =ah(m+nmeas,:)/sigm(2);
          m =m+nmeas;
        enddo;

        dz(:m) =bn(:m);
        if (mNoise) then;
          do i=1,m;
            dz(i) =dz(i) +rnor(0);
          enddo;
        endif;
        dzdp(:m,:n) =ahn(:m,:n);

        !*** stepwise regression ***
        isel(1:3) =.true.;
!        isel(1:3) =.false. !###
        isel(4:) =.false.;
        call estjmpj (xp,npest,rrms,  isel,  dz,dzdp,n,m,nrd);

        !** record number of times states and state sets were selected and number of times sets
        nj =0;
        jsel(:) =0;
        do i=1,n;
          if (isel(i)) then;
            nsel(i) =nsel(i)+1;
            nj =nj+1;
            jsel(nj) =i;
          endif;
        enddo;
        match =.false.;
        do i=1,nset;
          if (sum(jsel(:nj) -jset(:nj,i)) == 0) then;
            match =.true.;
            njset(i) =njset(i) +1;
            rrmsj(i) =rrmsj(i) +rrms**2;
          endif;
        enddo;
        if (.not.match) then;
          nset =nset+1;
          write (6,'("nset =",i3)') nset;
          jset(:nj,nset) =jsel(:nj);
          njset(nset) =1;
          rrmsj(nset) =rrmsj(nset) +rrms**2;
        endif;

        !*** old linear model
!        b(:) =matmul(h(:,1:nx),x(1:nx))

        if (isamp < 2) then;
          write (6,'(/"b =",3(/10g13.5))') b(:);
          debugP =.true.;
        else;
          debugP =.false.;
        endif;

        t1(1:ne,1:ne) =matmul(transpose(ahn(:,1:ne)),ahn(:,1:ne));
        c(1:ne) =matmul(transpose(ahn(:,1:n)),bn(:));
        l=0;
        do i=1,ne;
          do j=1,i;
            l=l+1;
            d(l) =t1(j,i);
          enddo;
          if (.not.rom) d(l) =d(l) +1.d2;
        enddo;
        eps =1.d-4;
        call sinv (ier,errm,  d,  ne,eps,debugP);

        if (ier >= 0) then;
          call symult (xe,  d,c,ne); !xe =d*c
          if (debugP) then;
            write (6,'(/"least-squares sigma and correlations")');
            l=0;
            do i=1,ne;
              l=l+i;
              t2(i) =sqrt(d(l));
              write (6,'(i3,f13.6,2x,10f8.4,2(/18x,10f8.4))') i,t2(i),  &
     &              (d(l-i+j)/(t2(i)*t2(j)),j=1,i);
            enddo;

            write (6,'(/"x-LS =",2(/10g13.5))') xe(1:ne);
            be(:) =b(:) -matmul(ah(:,1:ne),xe(1:ne));
            write (6,'(/"b-err =",4(/10g13.5))') be(:);
            write (6,'("RSS(b-err) =",g13.5)') sqrt(sum(be(:)**2));
          endif;
        else;
          call svd_est (xe,  u,w,v,b,nrd,ne,nrd,nx);
        endif;

        be(:) =b(:) -matmul(ah(:,1:ne),xe(1:ne));
        besum(:) =besum(:) +be(:)**2;
        bemax(:) =max(bemax(:),abs(be(:)));

        !*** now compute evaluation residuals
        call eval_resid (besume,bemaxe,                                 &
     &               x,xe,xref,ne,nx,mirrorLstEval,nscanE,              &
     &               GOEStype,testH,xdet0,delDet,estim,visible);

        if (debugP) then;
          write (6,'(/6("**********")/"Monte Carlo sample ",i3)') isamp;
          write (6,'("x-SVD =",2(/10g13.5))') x(1:ne);
          write (6,'(/"b-err (microrad) =",3(/10f13.4))') 1.d6*be(:);
          write (6,'("RSS(b-err) =",f13.4)') 1.d6*sqrt(sum(be(:)**2));
        endif;
      enddo;  !___  end Monte Carlo (isamp) loop

      write (6,'(/10("**********")/"For ",i3,                           &
     &       " Monte carlo samples, RMS(b-err) (microrad) =",           &
     &       4(/10f13.4))') nsamp,1.d6*sqrt(besum(:)/nsamp);
      write (6,'(/"max(b-err) (microrad) =",4(/10f13.4))') 1.d6*bemax(:);
      bem2(:) =0.d0;
      do i=1,nscan;
        j=2*i-1;
        bem2(:) =max(bem2,bemax(j:j+1));
      enddo;
      write (6,'(/"max(max(b-err)) =",f10.3," EW ",f10.3,               &
     &      " NS (microrad)")')1.d6*bem2(:);
      write (6,'("Overall RMS b-err =",f13.4," microrad")')             &
     &       1.d6*sqrt(sum(besum(:))/(nsamp*nrd));

      !*** evaluation
      bem2(:) =0.d0;
      bem2a(:) =0.d0;
      do i=1,nscanE;
        j=2*i-1;
        bem2(:) =max(bem2,bemaxE(j:j+1,1));
        bem2a(:) =max(bem2a,bemaxE(j:j+1,2));
      enddo;

      write (6,'(//"stepwise regression # times states selected"/       &
     &          "nsel =",13i4)') nsel(:n);
      write (6,'(/"state sets selected")');
      do i=1,nset;
        nj =njset(i);
        k =0;
        do j=1,n;
          if (jset(j,i) == 0) exit;
          k =j;
        enddo;
        write (6,'(i3,", # cases =",i3,", rrms =",f8.3,", states =",    &
     &       13i3)') i,nj,sqrt(rrmsj(i)/nj),jset(1:k,i);
      enddo;

      if (visible) then;
        write (6,'(/"Using visible detectors for evaluation")');
      else;
        write (6,'(/"Using IR detectors for evaluation")');
      endif;
      write (6,'("Evaluation max(max(b-err)) =",f10.3," EW ",f10.3,     &
     &      " NS (microrad)")') 1.d6*bem2(:);
      write (6,'("Overall RMS b-err =",f13.4," microrad")')             &
     &       1.d6*sqrt(sum(besumE(:,1))/(nsamp*nrdE));

      write (6,'(/"Alternate Evaluation using partials at origin: ",    &
     &       "max(max(b-err)) =",f10.3," EW ",                          &
     &      f10.3," NS (microrad)")') 1.d6*bem2a(:);
      write (6,'("Overall RMS b-err =",f13.4," microrad")')             &
     &       1.d6*sqrt(sum(besumE(:,2))/(nsamp*nrdE));

      stop;
      contains;
!**********************************************************************
      subroutine init_svd (mirrorList,mirrorLstEval,estim);
!** initialize variables and perform SVD analysis
!**      uses global variables
!              xref,xdet0,meanDet,DelDet,sigx,nx,nrd,nscan,nscanE,nmeas,
!              xnames,GOEStype,testH,ROM

      implicit none;
      logical(4),intent(out) :: estim(nx);
      real(8),intent(out) :: mirrorList(2,nscan),mirrorLstEval(2,nscanE);

      integer(4) n,ne,i;
      character(12) :: ename(nx);
      real(8) h(nrd,nx),hmag(nx),ah(nrd,nx),mirror(2),b(nrd);
      real(8) xdet1(3,nscan),scan0(2), smag, theta;
      real(8) sigxx(nx),sOcs0(3),scalx(nx);
      real(8) w(nx),v(nx,nx),u(nrd,nrd);
      real(8),parameter :: pi =3.141592653589793d0;
      !_____________________________________________________

      !** Imag EW and NS mirror angles (GOES NS angle must be multiplied by 2)
      mirrorList(:,1) = (/ -.067, -.067 /);  !11 deg circular FOR
!      mirrorList(:,1) = (/ -.10, -.10 /)
      mirrorList(:,2) = (/ -.10, +.0 /);
      mirrorList(:,3) = (/ -.067, +.067 /);  !11 deg circular FOR
!      mirrorList(:,3) = (/ -.10, +.10 /)
      mirrorList(:,4) = (/ .0, -.10 /);
      mirrorList(:,5) = (/ .020, .082 /);     !extra star rather than landmark ###
!      mirrorList(:,5) = (/ .0, .0 /)         !###  landmark at earth center
      mirrorList(:,6) = (/ .0, +.10 /);
      mirrorList(:,7) = (/ +.067, -.067 /);  !11 deg circular FOR
!      mirrorList(:,7) = (/ +.10, -.10 /)
      mirrorList(:,8) = (/ +.10, .0 /);
      mirrorList(:,9) = (/ +.067, +.067 /);  !11 deg circular FOR
!      mirrorList(:,9) = (/ +.10, +.10 /)
      mirrorList(:,10) = (/ -.0347, +.0695 /);
      mirrorList(:,11) = (/ +.0347, +.0695 /);
      mirrorList(:,12) = (/ -.0347, -.0695 /);
      mirrorList(:,13) = (/ +.0347, -.0695 /);
      mirrorList(:,14) = (/ -.0695, +.0347 /);
      mirrorList(:,15) = (/ +.0695, +.0347 /);
      mirrorList(:,16) = (/ -.0695, -.0347 /);
      mirrorList(:,17) = (/ +.0695, -.0347 /);

      mirrorLstEval(:,1) = (/ -.030, +.060 /);
      mirrorLstEval(:,2) = (/  .000, +.060 /);
      mirrorLstEval(:,3) = (/ +.030, +.060 /);
      mirrorLstEval(:,4) = (/ -.060, +.030 /);
      mirrorLstEval(:,5) = (/  .000, +.030 /);
      mirrorLstEval(:,6) = (/ +.060, +.030 /);
      mirrorLstEval(:,7) = (/ -.030, -.000 /);
      mirrorLstEval(:,8) = (/  .000,  .000 /);
      mirrorLstEval(:,9) = (/ +.030, +.000 /);
      mirrorLstEval(:,10) = (/ -.060, -.030 /);
      mirrorLstEval(:,11) = (/  .000, -.030 /);
      mirrorLstEval(:,12) = (/ +.060, -.030 /);
      mirrorLstEval(:,13) = (/ -.030, -.060 /);
      mirrorLstEval(:,14) = (/  .000, -.060 /);
      mirrorLstEval(:,15) = (/ +.030, -.060 /);

      if (GOEStype) then;
!        mirrorList(1,:) = mirrorList(1,:)
        mirrorList(2,:) = 2.d0*mirrorList(2,:);  !double GOES NS mirror angle
        mirrorLstEval(2,:) = 2.d0*mirrorLstEval(2,:);  !double GOES NS mirror angle
        if (ROM) then;
          estim(4)=.false.;
!          estim(5:6)=.false.      !detector offsets
          estim(7:8)=.false.;
          estim(9)=.false.;       !EW shaft 1   (doesn't help NS)
!          estim(10)=.false.       !EW shaft 2   (helps NS or detector offsets)
          estim(11)=.false.;      !EW shaft 3  (doesn't help NS)
          estim(12:13)=.false.;
          estim(14)=.false.;      !NS shaft 1  (slightly helps detector offsets)
          estim(15:16)=.false.;  !NS shaft 2,3
        else;
          estim(7)=.false.;   ! roll EW mirror misalign (not observable)
          estim(9)=.false.;   ! roll EW shaft misalign (not observable)
          estim(12:13)=.false.; !NS mirror misalign
        endif;
        write (6,'("GOES 1-mirror el-az")');

      else;
        write (6,'("Imag 2-mirror")');
        estim(4)=.false.;    !detector rotation not observable if using true locations
        estim(8)=.false.;    !EW mirror-2 not observable
        estim(12)=.false.;   !NS mirror-2 not observable
        if (ROM) then;
          !** best 5 parameter model is EW-s2 and NS-s3
          !** best 6 parameter model is EW-s2, NS-s2 and NS-s3
          !** best 7 parameter model is EW-s1, EW-s2, NS-s2 and NS-s3
          estim(5)=.false.;  !detector offset 1
          estim(6)=.false.;  !detector offset 2
          estim(7)=.false.;  !EW mirror 1
          estim(9)=.false.;    !EW shaft 1 (optional-2)
!          estim(10)=.false.    !EW shaft 2 (optional)
          estim(11)=.false.;   !EW shaft 3 (e-rotation)
          estim(13)=.false.;   !NS mirror 3
          estim(14)=.false.;   !NS shaft 1 (n-rotation)
          estim(15)=.false.;   !NS shaft 2 (optional)
!          estim(16)=.false.    !NS shaft 3 (optional-2)
        endif;
      endif;

      !** write star points for plotting
      write (10,3000);
 3000 format ('variables ="EW Scan Angle (deg)", "NS Scan Angle (deg)"')
      do i=1,nscan;
        write (10,'(2g13.5)') 2*57.296*mirrorlist(:,i);
      enddo;
      write (10,3010);
 3010 format ('ZONE T="eworigin", I=2, F=POINT'/5x,'-12.',5x,'0.'/      &
     &     5x,'12.',5x,'0.')
      write (10,3020);
 3020 format ('ZONE T="nsorigin", I=2, F=POINT'/5x,'0.',5x,'-12.'/      &
     &     5x,'0.',5x,'12.')
      write (10,3030);
 3030 format ('ZONE T="earth", I=101, F=POINT')
      do i=1,101;
        theta =0.01d0*i*2.d0*pi;
        write (10,'(2g13.5)') 8.667*cos(theta), 8.667*sin(theta);
      enddo;

      !*** reset estim flags if using off-nadir model
      if (ROM .and. testH) then;
        estim(4:14) =.false.;
        estim(15:16) =.true.;
      endif;
      write (6,'(/"xdet  =",3f12.6/"xdet0 =",3f12.6)') xdet,xdet0;
      write (6,'("mean detector y and z locations =",2f13.6)')meandet(:);
      write (6,                                                         &
     &   '("max range of uniform random detector y and z locations =",  &
     &      2f13.6)') deldet(:);
      write (6,'(/"1-sigma of Gaussian misalignments (microrad) =",     &
     &      2(/9f11.2))') 1.d6*sigx(1:nx);
      if (testH) write (6,'(/"Using off-axis misalignment model")');
      l =0;
      do j=1,nx;
        if (estim(j)) then;
          l =l+1;
          ename(l) =xnames(j);
          sigxx(l) =sigx(j);
        endif;
      enddo;
!      ne =l
      if (ROM) write (6,'(/"Using reduced-order model ",i4)') l;
      write (6,'("estimate: ",8(a12,2x)/8(a12,2x))') ename(1:l);
      write (6,'(/"scan #      mirror ang            scan vector",      &
     &     "        mag error")');

      do i=1,nscan;
        mirror(:) =mirrorList(:,i);
        xdet1(:,i) =xdet0(:);
        if (GOEStype) then;
          call scanGOES2 (sOcs0, mirror, xref,nx,xdet0);
        else;
          call scanImag2 (sOcs0, mirror, xref,nx,xdet0);
        endif;
        scan0(:) = (/ asin(sOcs0(1)), -atan(sOcs0(2)/sOcs0(3)) /); !GOES EW, NS
        smag =1.d0-sqrt(sum(sOcs0(:)**2));
        write (6,'(i3,2x,2f10.4,2x,3f8.4,2x,g12.5)') i,mirror,sOcs0,smag;
      enddo;

      !*** get partials at detector origin for no misalignments
      call hPartial (h,ah,n,                                            &
     &            mirrorlist,xref,xdet1,xdet0,nx,nrd,nscan,nmeas,       &
     &            GOEStype,testH,estim);

      ne =n;   !compressed states
!      nscan*nmeas =nrd  (checked in hpartial)
      do j=1,ne;
        hmag(j) =sqrt(sum(ah(1:nrd,j)**2));
      enddo;
      write (6,'(/"nrd,nx,ne =",3i3)') nrd,nx,ne;
      write (6,'("hmag =",2(/10g12.5))') hmag(:ne);
      write (6,'("sum(hmag2) =",g12.5)') sum(hmag(:ne)**2);
      write (6,'(/"estim =",10l5,2x,10l5/)') estim;

      !*** compute svd (output u is in h)
!      scalx(:ne) =1.d0
!      scalx(:ne) =sigxx(:ne)*1.d6     !normalize for svd using prior magnitude ###
      scalx(:ne) =1.d0/hmag(:ne);     !normalize for svd using norm of h columns ###
      forall (j=1:ne) u(:,j) =ah(:,j)*scalx(j);
!      call svdcmp(u,nrd,ne,nrd,nx,  w,v)  original version
!      call svdcmp (w,v,  u,  nrd,ne,nrd,nx)
      b(:) =0.d0;
      call svd_lap (w,v, u, b,nrd,ne,nrd,nx);

      !*** IMSL routine
!      call IMSL_svd (ah,nrd,ne,nrd,nx,  u,w,v)

      write (6,'(/"singular values and v columns")');
      do i=1,ne;
        write (6,'(/i3,g13.5,2(/17f8.4))') i,w(i),v(1:ne,i);
      enddo;

      !** test svd accuracy by inverting
      forall (j=1:ne) ah(:,j) =ah(:,j)*scalx(j);
      call tst_svd (ah,u,w,v,nrd,ne,nrd,nx);

      return;
      end subroutine init_svd;

      end program;
!*********************************************************************
      subroutine hPartial (h,ah,n,                                      &
     &               mirrorlist,xref,xdet1,xdet0,nx,nrd,nscan,nmeas,    &
     &               GOEStype,testH,estim);

      !*** compute partials numerically
      implicit none;

      integer(4),intent(in) :: nx,nrd,nscan,nmeas;
      real(8),intent(in) :: mirrorlist(2,nscan); !mirror angles for each scan
      real(8),intent(in) :: xref(nx);        !misalignment vector for 0 misalignments
      real(8),intent(in) :: xdet1(3,nscan);  !detector locations for each scan
      real(8),intent(in) :: xdet0(3);        !nominal detector location
      logical(4),intent(in) :: GOEStype, testH; !test flags
      logical(4),intent(in) :: estim(nx);    !denoted whether state is estimated in ROM
      real(8),intent(out) :: h(nrd,nx);      !meas partials (full)
      real(8),intent(out) :: ah(nrd,nx);     !meas partials (compressed)
      integer(4),intent(out) :: n;           !number of compressed states in ah

      real(8) x(nx),sOcs0(3),sOcs(3),x0,SE,CE,SN,CN,delm(2),xdet(3);
      real(8) mirror(2),scan0(2),scan(2),scan1(2),h1(nrd,nx);
      real(8) :: delx =50d-6;
      integer(4) i,j,l;
      !___________________________________________________

      h(:,:) =0.d0;
      l=0;
      do i=1,nscan;
        mirror(:) =mirrorList(:,i);
        delm(:) =(0.5d0/xdet1(1,i))*                                    &
     &          (/ -xdet1(2,i), -xdet1(3,i)/cos(2.d0*mirror(1)) /);
!        mirror(:) =mirror(:) +delm(:)
        xdet(:) =xdet1(:,i);
!        xdet(:) =xdet0(:)

        if (GOEStype) then;
          call scanGOES2 (sOcs0, mirror, xref,nx,xdet);
        else;
          call scanImag2 (sOcs0, mirror, xref,nx,xdet);
        endif;
        scan0(:) = (/ asin(sOcs0(1)), -atan(sOcs0(2)/sOcs0(3)) /); !GOES EW, NS
        x(:) =xref(:);

        do j=1,nx;
          x0 =x(j);
          x(j) =x0+delx;
          if (GOEStype) then;
            call scanGOES2 (sOcs, mirror, x,nx,xdet);
          else;
            call scanImag2 (sOcs, mirror, x,nx,xdet);
          endif;
          scan(:) = (/ asin(sOcs(1)), -atan(sOcs(2)/sOcs(3)) /); !GOES EW, NS

          x(j) =x0-delx;
          if (GOEStype) then;
            call scanGOES2 (sOcs, mirror, x,nx,xdet);
          else;
            call scanImag2 (sOcs, mirror, x,nx,xdet);
          endif;
          scan1(:) = (/ asin(sOcs(1)), -atan(sOcs(2)/sOcs(3)) /); !GOES EW, NS
          h(l+1:l+nmeas,j) =(scan(:)-scan1(:))/(2.d0*delx);  !no significant difference in results
!          h(l+1:l+nmeas,j) =(scan(:)-scan0(:))/delx
          x(j) =x0;
        enddo;

        !*** test off-nadir misalignment model
        if (testH) then;
          write (*,'(/                                                  &
     &      "Using Off-Nadir Misalignment model for states 15 and 16")');
          SE =sin(scan0(1));
          CE =cos(scan0(1));
          SN =sin(scan0(2));
          CN =cos(scan0(2));
          if (GOEStype) then;
            h1(l+1:l+nmeas,1:3) =h(l+1:l+nmeas,1:3);
            h1(l+1:l+nmeas,4:14) =0.d0;
            h1(l+1:l+nmeas,15) = (/ -SN, 0.d0 /);                    !-Mdy
            h1(l+1:l+nmeas,16) = (/ 1.d0-CN/CE, SN*(1.d0+SE)/CE /);  !-Mdz
          else;
            h1(l+1:l+nmeas,1:3) =h(l+1:l+nmeas,1:3);
            h1(l+1:l+nmeas,4:14) =0.d0;
            h1(l+1:l+nmeas,15) = (/ 0.d0, (1.d0+SE-CE)/CE  /);       !-Me2
            h1(l+1:l+nmeas,16) = (/ 1.d0-SN-CN, (CN-SN)*SE/CE /);    !-Mn3
!            h1(l+1:l+nmeas,15) = (/ 0.d0, 0.d0  /)                  !-Me1 (5 times worse)
!            h1(l+1:l+nmeas,16) = (/ 0.d0, -SE/CE /)                 !-Mn2
          endif;
        endif;
        l=l+nmeas;
      enddo;

      if (l /= nrd) stop 'l /= nrd';

!***   only use selected states
      n=0;
      do i=1,nx;
        if (estim(i)) then;
          n=n+1;
          if (testH) then;
            ah(:,n) =h1(:,i);  !** test GOES misalignment model
          else;
            ah(:,n) =h(:,i);
          endif;
        endif;
      enddo;

      return;
      end subroutine hPartial;
!*****************************************************************
      subroutine eval_resid (besume,bemaxe,                             &
     &               x,xe,xref,ne,nx,mirrorLstEval,nscanE,              &
     &               GOEStype,testH,xdet0,delDet,estim,visible);
!***    compute evaluation residuals
      implicit none;
      integer(4),intent(in) :: ne, nx, nscanE;
      logical(4),intent(in) :: GOEStype,testH,estim(nx),visible;
      real(8),intent(in) :: xe(ne);     !estimated misalignment states
      real(8),intent(in) :: x(nx);      !true misalignment states
      real(8),intent(in) :: xref(nx);   !null misalignment states
      real(8),intent(in) :: xdet0(3),deldet(2);
      real(8),intent(in) :: mirrorLstEval(2,nscanE);
      real(8),intent(inout) :: besume(2*nscanE,2),bemaxe(2*nscanE,2);

      integer(4) i,j,l,n;
      real(8) b(2*nscanE),be(2*nscanE),mirror(2),sOcs(3),sOcs0(3);
      real(8) x1(nx),h2(2),scan(2),scan0(2),x0,xdet(3),xdet1(3);
      real(8) :: delx =50d-6;
      real(8),external :: uni;
      logical(4),save :: first =.false.;
      real(8) :: delDet2(2) =(/ 16600d-6, 7850d-6 /);   !max y and z detector offset from mean
      real(8) :: meanDet(2) =(/ -3300d-6, 0.d0 /);      !mean y and z detector offset from origin (IR)
      !________________________________________________________

      l=0;
      do i=1,nscanE;
        mirror(:) =mirrorLstEval(:,i);
        !*** randomize detector locations
        xdet1(1) =xdet0(1);
        if (visible) then;
          xdet1(2) =xdet0(2) +2.d0*delDet(1)*(uni(0) -0.5);
          xdet1(3) =xdet0(3) +2.d0*delDet(2)*(uni(0) -0.5);
        else;
          !*** IR
          xdet1(2) =xdet0(2) +meanDet(1) +2.d0*delDet2(1)*(uni(0) -0.5);
          xdet1(3) =xdet0(3) +meanDet(2) +2.d0*delDet2(2)*(uni(0) -0.5);
        endif;

        !**  compute no misalignments and misalignment-affected scan angles
        if (GOEStype) then;
          call scanGOES2 (sOcs0, mirror, xref,nx,xdet1);
          call scanGOES2 (sOcs, mirror, x,nx,xdet1);
        else;
          call scanImag2 (sOcs0, mirror, xref,nx,xdet1);
          call scanImag2 (sOcs, mirror, x,nx,xdet1);
        endif;
        scan0(:) = (/ asin(sOcs0(1)), -atan(sOcs0(2)/sOcs0(3)) /); !GOES EW, NS
        b(l+1:l+2) = (/ asin(sOcs(1)), -atan(sOcs(2)/sOcs(3)) /);
        be(l+1:l+2) =b(l+1:l+2) -scan0(:);

        x1(:) =xref(:);
        n=0;
        do j=1,nx;
          if (estim(j)) then;
            n =n+1;
            x1(j) =xe(n);
          endif;
        enddo;

        if (GOEStype) then;
          call scanGOES2 (sOcs, mirror, x1,nx,xdet1);
        else;
          call scanImag2 (sOcs, mirror, x1,nx,xdet1);
        endif;
        scan(:) = (/ asin(sOcs(1)), -atan(sOcs(2)/sOcs(3)) /); !GOES EW, NS
        b(l+1:l+2) =b(l+1:l+2) -scan(:);

        !*** get partials wrt misalignments evaluated at nominal detector location
        xdet(:) =xdet0(:);

        !*** first evaluate for no misalignments
        if (GOEStype) then;
          call scanGOES2 (sOcs0, mirror, xref,nx,xdet);
        else;
          call scanImag2 (sOcs0, mirror, xref,nx,xdet);
        endif;
        scan0(:) = (/ asin(sOcs0(1)), -atan(sOcs0(2)/sOcs0(3)) /); !GOES EW, NS

        !*** get partials wrt misalignments
        x1(:) =xref(:);
        n =0;
        do j=1,nx;
          if (estim(j)) then;
            n =n+1;
            if (testH) then;
!              h2(:) =h1(:,j)   !** test GOES misalignment model
              write (6,'("testH not coded for evaluation")');
            else;
              x0 =x1(j);
              x1(j) =x0+delx;
              if (GOEStype) then;
                call scanGOES2 (sOcs, mirror, x1,nx,xdet);
              else;
                call scanImag2 (sOcs, mirror, x1,nx,xdet);
              endif;
              scan(:) = (/ asin(sOcs(1)), -atan(sOcs(2)/sOcs(3)) /); !GOES EW, NS
              h2(:) =(scan(:)-scan0(:))/(delx);
              be(l+1:l+2) =be(l+1:l+2) -h2(:)*xe(n);
              x1(j) =x0;
            endif;
          endif;
        enddo;

        l =l+2;
      enddo;   !_____ end scan loop

      if (first) then;
        write (6,'(/"Evaluation b =",4(/10g13.5))') 1.d6*b(:);
        write (6,'("RSS(b) =",g13.5)') 1.d6*sqrt(sum(b(:)**2));
        write (6,'(/"b-err =",4(/10g13.5))') 1.d6*be(:);
        write (6,'("RSS(b-err) =",g13.5)') 1.d6*sqrt(sum(be(:)**2));
        first =.false.;
      endif;

      besume(:,1) =besume(:,1) +b(:)**2;
      bemaxe(:,1) =max(bemaxe(:,1),abs(b(:)));
      besume(:,2) =besume(:,2) +be(:)**2;
      bemaxe(:,2) =max(bemaxe(:,2),abs(be(:)));

      return;
      end subroutine eval_resid;
!*********************************************************
      subroutine tst_svd (ah,u,w,v,nr,nc,nrd,ncd);
      implicit none;
      integer(4),intent(in) :: nr,nc,nrd,ncd;
      real(8),intent(in) :: ah(nrd,nc),u(nrd,nc),w(nc),v(ncd,nc);
      integer(4) i,j;
      real(8) c(nc),y(nr),t2(nc),t3(nc,nc);
      !________________________________

      !*** test svd
      c(:) =1.d0;
      y(:) =matmul(ah(1:nr,1:nc),c(1:nc));

      call svd_est (t2,  u,w,v,y,nr,nc,nrd,ncd);

      !** t2 =V *t1 -c (should = 0)
      t2(:nc) =t2(:) -c(:);
!      write (6,'("c=",2(/10g13.5))') c
      write (6,'(/"tst_svd err =")');
      do i=1,nc;
        if (abs(t2(i)) > 1.d-12) write (6,'(i3,g13.5)') i,t2(i);
      enddo;

      write (6,'("testing v^T*v-I")');
      t3(:,:) =matmul(transpose(v(1:nc,:)),v(1:nc,:));
      do i=1,nc;
        t3(i,i) =t3(i,i)-1.d0;
        do j=1,nc;
          if (abs(t3(i,j)) > 1.d-12) write (6,'(2i3,g13.5)') i,j,t3(i,j);
        enddo;
      enddo;

      write (6,'("testing u^T*u-I")');
      t3(:,:) =matmul(transpose(u(1:nr,:)),u(1:nr,:));
      do i=1,nc;
        t3(i,i) =t3(i,i)-1.d0;
        do j=1,nc;
          if (abs(t3(i,j)) > 1.d-12) write (6,'(2i3,g13.5)') i,j,t3(i,j);
        enddo;
      enddo;
      write (6,'("finished testing u^T*u-I")');

      return;
      end subroutine tst_svd;
!*********************************************************
      subroutine svd_est (x,  u,w,v,y,nr,nc,nrd,ncd);
      !*** computes least squares estimate using SVD factors
      !*** x = V*W^(-1)*(U^T*y)
      implicit none;
      integer(4),intent(in) :: nr,nc,nrd,ncd;
      real(8),intent(in) :: u(nrd,nc),w(nc),v(ncd,nc),y(nr);
      real(8),intent(out) :: x(nc);
      integer(4) i;
      real(8) t1(nc);
      !________________________________

      !**  t1 =U^T * y
      t1(:nc) =matmul(y(:),u(:nr,:nc));   !=transpose(u(1:nr,1:nc)),y(:))

      !**  t1 =W^(-1)*t1
      do i=1,nc;
        if (w(i)/w(1) > 1.d-13) then;
          t1(i) =t1(i)/w(i);
        else;
          t1(i) =0.d0;
          write (6,'("svd_est: pseudo-inverse at row ",i3)') i;
        endif;
      enddo;

      !** t2 =V *t1  (t2 should = c if y=h*c)
      x(:) =matmul(v(:nc,:nc),t1(:nc));

      return;
      end subroutine svd_est;
!************************************************************
      subroutine svd_lap (s,v, u, y,nr,nc,nrd,ncd);
      !*** computes SVD using LAPACK
      implicit none;
      integer(4),intent(in) :: nr,nc,nrd,ncd;
      real(8),intent(in) :: y(nr);
      real(8),intent(inout) :: u(nrd,nr);
      real(8),intent(out) :: s(nc),v(ncd,nc);

      integer(4) i,m,n,mmax,nd,ier;
      integer(4),parameter :: lwork=10000;
      integer(4) IWORK(8*min(ncd,nrd)),llwork;
      logical(4) :: scaleSVD =.false.;
      real(8) h1(nrd,nc),h2(nrd,nc),WORK(lwork),e(nc),dscal(nc);
      real(8) vt(ncd,nc);
      real(8) :: singTol =1.d-11;   !pseudo-inverse singular value tolerance  ###
      !________________________________

      m =nr;
      n =nc;
      mmax =nrd;
      nd =ncd;

      h1(:m,:n) =u(:m,:n);
      if (scaleSVD) then;
        forall (i=1:n) dscal(i) =sqrt(sum(h1(:m,i)**2));
        forall (i=1:n) h1(:m,i) =h1(:m,i)/dscal(i);
      else;
        dscal(:) =1.d0;
      endif;

      h2(:nr,:nc) =h1(:nr,:nc);

      !** LAPACK SVD - SDD is divide and conquer (S denotes thin, A denotes all))
      llwork =lwork;
      call DGESDD( "S", m, N, h2, mmax, s, U, mmax, VT, nd, WORK,       &
     &              llwork, IWORK, ier );

      if (llwork < 0) write (*,'("work(1) =",g13.5)') work(1);
      if (ier /= 0) then;
        write (*,'("DGESDD ier =",i5)') ier;
        stop;
      endif;

!      h2(:m,:n) =u(:m,:n)           !###
      v(:n,:n) =transpose(vt(:n,:n)); !###

      e(:n) =matmul(y(:m),u(:m,:n));!transpose(u(:m,:n)),y(:m))  != U**T * y
      !** xe = V * S^(-1) * U**T * y
      do i=1,n;
        if (s(i) > s(1)*singTol) then;
          e(i) =e(i)/s(i);
        else;
          e(i) =0.d0;
        endif;
      enddo;
!      xe(:n) =matmul(b(:n,:n),e(:n,1)) /dscal(:n)

      return;
      end subroutine svd_lap;
