!** run_LS_cov.f is a test driver for linear least-squares covariance error analysis.
!** LS_cov is general-purpose and may be used with other models.
!** author B. Gibbs  1/2011

!**************************************
! The author grants the user a non-exclusive, worldwide, royalty-free copyright license to

! 1. reproduce and modify the software for your own purposes, and
! 2. to distribute the software provided that you give credit to the author,
!    do not make claims against the author or Wiley-Interscience,
!    and mark any modifications as your own.  If the software is incorporated in
!    a commercial product, you agree to defend and indemnify this author and
!    Wiley-Interscience against any losses, damages and costs arising from claims,
!    lawsuits and other legal actions brought by a third party.

! The software is provided on an as is basis, without warranties or conditions
! of any kind, either express or implied (including any warranties or conditions
! of title, non-infringement, merchantability or fitness for a particular purpose).
! The user is solely responsible for determining the appropriateness of using and
! distributing the program and assumes all risks associated with its exercise of
! rights under this agreement.
!**************************************

!**   using Intel compiler:
!**        in ifort.cfg
!**   /check:bounds /traceback /include:\kf_book\distribSW\library /link /nodefaultlib:msvcrt.lib
!**     \kf_book\distribSW\library\libkf_bnd_ifort.lib

!**   set ifortcfg=\kf_book\distribSW\ifort.cfg
!**   ifort LSmodel.f90 /c
!**   ifort LS_cov.f90 /c
!**   ifort run_LScov.f90 LSmodel.obj LS_cov.obj
!**   run_LScov poly.cfg >ls_covp.lis
!**************************************

      use LSmodel, only:ntotal;
      use LScovMod, only:read_setup,LS_cov,xstate;
      implicit none;
      integer(4) lenferr,nargs,ier;
      character(80) line;
      character(80) setupFile;
      !_______________________________

      lenferr =nargs()-1;         !get number of command line arguments

! ....  Get run control file name
      if (lenferr == 0) then;
        write (6,'(" Input setup file:")');
        read (5,'(a40)') setupFile;
      else;
        call getarg(1,line,ier);    !get command line argument
        read (line,*) setupFile;
        if (ier < 0) stop 'ERROR IN RETRIEVING setupFile';
      endif;

      if (ntotal <= 0) call errhand ("error: model state size is <= 0");

!** read text control file to set the xstate structure.  Note that variables ntotal and xlabel
!** must have been set by module LSmodel before calling setup.  setup also sets arrays
!** mapxe,mapxu,mapxei,mapxui that define the mappings between estimated and unestimated states

      call read_setup (setupFile);

!**    run least-squares covariance error analysis
      call LS_cov;

      stop
      end program