!** Program montepoly_subOpt computes, using Monte Carlo simulation, the sample
!** covariance and fit/prediction residuals for suboptimal least-squares estimates
!** using a polynomial model.  It generates output plot files sigyMont.plt and
!** sigx.plt.  This is used in Example 6.2-3.

!** Note: montePoly_constr.f uses library libkf_bnd_ifort.lib (rnoruni, sinv, symult, prtmx)

!** sample compile and execute using Intel compiler:

!**     Include in ifort.cfg:
!**   /check:bounds /traceback /include:\kf_book\distribSWm\library /link /nodefaultlib:msvcrt.lib
!**     \kf_book\distribSWm\library\libkf_bnd_ifort.lib

!**     To compile and execute:
!**   set ifortcfg=\kf_book\distribSWm\ifort.cfg
!**   ifort monte_poly_subOpt.f90   (libkf_bnd_ifort.lib should be linked by default)
!**   monte_poly_subOpt >sigyMont.lis
!**************************************************

      implicit none;
      integer(4),parameter :: nd =13, mmax=101, mmax2 =2*mmax;
      integer(4),parameter :: nsamp =1000;
      integer(4) i,j,k,n,m,kk,ier,ix,nn2,nt,n2,nmax,isamp,it;
      real(8) au(nd*(nd+1)/2),au2(nd*(nd+1)/2),r(nd*(nd+3)/2);
      real(8) x(nd),xa(nd),hf(mmax,nd+1),hp(mmax,nd+1),aa(nd,nd);   !,xe(nd),dx(nd)
      real(8) y(mmax2),yr(mmax2),bb(nd),xe(nd),pa(nd),delc,c(nd,nd);
      real(8) dt,eps,cnb,errm,sigy(6),lnLike,sumstat(2,6,nd);
      real(8) ye(mmax2),sigx(nd),sigxx(nd),cc(mmax+nd),ypr(mmax);
      real(8) sum2_col,cost,costl,delcEst,delcEstT;
!      real(8) :: pi =3.141592653589793d0
      real(8),external :: rnor;
      !____________________________________________________

      m =mmax;
      ix =191919;
      cnb =rnor(ix);

      nmax =13;
      nt =9;   !true model order
!      pa(:) =1.d10        !###
      pa(:) =10.d0;
      sumstat(:,:,:) =0.d0;

      !** measurement model
      dt =1.d0/(mmax-1);
      call get_haPoly (hf,au,mmax,nd,nmax,dt,0.d0);      !polynomial span 0 to 1
      call get_haPoly (hp,au2,mmax,nd,nmax,dt,1.d0);    !span 1 to 2

      kk=0;
      do i=1,nmax;
        do j=1,i;
          kk =kk+1;
          aa(i,j) =au(kk);
          aa(j,i) =au(kk);
        enddo;
      enddo;

      !** set prior (mean) state values
      xa(:) =0.d0;
      do i=1,nmax;
!        xa(i) =0.5*pa(i)*rnor(0)    !###
      enddo;

      open (10,file='sigyMont.plt',status='unknown');
      write (10,1100);
 1100 format('VARIABLES ="n", "sigy1", "sigy51", "sigy101", ',          &
     &      '"sigy151", "sigy201", "2J+f-m"')

      open (11,file='sigx.plt',status='unknown');
      write (11,1200);
 1200 format('VARIABLES ="n", "sigma-x1", "sigma-x2", "sigma-x3", ',    &
     &    '"sigma-x4", "sigma-x5", "sigma-x6", "sigma-x7", "sigma-x8"')

!***    start sampling loop
      do isamp =1,nsamp;
        x(:) =xa(:);
        do i=1,nt;
          x(i) =xa(i) +pa(i)*rnor(0);      !add random component for 1:nt states
        enddo;
!       x(n+1:) =0.d0     !no model error ###

        do i=1,mmax2;
          ye(i) =rnor(0);        !###
        enddo;
        y(:m) =matmul(hf(:m,:nmax),x(:nmax));
        y(m+1:mmax2) =matmul(hp(:m,:nmax),x(:nmax));
        y(:) =y(:) +ye(:);    !add noise (default) ###
        costl =0.d0;
        delcEstT =0.;

        !**    start state order loop
        do n=1,nmax;
          nn2 =(n*(n+1))/2;

          !### test match between truth and estimator model
!          y(:m) =matmul(hf(:m,:n),x(:n)) +matmul(hf(:m,n+1:nmax),xa(n+1:nmax))+ye(:m)
          !### end test
          if (n < nmax) then;
            ypr(:m) =matmul(hf(:m,n+1:nmax),xa(n+1:nmax));
          else;
            ypr(:) =0.d0;
          endif;
          yr(:m) =y(:m)-ypr(:m);
          bb(:n) =matmul(yr(:m),hf(:m,:n));

          !************ normal equations using Cholesky factor
          r(:) =0.d0;
          kk =0;
          do j=1,n;
            do k=1,j;
              kk =kk+1;
              r(kk) =r(kk) +dot_product(hf(:m,j),hf(:m,k));
            enddo;
            bb(j) =bb(j) +xa(j)/pa(j)**2;
            r(kk) =r(kk) +1.d0/pa(j)**2;    !add prior
          enddo;

          eps =1.d-3;
          call sinv (ier,errm,  r,  n,eps,.false.);
          if (ier < 0) then;
            write (6,'("*** sinv error")');
            cycle;
          endif;

          xe(:nmax) =xa(:nmax);     !set default = prior for non-estimated states
          call symult(xe,  r,bb,n);
          !** test use of srif (no difference)    ###
!          call srif (xe,  hf,y(:mmax)-ypr,pa,xa,mmax,n)

!          dx(:n) =xe(:n) -x(:n)
!          rsos =sum(yr(:m)**2)
!          write (6,'("tr(covy) =",g22.15,", rsos =",g13.5)') trc,rsos

          yr(:m) =y(:m) -matmul(hf(:m,:nmax),xe(:nmax));  !** normalized by 1.0
          yr(m+1:) =y(m+1:) -matmul(hp(:m,:nmax),xe(:nmax));

          !** sum log likelihood
          cost = sum(yr(:m)**2) +sum(((xe(:n)-xa(:n))/pa(:n))**2);
          lnLike =cost; !n*log(2.d0*pi) +2.d0*sum(log(pa(:n)))

          !** compute noise-only state 1-sigma
          kk =0;
          do i=1,n;
            kk =kk+i;
            sigx(i) =sqrt(r(kk));
          enddo;
          delcEst =1.d0 +(xa(n)/pa(n))**2                               &
     &             -(xa(n)**2 +pa(n)**2)/sigx(n)**2;
          if (n > 3) then;
            delcEstT =delcEstT +delcEst;
            lnLike =cost -delcEstT;
          endif;

          if (isamp == 1) then;
            !** form R^(-1/2)*Hf*Px*Pa^(-1/2) and I-Pa^(-1/2)*Px*Pa^(-1/2)
            !** and sum last column for test on change in cost function
            k =((n-1)*n)/2;
            do i=1,m;
              cc(i) =dot_product(r(k+1:k+n),hf(i,:n))/pa(n);
            enddo;
            forall (i=1:n) cc(m+i) =-r(k+i)/(pa(i)*pa(n));
            cc(m+n) =cc(m+n) +1.d0;
            delc =sum(cc(:m+n)**2);
            call delcc (delc,sum2_col, r,hf,pa,n,mmax);   !no difference in delc ###
            write (6,'("n =",i3,", delc =",g13.5,", sum2_col =",g13.5,  &
     &          ", sigx(n) =",g13.5,", diff =",g13.5,", cost =",g13.5,  &
     &          ", delc2 =",2g13.5)')                                   &
     &         n ,delc, sum2_col, sigx(n), 1.d0-(sigx(n)/pa(n))**2-delc,&
     &          cost, cost-costl, delcEst;   !(xa(n)/pa(n))**2 -(xe(n)/sigx(n))**2
            costl =cost;

            if (nt > n) then;
              !** now compute total state 1-sigma including unadjusted states
              n2 =nt-n;
              do i=1,n2;
                call symult(c(:n,i),  r,aa(:n,i+n),n);
                c(:n,i) =c(:n,i)*pa(i+n);
              enddo;
              kk =0;
              sigxx(:) =pa(:);
              do i=1,n;
                kk =kk+i;
                sigxx(i) =sqrt(r(kk) +sum(c(i,:n2)**2));
              enddo;
              write (11,'(i3,8g13.5)') n,sigxx(:8);
            endif;
          endif;    !___ end isamp == 1 test

          do i=1,5;
            it =(i-1)*50+1;
            sumstat(1,i,n) =sumstat(1,i,n) +yr(it);
            sumstat(2,i,n) =sumstat(2,i,n) +yr(it)**2;
          enddo;
          sumstat(1,6,n) =sumstat(1,6,n) +lnLike-m;
          sumstat(2,6,n) =sumstat(2,6,n) +(lnLike-m)**2;
        enddo;     !___ end order loop

      enddo;   !___ end isamp loop

      do n=1,nmax;
        sigy(:5) =sqrt(sumstat(2,:5,n)/nsamp);
        write (10,'(i3,6g13.5)') n,sigy(:5),sumstat(1,6,n)/nsamp;
      enddo;

      stop;
      end program;
!*******************************************************************
      subroutine get_haPoly (h,au,  mmax,nd,n,dt,t0);
      implicit none;

      integer(4),intent(in) :: mmax,nd,n;
      real(8),intent(in) :: dt,t0;
      real(8),intent(out) :: h(mmax,nd),au(nd*(nd+1)/2);

      integer(4) it,j,k,kk;
      real(8) dtn,t,hnorm(mmax);
      !_________________________________________

      au(:) =0.d0;
      h(:,:) =0.d0;

      do it=1,mmax;
        t =(it-1)*dt +t0;
        dtn =1.d0;
        kk =0;
        do j=1,n;
          h(it,j) =dtn;
          dtn =dtn*t;
          do k=1,j;
            kk =kk+1;
            au(kk) =au(kk) +h(it,j)*h(it,k);
          enddo;
        enddo;
        hnorm(it) =sqrt(sum(h(it,1:n)**2));
      enddo;    !end time loop

      write (*,'("n =",i3,", hnorm =",10(/10g13.5))') n,hnorm(1:mmax);
      return;
      end subroutine get_haPoly;
!*******************************************************************
      subroutine get_haPolyC (h,au,  mmax,nd,n,dt,t0);
      !** Chebyshey polynomial
      implicit none;

      integer(4),intent(in) :: mmax,nd,n;
      real(8),intent(in) :: dt,t0;
      real(8),intent(out) :: h(mmax,nd+1),au(nd*(nd+1)/2);

      integer(4) it,j,k,kk;
      real(8) t,hnorm(mmax),c(n),cheb,dfdc(n),a,b;
      !_________________________________________

      au(:) =0.d0;
      h(:,:) =0.d0;
!      dt =1.d0/(mmax-1)
      a =0.d0;
      b =2.d0*(mmax-1)*dt;
      c(:) =0.d0;    !set chebyshey coeff

      do it=1,mmax;
        t =(it-1)*dt +t0;
        kk =0;
        call chebyPoly (cheb,dfdc,  a,b,c,n,t);
!        write (6,'("dfdc =",10g13.5)') dfdc(:n)
        h(it,:n) =dfdc(:n);
        do j=1,n;
          do k=1,j;
            kk =kk+1;
            au(kk) =au(kk) +h(it,j)*h(it,k);
          enddo;
        enddo;
        hnorm(it) =sqrt(sum(h(it,1:n)**2));
      enddo;    !end time loop

      write (*,'("n =",i3,", hnorm =",10(/10g13.5))') n,hnorm(1:mmax);

      return;
      end subroutine get_haPolyC;
!******************************************
      subroutine chebyPoly (fx,dfdc,  x1,x2,c,m,x);
      !** evaluates Chebyshev polynomial at a specific point x, and also computes
      !** partial derivatives WRT the polynomial coefficients
      implicit none;
      integer(4),intent(in) :: m;   !order of model
      real(8),intent(in) :: x;      !intermediate point between x1 and x2 for evaluation
      real(8),intent(in) :: x1,x2;  !start and end points of interval
      real(8),intent(in) :: c(m);   !chebyshev coeff
      real(8),intent(out) :: fx;    !interpolated value at x
      real(8),intent(out) :: dfdc(m);!interpolated partials of fx wrt c

      integer(4) j;
      real(8) d,dd,sv,y,y2,dddc(m),ddddc(m),ddddcL(m);
      !__________________________________________

!      if ((x-x1)*(x-x2) > 0.) stop 'chebyPoly error: x not in range'
      d =0.d0;
      dd =0.d0;
      y =(2.d0*x-x1-x2)/(x2-x1);
      y2 =2.d0*y;
      dddc(:) =0.d0;
      ddddc(:) =0.d0;

      do j =m,2,-1;
        sv =d;
        ddddcL(:) =dddc(:);
        dddc(:) =y2*dddc(:) -ddddc(:);
        dddc(j) =dddc(j) +1.d0;
        d =y2*d -dd +c(j);
        dd =sv;
        ddddc(:) =ddddcL(:);
      enddo;

      fx =y*d -dd +0.5d0*c(1);
      dfdc(:) =y*dddc(:) -ddddc(:);
      dfdc(1) =dfdc(1) +0.5d0;

      return;
      end subroutine chebyPoly;
!*******************************************************************
      subroutine get_haPolyF (h,au,  mmax,nd,n,dt,t0);
      !** Fourier series
      implicit none;

      integer(4),intent(in) :: mmax,nd,n;
      real(8),intent(in) :: dt,t0;
      real(8),intent(out) :: h(mmax,nd+1),au(nd*(nd+1)/2);

      integer(4) it,i,j,k,kk,nf;
      real(8) t,hnorm(mmax),deltt,omeg;
      real(8) :: pi =3.141592653589793d0;
      !_________________________________________

      nf =n/2;
      deltt =2.d0*(mmax-1)*dt;

      au(:) =0.d0;
      h(:,:) =0.d0;
!      dt =1.d0/(mmax-1)

      do it=1,mmax;
        t =(it-1)*dt +t0;
        omeg =2.d0*pi*t/deltt;
        h(it,1) =1.d0;
        h(it,2:) =0.d0;
        j =1;
        do i=2,n;
          j =i/2;
          if (mod(i,2) == 0) then;
            h(it,i) =sin(j*omeg);
          else;
            h(it,i) =cos(j*omeg);
          endif;
        enddo;
!        write (6,'("dfdc =",10g13.5)') h(it,:n)
        kk =0;
        do j=1,n;
          do k=1,j;
            kk =kk+1;
            au(kk) =au(kk) +h(it,j)*h(it,k);
          enddo;
        enddo;
        hnorm(it) =sqrt(sum(h(it,1:n)**2));
      enddo;    !end time loop

      write (*,'("n =",i3,", hnorm =",10(/10g13.5))') n,hnorm(1:mmax);

      return;
      end subroutine get_haPolyF;
!********************************************
      subroutine srif (xe,  hf,y,pa,xa,mmax,n);
      !*** solve ls using SRIF (QR)
      implicit none;
      integer(4),intent(in) :: mmax,n;
      real(8),intent(in) :: y(mmax),hf(mmax,n),pa(n),xa(n);
      real(8),intent(out) :: xe(n);

      integer(4) i,m,nstrt,nn2,ier;
      real(8) r(n*(n+3)/2),rsos,h1(mmax+n,n+1),y1(mmax);
      !______________________________________________

      nn2 =(n*(n+1))/2;
      r(:) =0.d0;
      h1(:,:) =0.d0;
      h1(:mmax,:n) =hf(:,:);     !already normalized to unit meas noise
      h1(:mmax,n+1) =y(:mmax);
      do i=1,n;
        h1(mmax+i,i) =1.d0/pa(i);
        h1(mmax+i,n+1) =xa(i)/pa(i);
      enddo;
      nstrt =1;
      rsos =0.d0;    !computes RHS
      m =mmax+n;
      call thhc_m (r,h1,rsos,  n,mmax+n,m,nstrt,.false.);
!      call thhc (r,n,h1,mmax+n,m,rsos,nstrt)

      y1(:n) =r(nn2+1:nn2+n);
      call rinz_m (xe,ier,  r,n,y1);
      if (ier /= 0) stop "rinz error";

      return;
      end subroutine srif;
!*******************************************
      subroutine delcc (delc,sum2_col, r,hf,pa,n,mmax);
      !** checks with short calculation
      implicit none;
      integer(4),intent(in) :: mmax,n;
      real(8),intent(in) :: r(n*(n+1)/2),hf(mmax,n),pa(n);
      real(8),intent(out) :: delc,sum2_col;

      integer(4) i,j,k;
      real(8) mm(mmax+n,n),px(n,n),a(mmax+n,mmax+n);
      !______________________________________________

      k =0;
      do i=1,n;
        do j=1,i;
          k =k+1;
          px(i,j) =r(k);
          px(j,i) =r(k);
        enddo;
      enddo;

      mm(:,:) =0.d0;
      mm(:mmax,:) =hf(:,:);
      do i=1,n;
        mm(mmax+i,i) =1.d0/pa(i);
      enddo;
      a(:,:) =-matmul(matmul(mm,px),transpose(mm));
      forall (i=1:mmax+n) a(i,i) =a(i,i)+1.d0;
      delc =sum(a(:,mmax+n)**2);
!      sum2_col =sum(a(:,:)**2)      !=mmax
      sum2_col =0.d0;
      if (n > 9) then;
!        do i=1,mmax+min(9,n)
        do i=mmax+10,mmax+n;
          sum2_col =sum2_col +sum(a(i,:)**2);     !=mmax
        enddo;
      endif;

      return;
      end subroutine delcc;
