!** Program tst_poly generates simulated measurements for a polynomial model
!** and computes optimal or reduced-order least squares fits to the data.
!** Both the normal equations and SVD solutions are used.  The residuals
!** and the residuals covariance are written to plot files resid04.plt,
!** covy04.plt and covyIndx04.plt.  This is used in Example 6.1-1.

!**************************************
! The author grants the user a non-exclusive, worldwide, royalty-free copyright license to

! 1. reproduce and modify the software for your own purposes, and
! 2. to distribute the software provided that you give credit to the author,
!    do not make claims against the author or Wiley-Interscience,
!    and mark any modifications as your own.  If the software is incorporated in
!    a commercial product, you  agree to defend and indemnify this author and
!    Wiley-Interscience against any losses, damages and costs arising from claims,
!    lawsuits and other legal actions brought by a third party.

! The software is provided on an as is basis, without warranties or conditions
! of any kind, either express or implied (including any warranties or conditions
! of title, non-infringement, merchantability or fitness for a particular purpose).
! The user is solely responsible for determining the appropriateness of using and
! distributing the program and assumes all risks associated with its exercise of
! rights under this agreement.
!**************************************

!** sample compile and execute using Intel compiler:

!**     Include in ifort.cfg
!**   /check:bounds /traceback /include:\kf_book\distribSWm\library /link /nodefaultlib:msvcrt.lib
!**     \kf_book\distribSWm\library\libkf_bnd_ifort.lib \library\lapack-3.1.1\lapack_win32.lib \library\lapack-3.1.1\blas_win32.lib

!**     To compile and execute:
!**   set ifortcfg=\kf_book\distribSWm\ifort.cfg
!**   ifort tst_poly.f90    (libkf_bnd_ifort.lib and lapack should be linked by default)
!**   tst_poly >tstPoly.lis
!**************************************

      implicit none;
      integer(4),parameter :: nd =40, mmax=101, mmax2=2*mmax-1;
      integer(4),parameter :: lwork=10000;
      integer(4) i,j,k,n,m,m1,kk,ier,ix,nn2,nt;
      integer(4) IWORK(8*min(nd,mmax)),llwork;
      logical(4) :: scaleSVD=.false.;    !normally all false
      real(8) dt;
      real(8) a(nd,nd),al(nd*(nd+1)/2),r(nd*(nd+3)/2);
      real(8) b(nd,nd),e(nd,nd),eps,d(nd),x(nd),sigx(nd);
      real(8) h(mmax2,nd+1),h1(mmax,nd+1),h2(mmax,nd+1),bb(nd),dx(nd);
      real(8) cnb,conds,xe(nd),y(mmax2),y1(mmax),errm,dscal(nd);
      real(8) covy(mmax,mmax),yr(mmax2),trc,rsos;
      real(8) U(mmax,mmax), VT(nd,nd), WORK(lwork);   !LAPACK SVD
      real(8),external :: rnor,l2normM,l1normM,condest;
      character(20) fname;
      !____________________________________________________

      m =mmax;
      ix =191919;
      cnb =rnor(ix);

      nt =8;   !true model
      n =4;    !estimate order ###
      nn2 =(n*(n+1))/2;

      !** measurement model
      dt =1.d0/(mmax-1);
      call get_haPoly (h,al,mmax2,nd,nt,dt);      !polynomial
      do i=1,nt;
        x(i) =50.*rnor(0);
      enddo;
!      x(n+1:) =0.d0     !no model error ###
      y(:) =matmul(h(:,:nt),x(:nt));
      do i=1,mmax2;
        y(i) =y(i) +rnor(0);        !###
      enddo;

      write (fname,'("resid",i2.2,".plt")') n;
      open (10,file=fname,status='unknown');
      write (10,1100);
 1100 format('VARIABLES ="t", "measurement-residual"')

      write (fname,'("covy",i2.2,".plt")') n;
      open (11,file=fname,status='unknown');

      write (fname,'("covyIndx",i2.2,".plt")') n;
      open (12,file=fname,status='unknown');
      write (12,1200);
 1200 format('VARIABLES ="column index", "row index"')

      !** measurement model
      h1(:,:) =h(:m,:);
      call get_haPoly (h1,al,mmax,nd,n,dt);      !polynomial

      kk =0;
      do j=1,n;
        do k=1,j;
          kk =kk+1;
          a(k,j) =al(kk);
          a(j,k) =al(kk);
        enddo;
      enddo;

      h(:,n+1) =y(:);
      forall (i=1:n) bb(i) =dot_product(y,h(:m,i));

      !************ method 1: normal equations using Cholesky factor
      eps =1.d-3;
      r(:nn2) =al(:nn2);
      call sinv (ier,errm,  r,  n,eps,.false.);
      if (ier < 0) then;
        write (6,'("*** sinv error")');
      else;
        call symult (xe,  r,bb,n);
!        dx(:n) =xe(:n) -x(:n)

        j =0;
        do i=1,n;
          j =j+i;
          sigx(i) =sqrt(r(j));
        enddo;
        write (6,'("sigx =",10g13.5)') sigx(:n);
        !sigx(4) =  0.38368  3.3394  7.7817  5.1139
        !sigx(8) =  0.68657  25.833  308.00  1612.4  4296.3  6083.9  4358.8  1242.1

        !** h1 =ainv*h^T
        do i=1,m;
          call symult (h1(i,:),  r,h(i,:),n);
        enddo;
        covy(:,:) =matmul(h(:m,:n),transpose(h1(:m,:n)));
        trc =0.d0;
        do i=1,m;
          covy(i,i) =1.d0 -covy(i,i);
          trc =trc +covy(i,i);
        enddo;
        call prtmx (covy,mmax,m,m,"covy");

        yr(:) =y(:) -matmul(h(:,:n),xe(:n));
        rsos =sum(yr(:m)**2);
        write (6,'("tr(covy) =",g22.15,", rsos =",g13.5)') trc,rsos;
!        do i=1,m
!!          write (10,'(2g13.5)') dt*(i-1),yr(i)
!          write (11,'(102f13.5)') dt*(i-1),covy(i,:)
!        enddo

!        do i=1,m
!          do j=1,m
!            if (abs(covy(i,j)) > 0.05) then
!              write (12,'(2i6)') i,j
!            endif
!          enddo
!        enddo
      endif;

      !********************* method 4: svd
      m1 =m;
      h1(:,:) =h(:m,:);
      y1(:m) =y(:m);
      if (scaleSVD) then;
        forall (i=1:n) dscal(i) =sqrt(sum(h(:m,i)**2));
        forall (i=1:n) h1(:m,i) =h1(:m,i)/dscal(i);
      else;
        dscal(:) =1.d0;
      endif;

      h2(:,:) =h1(:,:);
      !** LAPACK SVD - SDD is divide and conquer
      llwork =lwork;
      call DGESDD( "A", m1, N, h2, mmax, d, U, mmax, VT, nd, WORK,      &
     &              llwork, IWORK, ier );

      if (llwork < 0) write (*,'("work(1) =",g13.5)') work(1);
      if (ier /= 0) then;
        write (*,'("DGESDD ier =",i5)') ier;
        stop;
      endif;
      !** next 2 lines only for LAPACK
      h2(:m1,:n) =u(:m1,:n);          !###
      b(:n,:n) =transpose(vt(:n,:n)); !###

      e(:n,1) =matmul(transpose(h2(1:m1,1:n)),y1(:m1)); != U**T * y
      !** xe = V * S^(-1) * U**T * y
      do i=1,n;
        e(i,1) =e(i,1)/d(i);
      enddo;
      xe(:n) =matmul(b(:n,:n),e(:n,1)) /dscal(:n);
!      dx(:n) =xe(:n) -x(:n)
      yr(:) =y(:) -matmul(h(:,:n),xe(:n));
      rsos =sum(yr(:m)**2);
      write (6,'(" rsos =",g13.5)') rsos;
      do i=1,mmax2;
        write (10,'(2g13.5)') dt*(i-1),yr(i);
      enddo;

      !** e =U**T * H
      e(:n,:n) =matmul(transpose(h2(1:m1,1:n)),h1(1:m1,1:n));
      !** C = V * S^(-1) * U**T * H -I
      forall (i=1:n) e(i,:) =e(i,:)/d(i);

      conds =maxval(d(:n))/minval(d(:n));
      write (6,'("svd sing values =",5(/10g13.5))') d(:n);
!      call predict (ystat, xe,xer,x,n,m,dt,hdiag)
      trc =0.d0;
      do i=1,m;
        trc =trc +sum(u(i,n+1:m)**2);
      enddo;
      write (6,'("tr(U2*U2^T) =",g22.15)') trc;

      !** compute correlation matrix using singular vectors
      covy(:,:) =matmul(u(:,n+1:m),transpose(u(:,n+1:m)));

      do i=1,m;
        write (11,'(102f13.5)') dt*(i-1),covy(i,:);
        do j=1,m;
          if (abs(covy(i,j)) > 0.05) then;
            write (12,'(2i6)') i,j;
          endif;
        enddo;
      enddo;

!      close (10)

      stop;
      end program;
!*******************************************************************
      subroutine get_haPoly (h,al,mmax,nd,n,dt);
      implicit none;

      integer(4),intent(in) :: mmax,nd,n;
      real(8),intent(in) :: dt;
      real(8),intent(out) :: h(mmax,nd+1),al(nd*(nd+1)/2);

      integer(4) it,jt,j,k,kk;
      real(8) dtn,t,hnorm(mmax);
      !_________________________________________

      al(:) =0.d0;
      h(:,:) =0.d0;

      do it=0,mmax-1;
        t =it*dt;
        dtn =1.d0;
        kk =0;
        jt =it+1;
        do j=1,n;
          h(jt,j) =dtn;
          dtn =dtn*t;
          do k=1,j;
            kk =kk+1;
            al(kk) =al(kk) +h(jt,j)*h(jt,k);
          enddo;
        enddo;
        hnorm(jt) =sqrt(sum(h(jt,1:n)**2));
      enddo;    !end time loop

!      write (*,'("n =",i3,", hnorm =",10(/10g13.5))') n,hnorm(1:mmax)
      return;
      end subroutine get_haPoly;

