!** Program tst_poly_subOpt computes the covariance of fit/prediction residuals
!** for a suboptimal least squares estimate using polynomial model.  It generates
!** output plot files sigy.plt.  This is used in Example 6.2-3.

!**************************************
! The author grants the user a non-exclusive, worldwide, royalty-free copyright license to

! 1. reproduce and modify the software for your own purposes, and
! 2. to distribute the software provided that you give credit to the author,
!    do not make claims against the author or Wiley-Interscience,
!    and mark any modifications as your own.  If the software is incorporated in
!    a commercial product, you  agree to defend and indemnify this author and
!    Wiley-Interscience against any losses, damages and costs arising from claims,
!    lawsuits and other legal actions brought by a third party.

! The software is provided on an as is basis, without warranties or conditions
! of any kind, either express or implied (including any warranties or conditions
! of title, non-infringement, merchantability or fitness for a particular purpose).
! The user is solely responsible for determining the appropriateness of using and
! distributing the program and assumes all risks associated with its exercise of
! rights under this agreement.
!**************************************

!** sample compile and execute using Intel compiler:

!**     Include in ifort.cfg
!**   /check:bounds /traceback /include:\kf_book\distribSWm\library /link /nodefaultlib:msvcrt.lib
!**     \kf_book\distribSWm\library\libkf_bnd_ifort.lib

!**     To compile and execute:
!**   set ifortcfg=\kf_book\distribSWm\ifort.cfg
!**   ifort tst_poly_subOpt.f90    (libkf_bnd_ifort.lib should be linked by default)
!**   tst_poly_subOpt >sigyCov.lis
!**************************************

      implicit none;
      integer(4),parameter :: nd =40, mmax=101, mmax2 =2*mmax;
      integer(4) i,j,k,n,m,kk,ier,ix,nn2,nt,im,n2,nmax,nmod,jm;
      real(8) a(nd,nd),au(nd*(nd+1)/2),au2(nd*(nd+1)/2),r(nd*(nd+3)/2);
      real(8) hf(mmax,nd+1),hp(mmax,nd+1),ha(mmax2,nd),hh(mmax2,nd);   !,xe(nd),dx(nd)
      real(8) py(mmax,mmax),pyt(mmax,mmax),y(mmax2),htt(mmax,nd),x(nd);
      real(8) a2(nd,nd),a3(nd,nd),pa(nd),e(nd,nd),c(nd,mmax),pxx(nd,nd);
      real(8) dt,eps,cnb,errm,sigy(5),lnLike,dof,delcEstT,delcEst;
      real(8) :: pi =3.141592653589793d0;
      real(8),external :: rnor;
      !____________________________________________________

      m =mmax;
      ix =191919;
      cnb =rnor(ix);

      nmax =13;
      nt =9;   !true model order
!      pa(:) =1.d10        !###
      pa(:) =10.d0;

      !** measurement model
      dt =1.d0/(mmax-1);
      call get_haPoly (hf,au,mmax,nd,nmax,dt,0.d0);      !polynomial span 0 to 1
      call get_haPoly (hp,au2,mmax,nd,nmax,dt,1.d0);    !span 1 to 2
      do i=1,nt;
        x(i) =pa(i)*rnor(0);
      enddo;
!      x(n+1:) =0.d0     !no model error ###
      y(:mmax) =matmul(hf(:,:nt),x(:nt));
      y(mmax+1:mmax2) =matmul(hp(:,:nt),x(:nt));
      do i=1,mmax2;
        y(i) =y(i) +rnor(0);        !###
      enddo;

      open (10,file='sigy.plt',status='unknown');
      write (10,1100);
 1100 format('VARIABLES ="n", "sigy1", "sigy51", "sigy101", ',          &
     &      '"sigy151", "sigy201", "2J+f-m"')

      delcEstT =0.d0;

!***    start state order loop
      do n=1,nmax;
        nn2 =(n*(n+1))/2;
!        h(:,n+1) =y(:)
!        forall (i=1:n) bb(i) =dot_product(y,h(:m,i))

        kk =0;
        do j=1,nt;
          do k=1,j;
            kk =kk+1;
            a(k,j) =au(kk);
            a(j,k) =au(kk);
          enddo;
        enddo;

        !** a2 =transpose(hf1)*hf2*pa^(1/2)
        n2 =nt-n;
        if (n2 > 0) forall (i=1:n2) a2(:,i) =a(:n,n+i)*pa(i+n);

        !************ normal equations using Cholesky factor
        eps =1.d-3;
        r(:nn2) =au(:nn2);
        !** add prior on x1 states
        k =0;
        do i=1,n;
          k =k+i;
          r(k) =r(k) +1.d0/pa(i)**2;
        enddo;

        call sinv (ier,errm,  r,  n,eps,.false.);
        if (ier < 0) then;
          write (6,'("*** sinv error")');
          cycle;
        endif;

        !** set a1 =p1
        kk =0;
        do j=1,n;
          do k=1,j;
            kk =kk+1;
            a(k,j) =r(kk);
            a(j,k) =r(kk);
          enddo;
        enddo;

        !** py =hf1*px1*hf1^T  (a(:n,:n) is px1)
        ha(:m,:n) =matmul(hf(:m,:n),a(:n,:n));    !=hf1*px1
        ha(m+1:,:n) =matmul(hp(:m,:n),a(:n,:n));  !=hp1*px1
        pyt(:m,:m) =matmul(hf(:m,:n),transpose(ha(:m,:n)));
        nmod =max(n,nt);
        forall (i=1:nmod) htt(:m,i) =hf(:m,i)*pa(i);  !=hf*pa^1/2
        !** e =pa1^(-1/2)*px
        forall (i=1:n) e(:n,i) =a(i,:n)/pa(i);

        !** pyt =r -hf1*px*hf1^T
        pyt(:m,:m) =-pyt(:m,:m);
        dof =n;
        do i=1,m;
          pyt(i,i) =pyt(i,i) +1.d0;
          dof =dof +pyt(i,i);
        enddo;
        write (6,'(i2," py =",3g13.5,", dof =",g13.5)') n,              &
     &    sqrt(pyt(1,1)), sqrt(pyt(51,51)), sqrt(pyt(101,101)), dof;

        if (n2 > 0) then;
          !** compute cross term (a3) used in x1,x2 error covariance and in meas residual covar
          !** a3 =px1*a2 =p1*transpose(hf1)*hf2*pa^(1/2)
          a3(:n,:n2) =matmul(a(:n,:n),a2(:n,:n2));

          !***** compute x1,x2 error covariance

          !** a11 =p1 +p1*a2*transpose(p1*a2)
          a(:n,:n) =a(:n,:n) +matmul(a3(:n,:n2),transpose(a3(:n,:n2)));
          call prtmx (a,nd,n,n,"x1 covariance");

          !** a12 =-px1*a2*pa^(1/2) =-px1*transpose(hf1)*hf2*pa
          a(:n,n+1:nt) =-a3(:n,:n2);
          forall (i=n+1:nt) a(:n,i) =a(:n,i)*pa(i);
          a(n+1:nt,:n) =transpose(a(:n,n+1:nt));

          !** a22 =pa
          a(n+1:nt,n+1:nt) =0.d0;
          forall (i=n+1:nt) a(i,i) =pa(i)**2;

          !***** compute 1-sigma perturbation of yf due x2 states
          !** hh1 =hf2*pa^(1/2) -hf1*a3
          forall (i=n+1:nt) hh(:m,i-n) =hf(:m,i)*pa(i);
          hh(:m,:n2) =hh(:m,:n2) -matmul(hf(:m,:n),a3(:n,:n2));
          pyt(:m,:m) =pyt(:m,:m)                                        & !## delete this to compute noise-only covar
     &        +matmul(hh(:m,:n2),transpose(hh(:m,:n2)));   !a posteriori meas resid covar

          !***** compute 1-sigma perturbation of yp due x2 states
          !**  hh2 =hp2*pa^(1/2) -hp1*a3
          forall (i=n+1:nt) hh(m+1:,i-n) =hp(:m,i)*pa(i);
          hh(m+1:,:n2) =hh(m+1:,:n2) -matmul(hp(:m,:n),a3(:n,:n2));
        endif;

        !*** start of -2*log(likelihood) calculations:
        !*** First compute a priori measurement residual covariance (including x2 states)
        !*** used in computing a posteriori state error to ln(like)
        !** py =hf*pa1*hf^T +hp*pa2*hp^T +r
        py(:m,:m) =matmul(htt(:m,:nmod),transpose(htt(:m,:nmod)));
        forall (i=1:m) py(i,i) =py(i,i) +1.d0;
        write (6,'("1-sigma of a-priori meas resid covar =",            &
     &          20(/10g13.5))') (sqrt(py(i,i)),i=1,m);

        !** now compute multipliers in a posteriori state error covariance
        !** c =e*hf^T*r1^(-1/2) =pa1^(-1/2)*px1*hf^T*r1^(-1/2)
        c(:n,:m) =matmul(e(:n,:n),transpose(hf(:m,:n)));
!        write (6,'("diag of norm pxa1 =",20(/10g13.5)))')
!     &       (dot_product(matmul(c(i,:m),py(:m,:m)),c(i,:m)),i=1,n)
        pxx(:n,:n) =matmul(matmul(c(:n,:m),py(:m,:m)),                  &
     &                                 transpose(c(:n,:m)));
        call prtmx (pxx,nd,n,n,"norm_pxa1");
        write (6,'("********************"/)');

        !** sum log likelihood
        delcEst =-(1.-pa(n)**2/r(n*(n+1)/2));!n*log(2.d0*pi) +2.d0*sum(log(pa(:n)))
        if (n > 3) then;
          delcEstT =delcEstT +delcEst;
          lnLike =delcEstT;
        else;
          lnLike =0.d0;
        endif;
!        lnLike =(n+m)*log(2.d0*pi) +m*log(1.d0) +2.d0*sum(log(pa(:n)))
        !** add contribution of a posteriori meas residuals
        do i=1,m;
          lnLike =lnLike +pyt(i,i);
        enddo;
        !** add effect of prior
        do i=1,n;
          lnLike =lnLike +pxx(i,i);
!     &         +dot_product(matmul(c(i,:m),py(:m,:m)),c(i,:m))
        enddo;

        i =0;
        do im =1,mmax2,50;
          i =i+1;
          sigy(i) =1.d0;
          if (n2 > 0) sigy(i) =sigy(i) +sum(hh(im,:n2)**2);
          if (im <= mmax) then;
            !** sigy =diag(R1 -hf1*px1*transpose(hf1)
            !**             +(hf2*pa1^(1/2) -hf1*a3)*transpose(hf2*pa1^(1/2) -hf1*a3) )
            sigy(i) =pyt(im,im);
          else;
            !** sigy =diag(R2 +hp1*px1*transpose(hp1)
            !**             +(hp2*pa1^(1/2) -hp1*a3)*transpose(hp2*pa1^(1/2) -hp1*a3) )
            sigy(i) =sigy(i) +dot_product(hp(im-mmax,:n),ha(im,:n));
          endif;
          sigy(i) =sqrt(sigy(i));
        enddo;
        write (10,'(i3,6g13.5)') n,sigy(:5),lnLike-m;

!        call symult (xe,  r,bb,xe,n)
!        dx(:n) =xe(:n) -x(:n)
!        yr(:) =y(:) -matmul(h(:,:n),xe(:n))
!        rsos =sum(yr(:m)**2)
!        write (6,'("tr(covy) =",g22.15,", rsos =",g13.5)') trc,rsos
      enddo;

      stop;
      end program;
!*******************************************************************
      subroutine get_haPoly (h,au,  mmax,nd,n,dt,t0);
!***     compute measurement partials (h) and least squares information matrix (au) for polynomial model
      implicit none;

      integer(4),intent(in) :: mmax,nd,n;
      real(8),intent(in) :: dt,t0;
      real(8),intent(out) :: h(mmax,nd),au(nd*(nd+1)/2);

      integer(4) it,j,k,kk;
      real(8) dtn,t,hnorm(mmax);
      !_________________________________________

      au(:) =0.d0;
      h(:,:) =0.d0;

      do it=1,mmax;
        t =(it-1)*dt +t0;
        dtn =1.d0;
        kk =0;
        do j=1,n;
          h(it,j) =dtn;
          dtn =dtn*t;
          do k=1,j;
            kk =kk+1;
            au(kk) =au(kk) +h(it,j)*h(it,k);
          enddo;
        enddo;
        hnorm(it) =sqrt(sum(h(it,1:n)**2));
      enddo;    !end time loop

      write (*,'("n =",i3,", hnorm =",10(/10g13.5))') n,hnorm(1:mmax);
      return;
      end subroutine get_haPoly;
!*******************************************************************
      subroutine get_haPolyC (h,al,  mmax,nd,n,dt,t0);
!***     compute measurement partials (h) and least squares information matrix (a1) for Chebyshey polynomial
      implicit none;

      integer(4),intent(in) :: mmax,nd,n;
      real(8),intent(in) :: dt,t0;
      real(8),intent(out) :: h(mmax,nd+1),al(nd*(nd+1)/2);

      integer(4) it,jt,j,k,kk;
      real(8) t,hnorm(mmax),c(n),cheb,dfdc(n),a,b;
      !_________________________________________

      al(:) =0.d0;
      h(:,:) =0.d0;
!      dt =1.d0/(mmax-1)
      a =0.d0;
      b =2.d0*(mmax-1)*dt;
      c(:) =0.d0;    !set

      do it=0,mmax-1;
        t =it*dt +t0;
        kk =0;
        jt =it+1;
!        jt =mmax+1-(it+1)    !it+1      !test re-order ###
        call chebyPoly (cheb,dfdc,  a,b,c,n,t);
!        write (6,'("dfdc =",10g13.5)') dfdc(:n)
        h(jt,:n) =dfdc(:);
        do j=1,n;
          do k=1,j;
            kk =kk+1;
            al(kk) =al(kk) +h(jt,j)*h(jt,k);
          enddo;
        enddo;
        hnorm(jt) =sqrt(sum(h(jt,1:n)**2));
      enddo;    !end time loop

      write (*,'("n =",i3,", hnorm =",10(/10g13.5))') n,hnorm(1:mmax);

      return;
      end subroutine get_haPolyC;
!******************************************
      subroutine chebyPoly (fx,dfdc,  x1,x2,c,m,x);
      !** evaluates Chebyshev polynomial at a specific point x, and also computes
      !** partial derivatives WRT the polynomial coefficients
      implicit none;
      integer(4),intent(in) :: m;   !order of model
      real(8),intent(in) :: x;      !intermediate point between x1 and x2 for evaluation
      real(8),intent(in) :: x1,x2;  !start and end points of interval
      real(8),intent(in) :: c(m);   !chebyshev coeff
      real(8),intent(out) :: fx;    !interpolated value at x
      real(8),intent(out) :: dfdc(m);!interpolated partials of fx wrt c

      integer(4) j;
      real(8) d,dd,sv,y,y2,dddc(m),ddddc(m),ddddcL(m);
      !__________________________________________

!      if ((x-x1)*(x-x2) > 0.) stop 'chebyPoly error: x not in range'
      d =0.d0;
      dd =0.d0;
      y =(2.d0*x-x1-x2)/(x2-x1);
      y2 =2.d0*y;
      dddc(:) =0.d0;
      ddddc(:) =0.d0;

      do j =m,2,-1;
        sv =d;
        ddddcL(:) =dddc(:);
        dddc(:) =y2*dddc(:) -ddddc(:);
        dddc(j) =dddc(j) +1.d0;
        d =y2*d -dd +c(j);
        dd =sv;
        ddddc(:) =ddddcL(:);
      enddo;

      fx =y*d -dd +0.5d0*c(1);
      dfdc(:) =y*dddc(:) -ddddc(:);
      dfdc(1) =dfdc(1) +0.5d0;

      return;
      end subroutine chebyPoly;
!*******************************************************************
      subroutine get_haPolyF (h,au,  mmax,nd,n,dt,t0);
      !** Fourier series
      implicit none;

      integer(4),intent(in) :: mmax,nd,n;
      real(8),intent(in) :: dt,t0;
      real(8),intent(out) :: h(mmax,nd+1),au(nd*(nd+1)/2);

      integer(4) it,i,j,k,kk,nf;
      real(8) t,hnorm(mmax),deltt,omeg;
      real(8) :: pi =3.141592653589793d0;
      !_________________________________________

      nf =n/2;
      deltt =(mmax-1)*dt;

      au(:) =0.d0;
      h(:,:) =0.d0;
!      dt =1.d0/(mmax-1)

      do it=1,mmax;
        t =(it-1)*dt +t0;
        omeg =2.d0*pi*t/deltt;
        h(it,1) =1.d0;
        h(it,2:) =0.d0;
        j =1;
        do i=2,n;
          j =i/2;
          if (mod(i,2) == 0) then;
            h(it,i) =sin(j*omeg);
          else;
            h(it,i) =cos(j*omeg);
          endif;
        enddo;
!        write (6,'("dfdc =",10g13.5)') h(it,:n)
        kk =0;
        do j=1,n;
          do k=1,j;
            kk =kk+1;
            au(kk) =au(kk) +h(it,j)*h(it,k);
          enddo;
        enddo;
        hnorm(it) =sqrt(sum(h(it,1:n)**2));
      enddo;    !end time loop

      write (*,'("n =",i3,", hnorm =",10(/10g13.5))') n,hnorm(1:mmax);

      return;
      end subroutine get_haPolyF;

