!** quad_cost.f generates a cost function plot file using a mildly quadratic function
!** Program quad_cost uses a mildly quadratic function to demonstrate the convergence
!** of Newton and Gauss-Newton iterations for finding the minimum of the cost function.
!** The program creates output plot file "quad.plt" for use in Example 7.3-1.

!**  Author B. Gibbs, 12/2009

!**************************************
! The author grants the user a non-exclusive, worldwide, royalty-free copyright license to

! 1. reproduce and modify the software for your own purposes, and
! 2. to distribute the software provided that you give credit to the author,
!    do not make claims against the author or Wiley-Interscience,
!    and mark any modifications as your own.  If the software is incorporated in
!    a commercial product, you agree to defend and indemnify this author and
!    Wiley-Interscience against any losses, damages and costs arising from claims,
!    lawsuits and other legal actions brought by a third party.

! The software is provided on an as is basis, without warranties or conditions
! of any kind, either express or implied (including any warranties or conditions
! of title, non-infringement, merchantability or fitness for a particular purpose).
! The user is solely responsible for determining the appropriateness of using and
! distributing the program and assumes all risks associated with its exercise of
! rights under this agreement.
!**************************************

!** sample compile and execute using Intel compiler:

!**     Include in ifort.cfg:
!**   /check:bounds /traceback /include:\kf_book\distribSWm\library /link /nodefaultlib:msvcrt.lib
!**     \kf_book\distribSWm\library\libkf_bnd_ifort.lib

!**     To compile and execute:
!**   set ifortcfg=\kf_book\distribSWm\ifort.cfg
!**   ifort quad_cost.f90
!**   quad_cost
!**************************************

      implicit none;
      integer(4) i,j;
      real(8) a,x,xt,y,dt,yc,cost,grad,hess;
      !___________________________________________


      open (10,file='quad.plt',status='unknown');
      write (10,3000);
 3000 format ('variables ="x", "cost"')
      write (10,3010) 41;
 3010 format ('ZONE T="cost", I=',i2,', F=POINT')

      dt =0.1;
      a =-0.04d0;
!      a =-0.2d0      !-0.1 takes 4 iter using Hessian, 1 for Fisher.  -0.2 does not converge using Hessian
      xt =1.0d0;
      y =xt+a*xt**2 -0.5;

      do i=-20,20;
        x =i*dt +xt;
        call funct (cost,grad,hess,  x,a,y);
        write (10,'(3g13.5)') x,cost;
      enddo;

      write (10,3020) 41;
 3020 format ('ZONE T="grad", I=',i2,', F=POINT')

      do i=-20,20;
        x =i*dt +xt;
        call funct (cost,grad,hess,  x,a,y);
        write (10,'(3g13.5)') x,grad;
      enddo;

      !** Gauss-Newton iterationa
      x =2.d0;
      call funct (cost,grad,hess,  x,a,y);
      write (10,3040);
 3040 format ('ZONE T="line1", I=5, F=POINT')
      write (10,'(3g13.5)') x,grad;

      do i=1,2;
        write (6,'("cost,grad,hess,dx =",4g13.5)')                      &
     &         cost,grad,hess, -grad/hess;
        x =x -grad/hess;
        call funct (cost,grad,hess,  x,a,y);
        write (10,'(3g13.5)') x,0.;
        write (10,'(3g13.5)') x,grad;
      enddo;

      !** x axis
      write (10,3050);
 3050 format ('ZONE T="origin", I=2, F=POINT')
      write (10,'(3g13.5)') -1., 0.;
      write (10,'(3g13.5)') 3., 0.;

      stop;
      end;
!*******************************************
      subroutine funct (cost,grad,hess,  x,a,y);
      implicit none;
      real(8),intent(in) :: x,a,y;
      real(8),intent(out) :: cost,grad,hess;

      real(8) yc;
      !______________________________________

      yc =x+a*x**2;
      cost =0.5d0*(y-yc)**2;
      grad =-(y-yc)*(1.d0+2.d0*a*x);
      hess =(1.d0+2.d0*a*x)**2;     !test fisher
      hess =(1.d0+2.d0*a*x)**2 -(y-yc)*2.d0*a;

      return;
      end subroutine funct;
