!** Program Tst_KF1 simulates measurements for a 1st-order Markov model,
!** and computes the a posteriori variance and filter estimates for a
!** 1-state Kalman filter.  Three different levels (0.1, 1.0, 10) of
!** process noise variance Q are tested, with measurement noise variance R =100.
!** The output plot file is "kf1.plt".  This is used for example 8.1-1.

!**************************************
! The author grants the user a non-exclusive, worldwide, royalty-free copyright license to

! 1. reproduce and modify the software for your own purposes, and
! 2. to distribute the software provided that you give credit to the author,
!    do not make claims against the author or Wiley-Interscience,
!    and mark any modifications as your own.  If the software is incorporated in
!    a commercial product, you agree to defend and indemnify this author and
!    Wiley-Interscience against any losses, damages and costs arising from claims,
!    lawsuits and other legal actions brought by a third party.

! The software is provided on an as is basis, without warranties or conditions
! of any kind, either express or implied (including any warranties or conditions
! of title, non-infringement, merchantability or fitness for a particular purpose).
! The user is solely responsible for determining the appropriateness of using and
! distributing the program and assumes all risks associated with its exercise of
! rights under this agreement.
!**************************************

!** sample compile and execute using Intel compiler:

!**     Include in ifort.cfg:
!**   /check:bounds /traceback /include:\kf_book\distribSWm\library /link /nodefaultlib:msvcrt.lib
!**     \kf_book\distribSWm\library\libkf_bnd_ifort.lib

!**     To compile and execute:
!**   set ifortcfg=\kf_book\distribSWm\ifort.cfg
!**   ifort tst_KF1.f90
!**   tst_KF1
!**************************************

      implicit none;
      integer(4) i,j,ix,icase;
      integer(4),parameter :: nplt=3, m=100;
      real(8) pa,pf,xf,xa,q,r,dt,tau,phi,xt,g,dy,a,b,d;
      real(8) fout(2,0:m,nplt),yn(m),qn(m);
      real(8) :: qi(nplt) = (/ 0.1, 1.0, 10. /);
      real(8),external :: rnor;
      !_______________________________

      r =100.d0;
      dt =1.d0;
      tau =10.d0;
!      phi =1.d0      !### random walk
      phi =exp(-dt/tau);   !=0.81873
      ix =19191919;
      xf =rnor(ix);
      !** generate meas noise
      do i=1,m;
        yn(i) =sqrt(r)*rnor(0);
        qn(i) =rnor(0);
      enddo;

      do icase =1,nplt;
        pf =10.*r;
        xf =0.d0;
        xt =0.d0;
        fout(:,0,icase) =(/ xf-xt, sqrt(pf) /);
        do i=1,m;
          xt =phi*xt +sqrt(qi(icase))*qn(i);
          xa =phi*xf;
          pa =phi**2*pf +qi(icase);
          dy =xt+yn(i) -xa;
          g =pa/(pa+r);
          xf =xa+g*dy;
          pf =pa*r/(pa+r); !pa-g*pa
          fout(:,i,icase) =(/ xf-xt, sqrt(pf) /);
        enddo;
      enddo;

      do icase=1,3;
        a =qi(icase) +r*(1.d0-phi**2);
        d =sqrt(a**2 +4.d0*phi**2*qi(icase)*r);
        b =sqrt(qi(icase)/r);
        write (6,'("for q =",g13.5,", pss =",g13.5,", alpha =",g13.5)') &
     &    qi(icase), (-a+d)/(2.d0*phi**2), 2.*phi*b/(b**2+(1.d0-phi**2));
      enddo;

      write (6,'("steady-state Kalman gains =",3g13.5)')                &
     &   fout(2,m,:)**2/(r+fout(2,m,:)**2);

      open (10,file='kf1.plt',status='unknown');
      write (10,1000);
1000  format('variables ="t", "xe-qp1", "sigx-qp1", "xe-q1", ',         &
     &   '"sigx-q1", "xe-q10", "sigx-q10"')
      do i=0,m;
        write (10,'(7g13.5)') i*dt,(fout(1:2,i,j),j=1,nplt);
      enddo;

      stop;
      end;
