!** Program Tst_KF2 simulates measurements for a 2-state system, and computes the
!** a posteriori variance and filter estimates for a 2-state Kalman filter.
!** Process noise drives only the second state of the continuous system, and three
!** different levels (0.1, 1.0, 10) of process noise PSD are tested.
!** State #1 is the integral of the second, and only this state is measured.
!** The measurement noise PSD =100.  The output plot file is kf2.plt.
!** This is used for example 8.1-2.

!**************************************
! The author grants the user a non-exclusive, worldwide, royalty-free copyright license to

! 1. reproduce and modify the software for your own purposes, and
! 2. to distribute the software provided that you give credit to the author,
!    do not make claims against the author or Wiley-Interscience,
!    and mark any modifications as your own.  If the software is incorporated in
!    a commercial product, you agree to defend and indemnify this author and
!    Wiley-Interscience against any losses, damages and costs arising from claims,
!    lawsuits and other legal actions brought by a third party.

! The software is provided on an as is basis, without warranties or conditions
! of any kind, either express or implied (including any warranties or conditions
! of title, non-infringement, merchantability or fitness for a particular purpose).
! The user is solely responsible for determining the appropriateness of using and
! distributing the program and assumes all risks associated with its exercise of
! rights under this agreement.
!**************************************

!** sample compile and execute using Intel compiler:

!**     Include in ifort.cfg:
!**   /check:bounds /traceback /include:\kf_book\distribSWm\library /link /nodefaultlib:msvcrt.lib
!**     \kf_book\distribSWm\library\libkf_bnd_ifort.lib

!**     To compile and execute:
!**   set ifortcfg=\kf_book\distribSWm\ifort.cfg
!**   ifort tst_KF2.f90
!**   tst_KF2
!**************************************

      implicit none;
      integer(4) i,j,ix,icase;
      integer(4),parameter :: nplt=3, m=250;
      real(8) pa(2,2),pf(2,2),xf(2),xa(2),q(2),phi(2,2),xt(2);
      real(8) r,dt,tau,g(2),dy,a(2,2),b(2),d(2),hp(2);
      real(8) fout(4,0:m,nplt),yn(m),qn(m);
      real(8) :: qi(nplt) = (/ 0.1, 1.0, 10. /);
      real(8),external :: rnor;
      !_______________________________

      dt =0.2d0;
!      dt =1.d0
      r =100.d0*1.d0/dt;
      tau =10.d0;
      phi(1,:) =(/ 1.d0, dt /);
      phi(2,:) =(/ 0.d0, 1.d0 /);   !=0.81873
!      phi(2,:) =(/ 0.d0, exp(-dt/tau) /)    !=0.81873
      ix =19191919;
      xf(1) =rnor(ix);
      !** generate meas noise
      do i=1,m;
        yn(i) =sqrt(r)*rnor(0);
        qn(i) =rnor(0);
      enddo;

      do icase =1,nplt;
        pf(:,:) =0.d0;
        pf(1,1) =10.*r*dt;
        pf(2,2) =10.*r*dt;
        xf(:) =0.d0;
        xt(:) =0.d0;
        fout(:2,0,icase) =xf(:)-xt(:);
        fout(3:,0,icase) =(/ sqrt(pf(1,1)), sqrt(pf(2,2)) /);
        do i=1,m;
          xt =matmul(phi,xt);
          xt(2) =xt(2) +sqrt(qi(icase))*qn(i);
          xa =matmul(phi,xf);
          !** p =phi*p*phi^T
          a(:,:) =matmul(phi,pf);
          pa(:,:) =matmul(a,transpose(phi));
          !** add Q
          pa(1,1) =pa(1,1) +qi(icase)*dt**3/3.d0;
          pa(1,2) =pa(1,2) +qi(icase)*dt**2/2.d0;
          pa(2,1) =pa(1,2);   !pa(2,1) =pa(2,1) +qi(icase)*dt**2/2.d0
          pa(2,2) =pa(2,2) +qi(icase)*dt;
          !** meas update
          dy =xt(1)+yn(i) -xa(1);
          hp(:) =pa(1,:);
          g(:) =hp(:)/(hp(1)+r);
          xf(:) =xa(:) +g(:)*dy;
          pf(1,:) =pa(1,:) -g(1)*hp(:);
!          pf(2,:) =pa(2,:) -g(2)*hp(:)    !###
          pf(2,:) =(/ pf(1,2), pa(2,2)-g(2)*hp(2) /);
!          pf =pa*r/(pa+r)  !pa-g*pa
          fout(:2,i,icase) =xf(:)-xt(:);
          fout(3:,i,icase) =(/ sqrt(pf(1,1)), sqrt(pf(2,2)) /);
        enddo;
      enddo;

!      write (6,'("steady-state Kalman gains =",3g13.5)')
!     &   fout(2,m,:)**2/(r+fout(2,m,:)**2)

      open (10,file='kf2.plt',status='unknown');
      write (10,1000);
1000  format('variables ="t", '                                         &
     &   '"xe1-qp1", "xe2-qp1", "sigx1-qp1", "sigx2-qp1", ',            &
     &   '"xe1-q1",  "xe2-q1",  "sigx1-q1",  "sigx2-q1", ',             &
     &   '"xe1-q10", "xe2-q10", "sigx1-q10", "sigx2-q10"')
      do i=0,m;
        write (10,'(13g13.5)') i*dt,(fout(1:4,i,j),j=1,nplt);
      enddo;

      stop;
      end;
