!** Program Tst_INS compares results from the Friedland
!** bias-free/bias-restoring filter described in Section 8.4-1
!** with the standard Kalman filter.  This uses file "TAImes.dat"
!** generated by program INS_sim.

!**************************************
! The author grants the user a non-exclusive, worldwide, royalty-free copyright license to

! 1. reproduce and modify the software for your own purposes, and
! 2. to distribute the software provided that you give credit to the author,
!    do not make claims against the author or Wiley-Interscience,
!    and mark any modifications as your own.  If the software is incorporated in
!    a commercial product, you agree to defend and indemnify this author and
!    Wiley-Interscience against any losses, damages and costs arising from claims,
!    lawsuits and other legal actions brought by a third party.

! The software is provided on an as is basis, without warranties or conditions
! of any kind, either express or implied (including any warranties or conditions
! of title, non-infringement, merchantability or fitness for a particular purpose).
! The user is solely responsible for determining the appropriateness of using and
! distributing the program and assumes all risks associated with its exercise of
! rights under this agreement.
!**************************************

!** sample compile and execute using Intel compiler:

!**     Include in ifort.cfg:
!**   /check:bounds /traceback /include:\kf_book\distribSWm\library /link /nodefaultlib:msvcrt.lib
!**     \kf_book\distribSWm\library\libkf_bnd_ifort.lib

!**     To compile and execute:
!**   set ifortcfg=\kf_book\distribSWm\ifort.cfg
!**   ifort tst_ins.f90
!**   tst_ins >tst_ins.lis
!**************************************

      module modelDat;
      implicit none;
      integer(4),parameter :: ndim=27, n2dim=(ndim*(ndim+1))/2, mdim=3;

      real(8) phi(ndim,ndim),q(n2dim),h(mdim,ndim),sigr(mdim),y(mdim);
      real(8) x(ndim),sig0(ndim),q1(45);
      real(8) xt(9),at(3),f(3,ndim);

      end module modelDat;
!*****************************************************
      program tst_INS;
      use modelDat;
      implicit real*8 (a-h,o-z);
      integer(4),parameter :: maxjtim=90,maxjmp=20,nparm=1230,maxbias=30;
      character*8 gbname(maxbias),gjname(maxjmp);

      integer(4) mapxj(ndim);

      data tmax / 800.d0/;
      data mapxj /ndim*1/;
      data resprt /0/;
      data sigfac /0.5/;
      data gbname /'ab1','ab2','ab3','sfe1','sfe2','sfe3',              &
     &     'bda','bdb','bdc','sfeag','sfebg','sfecg',                   &
     &     'adas','adbs','adci','aadass','aadbss','aadcss',             &
     &     12*'    '/;
      data gjname /'ab1','ab2','ab3','sfe1','sfe2','sfe3',              &
     &     'bda','bdb','bdc','sfeag','sfebg','sfecg',                   &
     &       'adas','adbs','adci', 5*'    '/;
      !_____________________________________________________________

      nbias=0;
!      if (lbias > 0) nbias=18

!           input file of measurement data
      open(30,file='taimes.dat',status='old', form='unformatted');

      phi(:,:) =0.d0;
      q(:) =0.d0;
      t =0.d0;
      nmeas =0;

!         read header of initial standard deviations
      read(30) sig0;
      write (6,'("sigx0 =",5(/10g13.5))') sig0(:ndim);
      write (6,'(/"start time loop")');

!         time loop for reading and filtering measurement data

!****     time loop for reading and filtering measurement data
      do while (t < tmax);
        read(30,end=999) t,nt,m,nq,y(1:m),sigr(1:m);
!        print *,t,nt,m
        n =nt-nbias;    !set for bias case
        nx =n;
        nbias =nt-nx;
        nmeas =nmeas+m;
        if (nt > ndim .or. m > mdim) stop 'invalid nt or m';

        read(30) xt(:),at(:),(f(1:3,j),j=7,ndim);  !at is accel error, not total

        h(:,:) =0.d0;
        do i=1,m;
          read(30) j1,j2,(h(i,j),j=j1,j2);
          if (j1 < 1 .or. j1 > nt .or. j2 < 1 .or. j2 > nt) stop 40;
        enddo;

        q(:) =0.d0;
        do i=1,nq;
          read(30) l,q(l);
          if (l < 1 .or. l > n2dim) stop 43;
        enddo;

        phi(:,:) =0.d0;
        do i=1,nt;
          read(30) j1,j2,(phi(i,j),j=j1,j2);
          if (j1 < 1 .or. j1 > nt .or. j2 < 1 .or. j2 > nt) stop 41;
        enddo;

!        write(6,*) 't =',t
!        call prtmx(y,1,1,m,'y       ')
!        call prtmx(r,1,1,m,'rsig    ')
!        call prtmx(h,mdim,m,n,'h       ')
!        call prtmx(phi,ndim,n,n,'phi     ')
!        call prtsmx(q,n,'q       ')

        !** standard filter
        call filterBias (t,phi,q,h,sigr,y,sig0,n,nt,nx,m,ndim,mdim,rsum);
!        call filter (t,phi,q,h,sigr,y,sig0,n,nt,nx,m,ndim,mdim,rsum)
      enddo;   !___  end of filter loop

  999 write(6,'(" rsum =",g12.5,", nmeas =",i4)')  rsum,nmeas;
      nimub=0;

      stop;
      end program;
! *****************************************************************
      subroutine filter (t,phi,q,h,r,y,sig0,n,nt,nx,m,ndim,mdim,  rsum);
!         kalman filter subroutine with auxilliary eqns to compute
!         sensitivity to jump states

      implicit none;
      integer(4),intent(in) :: n,nt,nx,m,ndim,mdim;
      real(8),intent(in) :: t,sig0(n),phi(ndim,nt),q(n*(n+1)/2);
      real(8),intent(in) :: h(mdim,nt),r(m),y(m);
      real(8),intent(out) :: rsum;

      integer(4) i,j,k,l,im;
      integer(4),save :: ifirst =1;
      real*4 temp;
      real(8) sum,c,yp,dz;
      integer(4),parameter :: nn=27, n2dim=(nn*(nn+1))/2, mm=3;

      real(8),save :: p(n2dim),x(nn);
      real(8),save :: d1(nn),d2(nn,nn),d3(nn),hp(nn),g(nn);
      !_____________________________________

      if (ifirst.gt.0) then;
        ifirst=0;
        rsum=0.;
        x(:) =0.d0;
        p(:) =0.d0;
        !** initialize p using sig0
        l=0;
        do i=1,n;
          l=l+i;
          p(l)=sig0(i)**2;
        enddo;
      endif;

!         time update
      call smult (phi,x,d1,n,n,1,ndim,n,n);
      x(:n) =d1(:n);

!         p=phi*p*phi**t+q
      do i=1,n;
        d1(:n) =phi(i,:n);
        call symult (d2(1,i),  p,d1,n);
      enddo;

      l=0;
      do i=1,n;
        do j=1,i;
          l=l+1;
          p(l) =q(l) +dot_product(phi(i,:n),d2(:n,j));
        enddo;
      enddo;

!          propagate state sensitivity to jump parameters
!          add new parameters at current time
!      call jumptime (phi,t,nt,nx,m,ndim)    !###

!         measurement update (scalar)
      do im=1,m;
        yp=0.;
        d1(:nt) =h(im,:nt);
        yp =dot_product(h(im,:n),x(:n));
        dz =y(im)-yp;

        call symult (hp,  p,d1,n);
        c =dot_product(hp(:n),h(im,:n)) +r(im)**2;
        rsum =rsum+dz**2/c;
!        write(6,'('' t, rsum = '',f10.2,g12.5)') t, rsum

        l=0;
        do i=1,n;
          g(i) =hp(i)/c;
          x(i) =x(i)+g(i)*dz;
          do j=1,i;
            l=l+1;
            p(l) =p(l)-g(i)*hp(j);
          enddo;
        enddo;

!           update jump sensitivities
!        call jumpmes (d1,g,c,dz,im,nt,nx)  !###
      enddo;

      write (6,1000) t,x(:n);
 1000 format (50g13.5)

      return;
      end subroutine filter;
! *****************************************************************
      subroutine filterBias (t,phi,q,h,r,y,sig0,n,nt,nx,m,ndim,mdim,    &
     &                       rsum);
!         kalman filter using Friedland's bias-free/bias-restoring filter

      implicit none;
      integer(4),intent(in) :: n,nt,nx,m,ndim,mdim;
      real(8),intent(in) :: t,sig0(n),phi(ndim,nt),q(n*(n+1)/2);
      real(8),intent(in) :: h(mdim,nt),r(m),y(m);
      real(8),intent(out) :: rsum;

      integer(4) i,j,k,l,im;
      integer(4),save :: ifirst =1;
      real*4 temp;
      real(8) sum,c,cb,yp,dz,dzb;
      integer(4),parameter :: nd=9,nb =18,mm=3;

      real(8),save :: pd(nd,nd),xd(nd),hd(nd),hpd(nd),gd(nd);
      real(8),save :: pb(nb,nb),xb(nb),hb(nb),hpb(nb),gb(nb),tt(nb);
      real(8),save :: s(nd,nb),d2(nd,nd),x(nd+nb);
      !_____________________________________

      if (nd+nb /= n) stop 'filterBias error: nd +nb /= n';

      if (ifirst > 0) then;
        ifirst= 0;
        rsum=0.;
        xd(:) =0.d0;
        pd(:,:) =0.d0;
        xb(:) =0.d0;
        pb(:,:) =0.d0;
        s(:,:) =0.d0;
        !** initialize p using sig0
        forall (i=1:nd) pd(i,i) =sig0(i)**2;
        forall (i=1:nb) pb(i,i) =sig0(i+nd)**2;
      endif;

!         time update p=phi*p*phi**t+q
      d2(:,:) =matmul(phi(:nd,:nd),pd(:,:));
      xd(:) =matmul(phi(:nd,:nd),xd(:));
      s(:,:) =matmul(phi(:nd,:nd),s(:,:)) +phi(:nd,nd+1:nd+nb);

      l =0;
      do i=1,nd;
        do j=1,i;
          l =l+1;
          pd(i,j) =q(l) +dot_product(d2(i,:nd),phi(j,:nd));
          pd(j,i) =pd(i,j);
        enddo;
      enddo;
!      call prtmx (pd,nd,nd,nd,"pb-")

!         measurement update (scalar)
      do im=1,m;
        hd(:) =h(im,:nd);
        hb(:) =h(im,nd+1:nd+nb);
        yp =dot_product(hd(:),xd(:));
        dz =y(im)-yp;
        hpd(:) =matmul(hd(:),pd(:,:));
        c =dot_product(hpd(:),hd(:)) +r(im)**2;
!        rsum =rsum +dz**2/c
!        write(6,'(" t, rsum = ",f10.2,g12.5)') t, rsum
        gd(:) =hpd(:)/c;
        xd(:) =xd(:) +gd(:)*dz;
        forall (i=1:nd) pd(i,:) =pd(i,:) -gd(i)*hpd(:);

        do i=1,nb;
          tt(i) =dot_product(hd(:),s(:,i)) +hb(i);
          s(:,i) =s(:,i) -gd(:)*tt(i);
        enddo;

        hpb(:) =matmul(tt(:),pb(:,:));
        cb =dot_product(hpb(:),tt(:)) +c;
        gb(:) =hpb(:)/cb;

        dzb =dz -dot_product(tt,xb);
        rsum =rsum +dzb**2/cb;
        xb(:) =xb(:) +gb(:)*dzb;
        forall (i=1:nb) pb(i,:) =pb(i,:) -gb(i)*hpb(:);
!        write (6,1000) t,xd(:nd),xb(:nb)

        x(:nd) =xd(:nd) +matmul(s(:,:nb),xb(:nb));
        x(nd+1:nd+nb) =xb(:);
      enddo;

      write (6,1000) t,x(:n);
 1000 format (50g13.5)

      return;
      end subroutine filterBias;

