!** ACTrkFilSmo.f is a nonlinear filter-smoother that tracks an aircraft

!**************************************
! The author grants the user a non-exclusive, worldwide, royalty-free copyright license to

! 1. reproduce and modify the software for your own purposes, and
! 2. to distribute the software provided that you give credit to the author,
!    do not make claims against the author or Wiley-Interscience,
!    and mark any modifications as your own.  If the software is incorporated in
!    a commercial product, you agree to defend and indemnify this author and
!    Wiley-Interscience against any losses, damages and costs arising from claims,
!    lawsuits and other legal actions brought by a third party.

! The software is provided on an as is basis, without warranties or conditions
! of any kind, either express or implied (including any warranties or conditions
! of title, non-infringement, merchantability or fitness for a particular purpose).
! The user is solely responsible for determining the appropriateness of using and
! distributing the program and assumes all risks associated with its exercise of
! rights under this agreement.
!**************************************

!** sample compile and execute using Intel compiler:

!**     Include in ifort.cfg:
!**   /check:bounds /traceback /include:\kf_book\distribSWm\library /link /nodefaultlib:msvcrt.lib
!**     \kf_book\distribSWm\library\libkf_bnd_ifort.lib

!**     To compile and execute:
!**   set ifortcfg=\kf_book\distribSWm\ifort.cfg
!**   ifort acTrkFilSmo.f90       (libkf_bnd_ifort.lib should be linked by default)
!**   acTrkFilSmo >actrk.lis
!**************************************

      module modelDat;
      !** includes model info for funct
      integer(4),parameter :: nState =8,ndynam=8;
      integer(4),parameter :: mmax =3;
      integer(4) irecs;    !direct access record number for smoothing
      integer(4) nestF,nF3,nDynamEst;
      real(8) xt(nState),xtl(nState),xo(3);
      real(8) xp(nState),xa(nState),pa(nState,nState),pad(nState);
      real(8) xf(nState),pf(nState,nState),phi(nState,nState);
      real(8) q(nState,nState),qs(nState),tau(nstate);
      real(8) xref(nState);
      real(8) rn,rnoise(3);
      real(8) :: gacc = 9.80d0;    !m/s^2
      logical(4) cartesian, debugP, markov;
      logical(4) :: firstIter =.false.;     !normal mode linearizing all iter about estimate ###
!      logical(4) :: firstIter =.true.      !test linearizing partials about meas for 1st iter ###

      integer(4) nMnvr;         !number of maneuvers
      type mnvrDats;
        real(8) tstart,tend;
        real(8) tanphi,accel,zdot;
      end type;
      type (mnvrDats) mnvrDat(10);

      end module modelDat;
!****************************************************
      program ACTrkFil;
      !** simulation of aircraft tracking with radar RAE tracking every 1 second
      use modelDat;
      implicit none;
      integer(4) i,n,m,ix,nrec,imnvr,lenrec,ierr;
      integer(4) :: method =1;  !1=EKF, 2=IEKF   !little smoothed difference between 1 and 2 ###
      logical(4) IntegCov;
      real(8) r,y(3),yp(3),h(3,nState),dt;
      real(8) tend,tstart,tstop,tanphi,accel,zdot,t,tLst;
      real(8),external :: rnor;
      !____________________________________________________

      m =mmax;
      n =nState;
      nestF =nState;
      nF3 =(nestF*(nestF+3))/2;
      nDynamEst =nestF;    !(no biases in this problem)
      ix =191919;
      r =rnor(ix);      !initialize randum number generator
      rnoise(:) =(/ 1., 0.2e-3, 0.2e-3 /);  !RAE noise

      !** maneuvers
      open (10,file="mnvs.dat",status='old');
      nmnvr =1;
      do;
        read (10,*,end=100,err=100) tstart,tstop,tanphi,accel,zdot;
        nmnvr =nmnvr +1;
        mnvrDat(nmnvr)%tstart =tstart;
        mnvrDat(nmnvr)%tend =tstop;
        mnvrDat(nmnvr)%tanphi =tanphi;
        mnvrDat(nmnvr)%accel =accel;
        mnvrDat(nmnvr)%zdot =zdot;
      enddo;

!   45., 56., -0.5, 0., 0.
!   75., 87., -0.5, 0., -4.
!   90., 95., 0., -1., -6.
!  125., 510., 0., 0., 0.

  100 close(10);
      dt =1.d0;  !       !set
      tend =mnvrDat(nmnvr)%tstart -1.;
      cartesian =.true.;

      !** true state, prior and observer position
      xt(:) =(/ 1200.d0, 0.d0, 300.d0, 0.d0, 32.d0, 0.d0, 0.d0, 0.d0 /);
      xo(:) =0.d0;          !own position

      xp(:) =(/ 1100.d0, 10.d0, 330.d0,  20.d0, 0.d0, 0.d0, 0.d0, 0.d0/);   !prior estimate
      pad(:) =(/ 100.d0, 100.d0, 100.d0, 20.d0, 20.d0, 0.6d0,           &
     &           0.50d0, 0.63d0 /); !prior 1-sigma

      markov =.false.;  !true.
      tau(6:8) =(/ 5., 10., 10. /);
      pf(:,:) =0.d0;
      do i=1,nState;
        pf(i,i) =pad(i)**2;
      enddo;
      qs(:3) = 0.d0;
      qs(4:5) =0.001;
      qs(6) =0.08;
!      qs(7:8) =(/ 0.04, 0.4 /)
      qs(7:8) =(/ 0.01, 0.4 /);
      if (markov) forall (i=6:8) qs(i) =2.d0*pad(i)**2/tau(i);

      xf(:) =xp(:);     !initialize estimate with prior
!      xf(:) =xt(:)      !###  test initialized at truth
      xref(:) =0.d0;    !reference state is 0

      debugp =.false.;   !###

      open (11,file='ACTrk.plt',status='unknown');
      write (11,1200);
 1200 format('VARIABLES ="t (s)", "x (m)", "y (m)", "z (m)", ',         &
     &   '"xdot (m/s)", "ydot (m/s)", "zdot (/s)", ',                   &
     &   '"tanphi", "a-acc (m/s2"')

      open (20,file='Filt_Xerr.plt',status='unknown');
      write (20,1100);
 1100 format('VARIABLES ="t", "x-error", "y-error", "z-error", ',       &
     &   '"xdot-error", "ydot-error", "zdot-error", "tanphi-error", ',  &
     &   '"accel-a-error"')
      open (21,file='Filt_sig.plt',status='unknown');
      write (21,1110);
 1110 format('VARIABLES ="t", "sig-x", "sig-y", "sig-z", "sig-xd", ',   &
     &   '"sig-yd", "sig-zd", "sig-tanphi", "sig-acca"')

      !** smoothRTS
      lenrec =2*(n*(3+3*n)+2);  !record length in 4-byte words
      open (42,form='unformatted',access='direct',recl=lenrec);
      irecs =0;

      !** scratch direct access for ephemeris
      open (43,form='unformatted',access='direct',recl=2+2*n);

      !** sequential smoothed track file (to be read in next run on unit 21)
      open (44,file='smoothephem.dat',form='binary',status='unknown',   &
     &        iostat=ierr);
      if (ierr /= 0)                                                    &
     &        call errhand('error opening output track file');
      write (44) n;

      !** output measurement file (for program filter_smoother)
      open (23,file='acMeas.txt',status='unknown',iostat=ierr);
      if (ierr /= 0)                                                    &
     &        call errhand('error opening measurement output file');

      t =0.d0;
      tLst =t;
      zdot =0.d0;
      imnvr =1;

      do while (t < tend);
        t =t+dt;
        tanphi =0.d0;
        accel =0.d0;
        xtl(:) =xt(:);

        if (t >= mnvrDat(imnvr)%tend) imnvr =imnvr+1;
        if (t >= mnvrDat(imnvr)%tstart) then;
          xt(6) =mnvrDat(imnvr)%zdot;
          tanphi =mnvrDat(imnvr)%tanphi;
          accel =mnvrDat(imnvr)%accel;
        endif;
        xt(7) =tanphi;
        xt(8) =accel;

        !** propagate state over dt and compute phi,q
        call integx (xt,  xt,dt,n);
        write (11,'(9g13.5)') t,xt(:);

        !** meas (r,ax,el)
        call get_yh (yp,h,  xt,xo,n);
        do i=1,3;
          y(i) =yp(i) +rnoise(i)*rnor(0);
        enddo;
        write (23,'(g15.7,i5,6g15.7)') t,1,y(:)
        call ekf (t,dt,y,rnoise,qs,method);     !extended Kalman filter or iterated EKF
!        call iekfs (t,dt,y,rnoise,qs,method)   !iterated extended Kalman filter-smoother
        tLst =t;
      enddo;

      !***   execute RTS smoother
      IntegCov =.true.;         !compute smoothed covariance
      nrec =irecs;
      call RTSsmoother (t,nrec,integCov);

      stop;
      end program;
!******************************************************************
      subroutine ekf (t,dt,y,rnoise,qs,method);
!**     extended Kalman filter with output for RTS smoothing
      use modelDat, only:xref,xt,xtl,xo,xa,pa,xf,pf,phi,q,mmax,         &
     &                    firstIter,irecs,nestF,debugP;
      implicit none;

      integer(4),intent(in) :: method;
      real(8),intent(in) :: t,dt,y(mmax),rnoise(mmax);
      real(8),intent(in) :: qs(nestF);

      integer(4) j,i,k,n,im;
      real(8) h(nestF),hp(nestF),g(nestF),xfi(nestF),xfil(nestF);
      real(8) h3(mmax,nestF),yp(mmax),dy;
      !_________________________________________

      n =nestF;

      if (dt > 0.d0) then;
        call get_phix (xa,phi,q,   xf,dt,qs,n);
        pa(:,:) =matmul(matmul(phi,pf),transpose(phi)) +q(:,:);
      else;
        xa(:) =xf(:);
        pa(:,:) =pf(:,:);
      endif;

      irecs =irecs+1;
      xref(:) =xtl(:);
      write (42,rec=irecs) t,t-dt,xref,xa,xf,pf,phi,q;

      !** now measurement update one-component-at-a-time
      do im=1,mmax;
        call get_yh (yp,h3,  xa,xo,n);
        h(:) =h3(im,:);
        hp(:) =matmul(h,pa);
        g(:) =hp(:)/(dot_product(h,hp)+rnoise(im)**2);
        dy =y(im)-yp(im);     !/rn
        xf(:) =xa(:) +g(:)*dy;

        if (method == 2) then;
          !** iterated EKF
          xfi(:) =xf(:);
          xfil(:) =1.d20;
          do k=1,100;
            call get_yh (yp,h3,  xf,xo,n);
            h(:) =h3(im,:);
            hp(:) =matmul(h,pa);
            g(:) =hp(:)/(dot_product(h,hp) +rnoise(im)**2);
            dy =y(im) -yp(im) -dot_product(h,xa(:)-xfi(:));     !/rn
            xf(:) =xa(:) +g(:)*dy;
            if (sqrt(sum((xfil(:n) -xfi(:n))**2)) < 0.001) exit;
            xfil(:) =xfi(:);
!            write (6,'("iekf x =",g13.5)') xfi
          enddo;
          xf(:) =xfi(:);
        endif;

        !** covariance update and make symmetric
        forall (i=1:n) pf(:,i) =pa(:,i) -g(:)*hp(i);
        do i=2,n;
          do j=1,n-1;
            pf(j,i) =0.5d0*(pf(j,i)+pf(i,j));
            pf(i,j) =pf(j,i);
          enddo;
        enddo;

        !** reset prior for next meas
        xa(:) =xf(:);
        pa(:,:) =pf(:,:);
      enddo;   !___ end measurement loop

      write (20,'(9g13.5)') t,xf(:)-xt(:);
      write (21,'(9g13.5)') t,(sqrt(pf(i,i)),i=1,n);

      return;
      end subroutine ekf;
!******************************************************************
      subroutine iekfs (t,dt,y,rnoise,qs,method);
      !*** iterated extended filter-smoother
      use modelDat, only:xref,xt,xtl,xo,xa,pa,xf,pf,phi,q,mmax,         &
     &                    firstIter,irecs,nestF,debugP;
      implicit none;

      integer(4),intent(in) :: method;
      real(8),intent(in) :: t,dt,y(mmax),rnoise(mmax);
      real(8),intent(in) :: qs(nestF);

      integer(4) j,i,k,n,iter,ier,itmax;
      real(8) hp(mmax,nestF),g(nestF,mmax),eps,epsm;
      real(8) h3(mmax,nestF),yp(mmax),dy(mmax),pas(nEstF*(nEstF+1)/2);
      real(8) x0smo(nestF),x1mod(nestF),x1modl(nestF),pai(nestf,nestf);
      real(8) pz(mmax*(mmax+1)/2),s(nestF,nestF);
      real(8) t1(nestF,nestF),t2(nEstF),t3(nestF,mmax);
      !_________________________________________

      n =nestF;
      x0smo(:) =xf(:);
      x1modl(:) =1.d20;
      itmax =100;
      if (t < 5) itmax =1;    !###

      do iter=1,itmax;
        if (dt > 0.d0) then;
          call get_phix (xa,phi,q,   x0smo,dt,qs,n);
          t1(:n,:n) =matmul(phi(:n,:n),pf(:n,:n));
          pa(:n,:n) =matmul(t1(:n,:n),transpose(phi(:n,:n))) +q(:n,:n);
          xa(:) =xa(:) +matmul(phi(:n,:n),(xf(:n)-x0smo(:n)));
        else;
          xa(:) =xf(:);
          pa(:,:) =pf(:,:);
        endif;
        if (iter == 1) then;
          x1mod(:) =xa(:);
          write (6,'("iter ",i3," x1mod =",8g13.5)') iter,x1mod(:);
        endif;

        if (method == 1) then;
          !** t1 =pa^(-1)*phi*pf,  s =transpose(t1)
          pai(:n,:n) =pa(:n,:n);
          call matInvGJ (ier,  pai,t1,  n,n,n);  !was gaussj (ier,pai,n,n,t1,n,n)
          if (ier > 0) stop 'inversion error on pai';
          s(:,:) =transpose(t1(:,:));
          write (6,'("method 1")');

        else;
          !** s =transpose [pa^(-1)*phi*pf]
          !** create symmetric version of pa
          k =0;
          do i=1,n;
            do j=1,i;
              k=k+1;
              pas(k) =pa(j,i);
            enddo;
          enddo;

          !** invert pa
          eps =1.d-4;
          call sinv (ier,epsm,  pas,  n,eps,.true.);
          if (ier /= 0) stop 'iekfs error: sinv pa failure';

          !** s =transpose[pa^(-1)*t1]
          do i=1,n;
            t2(:) =t1(:,i);
            call symult (s(i,:),  pas,t2(:n), n);
          enddo;
          write (6,'("method 2")');
        endif;  !___ end calculation of s
        call prtmx (s,n,n,n,"s");

        call get_yh (yp,h3,  x1mod,xo,n);
        hp(:,:) =matmul(h3,pa);
        k =0;
        do i=1,mmax;
          do j=1,i;
            k=k+1;
            pz(k) =dot_product(hp(j,:n),h3(i,:n));
          enddo;
          pz(k) =pz(k) +rnoise(i)**2;
        enddo;

        eps =1.d-4;
        call sinv (ier,epsm,  pz,  mmax,eps,.true.);
        if (ier /= 0) stop 'ekfs error: sinv pz failure';
        do i=1,n;
          call symult (g(i,:mmax),  pz,hp(:mmax,i),mmax);
        enddo;

        dy(:) =y(:) -yp(:) -matmul(h3(:mmax,:n),(xa(:n)-x1mod(:n)));     !/rn
        x1mod(:n) =xa(:n) +matmul(g(:n,:mmax),dy(:mmax));
        x0smo(:n) =xf(:n) +matmul(s(:n,:n),x1mod(:n)-xa(:n));
        if (itmax > 1 .and. iter == itmax)                              &
     &         stop 'ekfs error: max iterations';
        if (sum((x1mod(:)-x1modl(:))**2) < 0.001**2) exit;
        x1modl(:) =x1mod(:);
        write (6,'("t =",g13.5,", iter ",i3," x1mod =",8g13.5)')        &
     &        t,iter,x1mod(:);
      enddo;   !___ end iteration loop

      irecs =irecs+1;
      xref(:) =xtl(:);
      write (42,rec=irecs) t,t-dt,xref,xa,xf,pf,phi,q;

      xf(:) =x1mod(:);
      pa(:,:) =pa(:,:) -matmul(g,hp);
      t3(:n,:mmax) =matmul(pa,transpose(h3));
      forall (i=1:mmax) t3(:,i) =-t3(:,i) +g(:,i)*rnoise(i)**2;
      pf(:,:) =pa(:,:) +matmul(t3,transpose(g));
      do i=2,n;
        do j=1,n-1;
          pf(j,i) =0.5d0*(pf(j,i)+pf(i,j));
          pf(i,j) =pf(j,i);
        enddo;
      enddo;

      write (20,'(9g13.5)') t,xf(:)-xt(:);
      write (21,'(9g13.5)') t,(sqrt(pf(i,i)),i=1,n);

      return;
      end subroutine iekfs;
!*********************************************************
      subroutine get_yh (yp,h,  xt,xo,n);
!***    computes measurements range, az, el and partials wrt state
      use modelDat, only:nState;
      implicit none;
      integer(4),intent(in) :: n;       !state dimension
      real(8),intent(in) :: xo(3);      !own cartesian position
      real(8),intent(in) :: xt(8);      !target cartesian position, velocity, tan phi, long accel
      real(8),intent(out) :: yp(3);      !range, az, el
      real(8),intent(out) :: h(3,n);     !partials of y wrt xt

      real(8) dx(3),r,r2,az,el,rh,rh2;
      !_______________________________________

      dx(:) =xt(:3) -xo(:);
      r2 =sum(dx(1:3)**2);
      r =sqrt(r2);
      rh2 =dx(1)**2 +dx(2)**2;
      rh =sqrt(rh2);
      az =atan2(dx(1),dx(2));
      el =asin(dx(3)/r);
      yp(:) =(/ r,az,el/);

      h(:,:) =0.d0;
      h(1,:3) =(/ dx(1), dx(2), dx(3) /)/r;
      h(2,:3) =(/ dx(2), -dx(1), 0.d0 /)/rh2;
      h(3,:3) =(/ -dx(1)*dx(3), -dx(2)*dx(3), rh2 /)/(rh*r2);

      return;
      end subroutine get_yh;
!*********************************************************
      subroutine get_phix (xa,phi,q,   xf,dt,qs,n);
!***    computes state transition and process noise matrix and integrates state
      use modelDat, only:nState,ndynam,gacc,debugP;
      implicit none;
      integer(4),intent(in) :: n;       !state dimension
      real(8),intent(in) :: xf(n);     !target state
      real(8),intent(in) :: dt;         !time step (s)
      real(8),intent(in) :: qs(n);      !state noise psd
      real(8),intent(out) :: xa(n);     !predicted state at t+dt
      real(8),intent(out) :: phi(nState,n); !state transition matrix
      real(8),intent(out) :: q(nState,n);   !state noise matrix

      integer(4) i;
      real(8) xdot(nState),f(n,n),phii(n,n),dti,t;
!      real(8) :: dtmax =1.1  !###
      real(8) :: dtmax =0.2;     !little difference between 0.1 and 0.2
      !_______________________________________

      if (n /= nState) stop 'get_phix error: n /= nState';

      t =0.d0;
      if (dt <= 0.) stop 'get_phix error: dt <= 0';
      xa(:) =xf(:);
      phi(:,:) =0.d0;
      forall (i=1:n) phi(i,i) =1.d0;
      q(:,:) =0.d0;

      do while (t < dt);
        dti =min(dtmax,dt-t);
        t =t+dti;
        call get_F (f,xdot,   xa,n);
        phii(:,:) =dti*f(:,:);
!        phii(:,:) =dti*(f(:,:) +0.5*dti*matmul(f,f))        !###
        forall (i=1:n) phii(i,i) =phii(i,i) +1.d0;
        phi(:,:) =matmul(phii(:,:),phi(:,:));
!        call prtmx (phi,n,n,n,"phii")
        q(:,:) =matmul(matmul(phii,q),transpose(phii));
        forall (i=1:n) q(i,i) =q(i,i) +qs(i)*dti;
        xa(:) =xa(:) +dti*xdot(:);
      enddo;

      if (debugP) then;
        write (6,'("get_phix t =",g13.5)') t;
        call prtmx (phi,n,n,n,"phi");
        call prtmx (q,n,n,n,"q");
      endif;

      return;
      end subroutine get_phix;
!*********************************************************
      subroutine get_phix2 (xa,phi,q,   xf,dt,qs,n);
!***    computes state transition and process noise matrix and integrates state
!***    (little difference between get_phiq and get_phiq2)
      use modelDat, only:nState,ndynam,gacc,debugP;
      implicit none;
      integer(4),intent(in) :: n;       !state dimension
      real(8),intent(in) :: xf(n);     !target state
      real(8),intent(in) :: dt;         !time step (s)
      real(8),intent(in) :: qs(n);      !state noise psd
      real(8),intent(out) :: xa(n);     !predicted state at t+dt
      real(8),intent(out) :: phi(nState,n); !state transition matrix
      real(8),intent(out) :: q(nState,n);   !state noise matrix

      integer(4) i;
      real(8) xp1(n),xp2(n);
      real(8) :: xscale(n),epsx,delx;
      !_______________________________________

      if (n /= nState) stop 'get_phix2 error: n /= nState';
      xscale(:6) =10.;
      xscale(7:8) =(/ 0.1, 0.5/);
      epsx =1.d-5;

      if (dt <= 0.) stop 'get_phix2 error: dt <= 0';
      xa(:) =xf(:);

      do i=1,n;
        delx =epsx*xscale(i);
        xa(i) =xf(i) +delx;
        call integx (xp1,  xa,dt,n);
        xa(i) =xf(i) -delx;
        call integx (xp2,  xa,dt,n);
        xa(i) =xf(i);
        phi(:,i) =(xp1(:) -xp2(:))/(2.d0*delx);
      enddo;

      q(:,:) =0.d0;
      forall (i=1:n) q(i,i) =qs(i)*dt;
      q(:,:) =matmul(matmul(phi,q),transpose(phi));
      call integx (xa,  xf,dt,n);

      if (debugP) then;
        write (6,'("get_phix2 t =",8g13.5)') dt;
        call prtmx (phi,n,n,n,"phi");
        call prtmx (q,n,n,n,"q");
      endif;

      return;
      end subroutine get_phix2;
!*********************************************************
      subroutine integx (xa,  xf,dt,n);
!***    integrates state in small intervals for improved accuracy
      use modelDat, only:nState;
      implicit none;
      integer(4),intent(in) :: n;       !state dimension
      real(8),intent(in) :: xf(n);     !target state
      real(8),intent(in) :: dt;         !time step (s)
      real(8),intent(out) :: xa(n);     !predicted state at t+dt

      real(8) xdot(nState),f(n,n),dti,t;
!      real(8) :: dtmax =1.1  !###
      real(8) :: dtmax =0.2;
      !_______________________________________

      if (n /= nState) stop 'integx error: n /= nState';

      t =0.d0;
      if (dt <= 0.) stop 'integx error: dt <= 0';
      xa(:) =xf(:);

      do while (t < dt);
        dti =min(dtmax,dt-t);
        t =t+dti;
        call get_F (f,xdot,   xa,n);
        xa(:) =xa(:) +dti*xdot(:);
      enddo;
!      write (6,'("integx t =",g13.5)') t

      return;
      end subroutine integx;
!**********************************************************************
      subroutine get_F (F,xdot,   xf,n);
!***    computes state transition and process noise matrix and integrates state
      use modelDat, only:nState,ndynam,gacc,tau,markov;
      implicit none;
      integer(4),intent(in) :: n;       !state dimension
      real(8),intent(in) :: xf(n);     !target state
      real(8),intent(out) :: F(n,n); !state transition matrix
      real(8),intent(out) :: xdot(n);     !steta derivative

      integer(4) i;
      real(8) ac,aa,vh,vh2;
      !_______________________________________

      if (n /= nState) stop 'get_f error: n /= nState';

      ac =gacc*xf(7);   !crosstrack accel
      aa =xf(8);        !alongtrack accel
      vh2 =xf(4)**2 +xf(5)**2;
      vh =sqrt(vh2);
      xdot(:3) =xf(4:6);
      xdot(4) =(ac*xf(5) +aa*xf(4))/vh;
      xdot(5) =(-ac*xf(4) +aa*xf(5))/vh;
      xdot(6:8) =0.d0;
      if (markov) forall (i=6:8) xdot(i) =-xf(i)/tau(i);

      f(:,:) =0.d0;
      forall (i=1:3) f(i,i+3) =1.d0;
      f(4,4:8) =(/  aa -xdot(4)*xf(4)/vh, ac -xdot(4)*xf(5)/vh,         &
     &            0.d0,  gacc*xf(5), xf(4) /) /vh;
      f(5,4:8) =(/ -ac -xdot(5)*xf(4)/vh, aa -xdot(5)*xf(5)/vh,         &
     &            0.d0, -gacc*xf(4), xf(5) /) /vh;
      if (markov) forall (i=6:8) f(i,i) =-1.d0/tau(i);

      return;
      end subroutine get_F;
!***********************************************************************
      subroutine RTSsmoother (tFinal,nrec,IntegCov);
!***   RTS fixed-interval smoother.

      use modelDat, only:xt,xa,pa,xf,pf,phi,q,mmax,irecs,               &
     &              nState,ndynam,nestF,nF3,nDynamEst;
!      use GPScom, only:lprint,udFilt,irecs,nbias,nstate,nsat,GPSsat
!      use models, only:nEstF,nF3,nDynamEst,xf,pf,phiqs,nPhiQrow,
!     &                 mapx,mapxi
!      use MSSVpassBias, only:biasSVMS,biasSVMSL
!      use ephgenMod, only: nDynamSat

      implicit none;

!  .... passed variables
      integer(4),intent(in) :: nrec;       !index of last record in unit 40
      logical(4),intent(in) :: IntegCov;   !if true, smoothed covariance is computed and
                                           ! printed in file sigout.dat
      real(8),intent(inout) :: tFinal;     !time of 1st ob (min from J2k)

!  ....  local
!      logical(4) :: tests =.false.      !true if using tst_rts.f to test smoother  ###
      integer(4),parameter :: maxbias =100;     !max # of reconstructed biases allowed in this step
      integer(4) i,j,k,l,ier,nd;    !,irow,icol,nb,isat,i1,i2
!      integer(4) ibrow(maxbias),imax
      real(8),allocatable :: phid(:,:),g(:,:),pbias(:,:);
      real(8),allocatable :: t1(:,:),t2(:,:),t3(:),dx(:),qds(:);
      real(8),allocatable :: pbs(:),xb(:),ps(:),pas(:),sigx(:),xref(:);
      real(8) tobsK,tKLst,td;
      !__________________________________________________________

!      if (udFilt)
!     &     call errhand ("RTSsmoother does not yet work with UDfilt")
      nd =nDynamEst;

      allocate (phid(nDynamEst,nDynamEst),g(nDynamEst,nestF),           &
     &         t1(nestF,nestF),t2(nestF,nestF),t3(nestF),dx(nestF),     &
     &         pbs(nF3),xb(nestF),sigx(nestF),ps(nF3),pas(nF3),         &
     &         qds(nDynamEst*(nDynamEst+1)/2),pbias(nestF,maxbias),     &
     &         xref(nState),stat=ier);
      if (ier > 0) call errhand ('RTSsmoother allocate failed');

      !** initialize smoother state & cov to last filter state & cov
      xb(:) =xf(:);
      k =0;
      do i=1,nestF;
        do j=1,i;
          k=k+1;
          pbs(k) =pf(j,i);
        enddo;
      enddo;

!      write (6,'(/"smoother residuals"/
!     &           "  time       station   mtype     filter N/E",
!     &            "              smoother  N/E")')
      write (6,1070);
 1070 format ('TITLE ="smoother output"'/                               &
     &   'VARIABLES ="t", "xb-xt", "yb-yt", "zb-zt", "xdb-xtd", ',      &
     &   '"ydb-ytd", "zdb-ztd", "tanphib-tanphit", "accab-accat"')

!      open (24,file='xsmooth.plt',status='unknown',access='direct',
!     &       recl=(nEstF+1)*2)
!***   open plot file for saving state estimates
      open (24,file='xerr_smo.plt',status='unknown');
      write (24,1130);
! 1130 format('VARIABLES ="t", "xb-xt", "yb-yt", "zb-zt", "xdb-xtd", ',
!     &   '"ydb-ytd", "zdb-ztd", "tanphib-tanphit", "accab-accat"')
 1130 format('VARIABLES ="t", "x-error", "y-error", "z-error", ',       &
     &   '"xdot-error", "ydot-error", "zdot-error", "tanphi-error", ',  &
     &   '"accel-a-error"')

      if (IntegCov) then;
        open (22,file='Smo_sig.plt',status='unknown');
        write (22,1100);
 1100   format('TITLE="smoothed state 1sigma"')
          write (22,1120);
 1120   format('VARIABLES ="t","sig-xb", "sig-yb", "sig-zb", ',         &
     &   '"sig-xdb", "sig-ydb", "sig-zdb", "sig-tanphib", "sig-accab"')
      endif;

      !*** process filter data backward in time
      irecs =nrec+1;

      do;
        !** read filter variables
        irecs =irecs-1;
        if (irecs <= 0) exit;
        read (42,rec=irecs) tobsK,tKLst,xref,xa,xf,pf,phi,q;

        !**  form pa =pf(i+1/i) using sparse arrays
        pa(:nd,:nestF) =matmul(phi(:nd,:nestF),pf(:nestF,:nestF));
        t2(:nd,:nd) =matmul(pa(:nd,:nd),transpose(phi(:nd,:nd)));

        k=0;
        do i=1,nEstF;
          do j=1,i;
            k =k+1;
            if (i <= nd) then;
              qds(k) =q(j,i);
              pas(k) =t2(j,i) +qds(k);
            else if (j <= nd) then;
              pas(k) =pa(j,i);
            else;
              pas(k) =pf(j,i);
            endif;
          enddo;
        enddo;

!        call prtmx(phi,nstate,nEstF,nEstF,'phi');
!        call prtmx (pf,nstate,nEstF,nEstF,'pf')
!        call prtsmx (pas,nEstF,'pas')

        td =tKlst/1440.d0;

        if (abs(tobsK-tKlst) <= 1.d-6) then;    !should not happen
          stop 'error: tobsK =tKlst';
          !** xa =xf, pa =pf
          l =0;
          do i=1,nestF;
            if (xa(i) /= xf(i)) stop 'xa(i) /= xf(i)';
            l =l+i;
            if (pas(l) /= pf(l,1)) stop 'pas(l) /= pf(l)';
          enddo;
          cycle;     !bypass xb & pb computations and write to files
        endif;

!        call prtsmx (pas,nEstF,'pas+')
!        call prtmx (pbias,nEstF,nEstF,nb,'pbias')
        pbs(:) =pbs(:) -pas(:); !compute before pas inversion

        !*** compute smoother gain g  (same results with RTSgain0 or RTSgain1,
        !*** but RTSgain1 run time is only 61% of RTSgain0)
        phid(:nd,:nd) =phi(:nd,:nd);
        call RTSgain1 (g,  pas,t1,t2,  phid,qds);

!        write (6,'("xa =",13g13.5)') xa(:)
!        write (6,'("xf =",13g13.5)') xf(:)
!        write (6,'("xb =",13g13.5)') xb(:)
        dx(:) =xb(:)-xa(:);
        xb(1:nd) =xf(1:nd) +matmul(g,dx);
        !** xb(nDynamEst+1:) is unchanged (not smoothable)
        write (6,'(/"time ",f14.5/"xb    =",10g13.5,5(/7x,10g13.5))')    &
     &          tKlst, xb(1:nd);

        !** write smoothed states: compute total state estimate
!        xref(:nestF) =xref(:nestF) +xb(:nestF)     !need to use mapx if missing states
        write (43,rec=irecs) tKlst, xref(:nestF);
        write (24,'(9g13.5)') tKlst, xb(:nestF) -xref(:nestF);

        if (IntegCov) then;
          !** compute smoother covariance
          !** t2(:,:) =(pb-pa)*transpose(g)
          !** Note: pbs(:) =pbs(:) -pas(:) from above
          t1(1:nd,:) =g(:nd,:);
          t1(nd+1:,:) =0.d0;
          forall (i=nd+1:nEstF) t1(i,i) =1.d0;

          !** t2 =pbs*t1^T
          do i=1,nestF;
            t3(:) =t1(i,:);
            call sysmult (t2(:,i),  pbs,t3,nestF);
          enddo;

          !** pbs =pf +t1*t2
          l =0;
          do i=1,nestF;
            t3(:) =t1(i,:);
            call smult (dx,  t3,t2, 1,nEstF,i, 1,nestF,1);
            do j=1,i;
              l=l+1;
              pbs(l) =pf(i,j) +dx(j);
            enddo;
            sigx(i) =sqrt(pbs(l));
          enddo;
!          call prtsmx (pbs,nEstF,'pbs+')

          write (6,'("sigxb =",10g13.5,5(/7x,10g13.5/))') sigx(1:nestF);
          !*** write error sigma to file sigout.dat
          write (22,'(9g13.5)') tKlst,sigx(:);
        endif;

      enddo;   !____ end backward time loop

      close (42,status='delete');
      if (IntegCov) close (22);

      !** now write ephemeris file in forward order
      do irecs=1,nrec;
        read (43,rec=irecs) tKlst, xref(:);
        write (44) tKlst, xref(:);
      enddo;
      close (43,status='delete');
      close (44);
!      call cpu_time(tt1)
!      write (6,'(/" Time for RTSsmoother:   =",f10.3)') tt1-tt0

      xf(:) =xb(:);
      tFinal =tKLst;

!      call residual_stat ()
      deallocate (phid,g,t1,t2,t3,dx,pbs,xb,sigx,ps,pas,qds,            &
     &            pbias,xref);

      return;
      end subroutine RTSsmoother;
!*****************************************************************
      subroutine RTSgain1 (g,  pas,t1,t2,  phid,qds);
!***   computes RTS smoother gain by computing
!***   G =inv(phi)*[I-Q*inv(P(i+1/i))].  This version gives the same results
!***   as the version using G =P(i/i)*trn(phi)*inv(P(i+1/i)

      use modelDat, only:nEstF,nF3,nDynamEst,q;
      implicit none;

!      real(8),intent(in) :: tlst,t
      real(8),intent(in) :: phid(nDynamEst,nDynamEst);      !state transition matrix
      real(8),intent(in) :: qds(nDynamEst*(nDynamEst+1)/2); !process noise matrix
      real(8),intent(inout) :: pas(nF3);                    !p(i+1/i) [inverse on output]
      real(8),intent(inout) :: t1(nDynamEst,nEstF);         !scratch array (actually dimensioned as nEstF,nEstF, but we don't need all)
      real(8),intent(inout) :: t2(nDynamEst,nDynamEst);     !scratch array (actually dimensioned as nEstF,nEstF, but we don't need all)
      real(8),intent(out) :: g(nDynamEst,nestF);            !RTS gain for dynamic states

      !** local
      logical(4) :: check =.false.;
      integer(4) i,j,k,ier;
      real(8) t5(nDynamEst,nDynamEst);   !explicitly allocate or pass  !!!
      real(8) eps,err,epsm;
      !_______________________________________________________

      !*** invert pas = filter P(i+1/i)
      eps =1.d-5;
      call sinv (ier,epsm,  pas,  nestF,eps,.false.);      !change to use Cholesky factor (20% faster) !!!
      if (ier < 0) stop 'pas inversion error';

      !** temporarily set g =inv(pas) and q =-qds
      k =0;
      do i=1,nestF;
        do j=1,i;
          k=k+1;
          if (j <= nDynamEst) g(j,i) =pas(k);
          if (i <= nDynamEst) then;
            g(i,j) =pas(k);
            q(i,j) =-qds(k);
            q(j,i) =-qds(k);
          endif;
        enddo;
      enddo;

      !** t1 =q*g =I-qds*inv(pas)
      call smult (t1,  q,g, nDynamEst,nDynamEst,nEstF,                  &
     &                    nDynamEst,nDynamEst,nDynamEst);
      forall (i=1:nDynamEst) t1(i,i) =t1(i,i) +1.d0;

      !** RTS gain gn =inv(phid)*(I-qd*inv(pp))
      !** first compute t2 =inv(phid(1,1))            !very slow for some problems !!!!
      t2(:,:) =phid(:,:);
      call matInvGJ (ier,  t2,t5, nDynamEst,nDynamEst,1);
      if (ier > 0) stop 'inversion error on phid';

      if (check) then;
        !** check inversion
        t5(:,:) =matmul(t2,phid);
        do i=1,ndynamEst;
          do j=1,nDynamEst;
            err =t5(i,j);
            if (j == i) err =err-1.d0;
            if (abs(err) > 1.d-13)                                      &
     &        write(6,'("phid inversion error: i,j =",2i4,              &
     &              ", phid*inv(phid) =",g13.5)') i,j,err;
          enddo;
        enddo;
!        call prtmx (t2,nDynamEst,nDynamEst,nDynamEst,"inv(phid)")
      endif;

      !** finally g =t2*t1 =inv(phi)*(I-qd*inv(pp))
      call smult (g,  t2,t1, nDynamEst,nDynamEst,nEstF,                 &
     &                     nDynamEst,nDynamEst,nDynamEst);
      if (check) call prtmx (g,nDynamEst,nDynamEst,nEstF,"g");

      return;
      end subroutine RTSgain1;
!***************************************************************
      subroutine zeroRowCol (v,  p, j,n);
!**     sets the j-th row & column of symmetric, upper-triangular stored array p to 0
!**     and returns the original row/colum in vector v
      implicit none;
      integer(4),intent(in) :: n;       !dimension of p
      integer(4),intent(in) :: j;       !row/column to zero
      real(8),intent(inout) :: p(n*(n+1)/2);    !symmetric array p
      real(8),intent(out) :: v(n);      !original j-th row/column

      !** local
      integer(4) i,k;
      !___________________________________________

      k =((j-1)*j)/2;
      v(1:j) =p(k+1:k+j);
      p(k+1:k+j) =0.d0;
      k =k+j;
      do i=j+1,n;
        v(i) =p(k+j);
        p(k+j) =0.d0;
        k =k+i;
      enddo;

      return;
      end subroutine zeroRowCol;
