!** condMean.f generates a plot file of the conditional distribution
!   for a 1-state model nonlinear model y=x+alpha*x*2+r with Rs =1

!** Program condMean generates a plot file of the conditional probability
!** distribution for the 1-state model nonlinear model y=x+alpha*x*2+r for five
!** values of measurement noise (-1.5,-0.75,0,0.75,1.5).  It also tests an
!** extended Kalman filter (EKF) and iterated extended Kalman filter (IEKF)
!** solutions for 10000 time points with simulated N(0,1) values for r.
!** Plot files condProb.plt lists the conditional probability distribution
!** versus x for each r-value, and EKF.plt lists the state estimate and 1-sigma
!** uncertainty versus time as computed by the EKF and IEKF.  This is used in
!** Examples 9.1-1 and 9.1-2.

!**************************************
! The author grants the user a non-exclusive, worldwide, royalty-free copyright license to

! 1. reproduce and modify the software for your own purposes, and
! 2. to distribute the software provided that you give credit to the author,
!    do not make claims against the author or Wiley-Interscience,
!    and mark any modifications as your own.  If the software is incorporated in
!    a commercial product, you agree to defend and indemnify this author and
!    Wiley-Interscience against any losses, damages and costs arising from claims,
!    lawsuits and other legal actions brought by a third party.

! The software is provided on an as is basis, without warranties or conditions
! of any kind, either express or implied (including any warranties or conditions
! of title, non-infringement, merchantability or fitness for a particular purpose).
! The user is solely responsible for determining the appropriateness of using and
! distributing the program and assumes all risks associated with its exercise of
! rights under this agreement.
!**************************************

!** sample compile using Intel compiler:
!**     Include in ifort.cfg:
!**   /check:bounds /traceback /include:\kf_book\distribSWm\library /link /nodefaultlib:msvcrt.lib
!**     \kf_book\distribSWm\library\libkf_bnd_ifort.lib

!**     To compile and execute:
!**   set ifortcfg=\kf_book\distribSWm\ifort.cfg
!**   ifort condMean2.f90
!**   condmean2 >condm.lis
!**************************************

      implicit none;
      integer(4),parameter :: nplt=5,nx=350;
      integer(4) i,j,nsum(2),nsamp;
      integer(4) :: ix =19191927;
      real(8) :: rsig=1.0, xbar =0.d0, xt =0.d0;   !### change xbar to test different prior
      real(8) :: sigx =1.0, dxx=0.01;
      real(8) y(nplt),yp,r,h;
      real(8) xmean(nplt),xe(nplt),xei(nplt);
      real(8) xa,pa,xf,pf,xai,pai,xfi,pfi,yt;
      real(8) xmax(nplt),rrms(4),rmean(4),xesum(2);
      real(8),external :: rnor;
      !___________________________________________________

      !** compute true values of state and meas (xt is always 0)
      do j=1,nplt;
        r =(j-3)*0.75;
        call get_h (y(j),h,  xt);
        y(j) =y(j) +r*rsig;
      enddo;

      open (10,file="condProb.plt",status='unknown');
      write (10,1000);
 1000 format('variables ="x", "pxy-rm150", "pxy-rm075", "pxy-r00",',    &
     &   ' "pxy-rp075", "pxy-rp150"')

      call nonlEst (xmax,xmean,xe,xei,                                  &
     &                   y,xt,xbar,rsig,sigx,dxx,nx,nplt,.true.);
      close (10);

!** xbar =0.0
!asum*dx =  0.99954
!psum*dx =  0.11789      0.23716      0.29991      0.25610      0.16612
!mean x = -0.61267     -0.49707     -0.31761     -0.40143E-01  0.35294
!ekf x = -0.75000     -0.37500       0.0000      0.37500      0.75000
!iekf x = -0.52699     -0.32351       0.0000      0.39175      0.76927
!  (iterations are 62, 5, 2, 3, 3)
!xmode = -0.53000     -0.32000       0.0000      0.39000      0.77000
!xmax = -0.52654     -0.32373     -0.24998E-04  0.39174      0.76927

!500 sample Monte Carlo: RMS error =  0.77171      0.82658      0.82462      0.81533
!mean error = -0.45515E-01  0.23584      0.21773      0.22081


!** xbar =1.0
!asum*dx =  0.99954
!psum*dx =  0.46323E-01  0.10878      0.17195      0.19983      0.18944
!mean x = -0.26355     -0.78061E-01  0.17252      0.48736      0.83603
!ekf x = -0.20000      0.10000      0.40000      0.70000       1.0000
!  (iterations are 21, 5, 3, 2, 2)
!iekf x = -0.15194      0.86580E-01  0.37872      0.69227       1.0000
!xmode = -0.15000      0.90000E-01  0.38000      0.69000       1.0000
!xmax = -0.15230      0.86679E-01  0.37878      0.69223      0.99998

      !**  now monte carlo simulation with random true x and r

      yp =rnor(ix);     !initialize random number generator
      rrms(:) =0.d0;
      rmean(:) =0.d0;

      nsamp =200;
      do i=1,nsamp;
        xt =xbar +sigx*rnor(0);
        do j=1,nplt;
          call get_h (y(j),h,  xt);
          r =rsig*rnor(0);
          y(j) =y(j) +r;
        enddo;
        call nonlEst (xmax,xmean,xe,xei,                                &
     &                   y,xt,xbar,rsig,sigx,dxx,nx,nplt,.false.);
        rrms(1) =rrms(1) +sum((xmean(:)-xt)**2); !conditional mean
        rrms(2) =rrms(2) +sum((xmax(:)-xt)**2);  !conditional mode
        rrms(3) =rrms(3) +sum((xe(:)-xt)**2);  !EKF
        rrms(4) =rrms(4) +sum((xei(:)-xt)**2);  !IEKF
        rmean(1) =rmean(1) +sum((xmean(:)-xt)); !conditional mean
        rmean(2) =rmean(2) +sum((xmax(:)-xt));  !conditional mode
        rmean(3) =rmean(3) +sum((xe(:)-xt));  !EKF
        rmean(4) =rmean(4) +sum((xei(:)-xt));  !IEKF
      enddo;
      write (6,'("1000 sample Monte Carlo: RMS error =",4g13.5/         &
     &  "mean error =",4g13.5/)')                                       &
     &          sqrt(rrms(:)/(nsamp*nplt)), rmean(:)/(nsamp*nplt);

      open (10,file="EKF.plt",status='unknown');
      write (10,1010);
 1010 format('variables ="t", "xf", "sig-xf", "xfi", "sig-xfi"')
      !** Monte Carlo simulation of EKF for 10000 time steps
      xt =xbar;     !true = mean
      call get_h (yt,h,  xt);
      xesum(:) =0.d0;
      nsum(:) =0;

      do j=1,500;
        xa =xbar;     !initialize at truth
        pa =sigx**2;
        xai =xa;
        pai =pa;
        do i=1,10000;
          y(1) =yt +rsig*rnor(0);
          call iekf (xf,pf,   y(1),xa,pa,rsig,1);
          xa =xf;
          pa =pf;
          call iekf (xfi,pfi,   y(1),xai,pai,rsig,2);
          xai =xfi;
          pai =pfi;
          if (j == 1 .and. mod(i-1,5) == 0)                             &
     &       write (10,'(5g13.5)') i*1., xf, sqrt(pf), xfi, sqrt(pfi);
        enddo;

        if (abs(xf-xt) < 0.5) then;
          xesum(1) =xesum(1) +xf-xt;
          nsum(1) =nsum(1) +1;
        endif;
        if (abs(xfi-xt) < 0.5) then;
          xesum(2) =xesum(2) +xfi-xt;
          nsum(2) =nsum(2) +1;
        endif;
      enddo;

      write (6,'("average final bias =",2g13.5,2i5)')                   &
     &         xesum(:)/nsum(:), nsum(:);

      stop;
      end;
!*************************************************************
      subroutine get_h (y,h,  x);
      !** evaluates measurement model y =h(x) and partial H =dy/dx
      implicit none;
      real(8),intent(in) :: x;
      real(8),intent(out) ::y,h;

      integer(4) :: order = 2;
      real(8) :: alpha =0.5; !###
      !_______________________________________________

      if (order == 2) then;
        y =x*(1.d0+alpha*x);
        h =1.d0 +2.d0*alpha*x;
      else;
        y =x*(1.d0+alpha*x**2);
        h =1.d0 +3.d0*alpha*x**2;
      endif;

      return;
      end subroutine get_h;
!*********************************************************
      subroutine iekf (xf,pf,   y,xa,pa,rsig,method);
      !** extended or iterated-extended Kalman filter
      implicit none;
      integer(4),intent(in) :: method;
      real(8),intent(in) :: y,xa,pa,rsig;
      real(8),intent(out) :: xf,pf;

      integer(4) k;
      real(8) yp,h,g,xfi,xfil;
      !_________________________________________

      call get_h (yp,h,  xa);

      !** extended kf
      g =pa*h/(pa*h**2+rsig**2);
      xf =xa +g*(y-yp);
      pf =pa -g*h*pa;

      if (method == 2) then;
        !** iterated extended kf
        xfi =xf;
        xfil =1.d20;
        do k=1,100;
          call get_h (yp,h,  xfi);
          g =pa*h/(pa*h**2+rsig**2);
          xfi =xa +g*(y-yp-h*(xa-xfi));
          if (abs(xfil -xfi) < 0.001) exit;
          xfil =xfi;
!          write (6,'("iekf x =",g13.5)') xfi
        enddo;
        xf =xfi;
        pf =pa -g*h*pa;
      endif;

      return;
      end subroutine iekf;
!******************************************************
      subroutine nonlEst (xmax,xmean,xe,xei,                            &
     &                   y,xt,xbar,rsig,sigx,dxx,nx,nplt,lprint);
      !** compute state estimates for scalar state-measurement nonlinear model
      !** Estimates are : conditional mean (xmean), conditional mode (xmode),
      !** extended Kalman filter (xe) and iterated-EKK (xei)
      implicit none;

      integer(4),intent(in) :: nx,nplt;
      logical(4),intent(in) :: lprint;
      real(8),intent(in) :: y(nplt),xt,xbar,rsig,sigx,dxx;
      real(8),intent(out) :: xmax(nplt),xmean(nplt),xe(nplt),xei(nplt);

      integer(4) i,j,k;
      real(8) x,yp,a,b(nplt),sqr2p,g,h,px,asum,h1;
      real(8) plt(-nx:nx,nplt),psum(nplt);
      real(8) pmax(nplt),yp1,ax(nplt,3);
      real(8) c0,c1,c2,xeil,xmode(nplt);
      real(8) :: pi =3.141592653589793d0;
      !________________________________________________

      sqr2p =sqrt(2.d0*pi);
      psum(:) =0.d0;
      xmean(:) =0.d0;
      asum =0.d0;
      xmode(:) =-1.d10;
      pmax(:) =-1.d10;

      !** compute conditional density for plotting and computing mean and mode
      do i=-nx,nx;
        x =xbar+i*dxx;
        a =exp(-0.5d0*((x-xbar)/sigx)**2) /(sqr2p*sigx);
        asum =asum +a;
        call get_h (yp,h,  x);
        do j=1,nplt;
          b(j) =exp(-0.5d0*((y(j)-yp)/rsig)**2) /(sqr2p*rsig);
        enddo;
        plt(i,:) =a*b(:);
        psum(:) =psum(:) +a*b(:);
        xmean(:) =xmean(:) +x*a*b(:);
      enddo;

      xmean(:) =xmean(:)/psum(:);
      if (lprint) then;
        write (6,'("asum*dx =",8g13.5)') asum*dxx;
        write (6,'("psum*dx =",8g13.5)') psum(:)*dxx;
        write (6,'("mean x =",8g13.5)') xmean(:);
      endif;

      !** compute EKF and IEKF estimates
      call get_h (yp,h,  xbar);     !linearize about a priori estimate xbar
      px =sigx**2;
      do j=1,nplt;
        !** extended kf
        g =px*h/(px*h**2+rsig**2);
        xe(j) =xbar +g*(y(j)-yp);
        !** iterated extended kf
        xei(j) =xe(j);
        xeil =1.d20;
        do k=1,100;
          call get_h (yp1,h1,  xei(j));
          g =px*h1/(px*h1**2+rsig**2);
          xei(j) =xbar +g*(y(j)-yp1-h1*(xbar-xei(j)));
!          write (6,'("iekf x =",8g13.5)') xei(j)
          if (abs(xeil -xei(j)) < 0.001) exit;
          xeil =xei(j);
        enddo;
        if (lprint)                                                     &
     &        write (6,'("iekf iter ",i3,", x =",g13.5/)') k, xei(j);
      enddo;

      if (lprint) then;
        write (6,'("ekf x =",8g13.5)') xe(:);
        write (6,'("iekf x =",8g13.5)') xei(:);
      endif;

      !** compute mode of conditional density function
      ax(:,:) =1.d20;
      do i=-nx,nx;
        x =xbar+i*dxx;
        ax(:,1) =plt(i,:)/(psum(:)*dxx);
        if (lprint) write (10,'(8g13.5)') x,ax(:,1);
        do j=1,nplt;
          if (ax(j,1) > pmax(j)) then;
            !* select maximum value
            pmax(j) =ax(j,1);
            xmode(j) =x;
          else if (ax(j,2) > ax(j,1) .and. ax(j,2) > ax(j,3)) then;
            !** compute more accurate peak by fitting quadratic and evaluate peak
            c0 =ax(j,3);
            c1 =0.5d0*(c0-ax(j,1) +4.d0*(ax(j,2)-c0));
            c2 =ax(j,2)-c0-c1;
            xmax(j) =x +(-2.d0 -c1/(2.*c2))*dxx;
          endif;
        enddo;
        ax(:,3) =ax(:,2);
        ax(:,2) =ax(:,1);
      enddo;

      if (lprint) then;
        write (6,'("xmode =",8g13.5)') xmode(:);
        write (6,'("xmax =",8g13.5)') xmax(:);
      endif;

      return;
      end subroutine nonlEst;
