!** Program Tst_KFsmo tests three filter-smoothers (forward-backward, RTS and mBF)
!** using the 2-state problem x1(i) =x1(i-1) +delta-t*x2*i01),  x2*i) =x2*i-1) +q(i).
!** q(i) is N(0,1)  noise.  The measurement is y(i) =x1(i) +r(i) where r(i) is
!** N(0,500) when delta-t =0.2.  Plot file smooth.plt lists the state estimate
!** and 1-sigma uncertainty versus time as computed by the three smoothers.  Plot file
!** xerr_smo.plt lists the error in the RTS smoother state estimates versus time,
!** and Smo_sig.plt lists the RTS smoother 1-sigma uncertainty.  The results for the
!** forward-backward smoother are used in Examples 9.2-1.

!**************************************
! The author grants the user a non-exclusive, worldwide, royalty-free copyright license to

! 1. reproduce and modify the software for your own purposes, and
! 2. to distribute the software provided that you give credit to the author,
!    do not make claims against the author or Wiley-Interscience,
!    and mark any modifications as your own.  If the software is incorporated in
!    a commercial product, you agree to defend and indemnify this author and
!    Wiley-Interscience against any losses, damages and costs arising from claims,
!    lawsuits and other legal actions brought by a third party.

! The software is provided on an as is basis, without warranties or conditions
! of any kind, either express or implied (including any warranties or conditions
! of title, non-infringement, merchantability or fitness for a particular purpose).
! The user is solely responsible for determining the appropriateness of using and
! distributing the program and assumes all risks associated with its exercise of
! rights under this agreement.
!**************************************

!** sample compile and execute using Intel compiler:

!**     Include in ifort.cfg:
!**   /check:bounds /traceback /include:\kf_book\distribSWm\library /link /nodefaultlib:msvcrt.lib
!**     \kf_book\distribSWm\library\libkf_bnd_ifort.lib

!**     To compile and execute:
!**   set ifortcfg=\kf_book\distribSWm\ifort.cfg
!**   ifort tst_KFsmo.f90    (libkf_bnd_ifort.lib should be linked by default
!**   tst_KFsmo >kfs.lis
!**************************************

      module modelDat;
      !** includes model info for kf
      integer(4),parameter :: nState =2,ndynam=2;
      integer(4),parameter :: mmax =1;
      integer(4),parameter :: nplt=4,m=250;
      integer(4) irecs;    !direct access record number for smoothing
      integer(4) nestF,nF3,nDynamEst;
      real(8) xt(nState),xtl(nState);
      real(8) xp(nState),xa(nState),pa(nState,nState),pad(nState);
      real(8) xf(nState),pf(nState,nState),phi(nState,nState);
      real(8) q(nState,nState),qs(nState),tau(nstate);
      real(8) xref(nState);
      real(8) rn,rnoise(mmax);
      real(8) fout(2,0:m,nplt);

      end module modelDat;
!*********************************************************************
      use modelDat;
      implicit none;
      integer(4) i,j,ix,lenrec;
      real(8) r,dt,t,yi,tfinal;
      real(8) yn(m),qn(m),x1,p1,y(0:m);
      real(8) xff(nstate),pff(nstate,nstate);
      real(8) :: qi = 1.0;
      real(8),external :: rnor;
      logical(4) :: IntegCov =.true.;
      !_______________________________

      dt =0.2d0;
      r =100.d0*1.d0/dt;   !meas variance
      rnoise(1) =r;
      phi(1,:) =(/ 1.d0, dt /);
      phi(2,:) =(/ 0.d0, 1.d0 /);   !=0.81873
!      phi(2,:) =(/ 0.d0, exp(-dt/tau) /)    !=0.81873
      q(:,:) =0.d0;
!      q(1,1) =qi   !for forward-backward
      q(2,2) =qi;
      pad(:) =sqrt(10.d0*r*dt);
      ix =19191919;
      x1 =rnor(ix);
      nestF =nstate;
      nDynamEst =nestF;
      nF3 =(nEstF*(nEstf+1))/2;

      !** smoothRTS direct access file
      lenrec =2*(nestf*(3+3*nestf+2*mmax)+2+2*mmax);  !record length in 4-byte words
      open (42,form='unformatted',access='direct',recl=lenrec);
      irecs =0;

      !** generate meas noise
      do i=1,m;
        yn(i) =sqrt(r)*rnor(0);
        qn(i) =rnor(0);
      enddo;

      !** initialize filters
      t =0.d0;
      xt(:) =0.d0;
      xtl(:) =xt(:);
      y(0) =0.d0;
      yi =y(0);
      !** initialize filter
      call kf (x1,p1,  yi,t,dt,r,phi,qi);

      fout(1,0,1:nplt) =x1-xt(1);
      fout(2,0,1:nplt) =p1;
      do i=1,m;
        t =i*dt;
        xtl(:) =xt(:);
        xt =matmul(phi,xt);
        xt(2) =xt(2) +sqrt(qi)*qn(i);
        y(i) =xt(1)+yn(i);
        call kf (x1,p1,  y(i),t,dt,r,phi,qi);
        fout(1,i,1) =x1-xt(1);
        fout(2,i,1) =p1;
      enddo;

      fout(1,m,2:) =x1-xt(1);
      fout(2,m,2:) =p1;

      xff(:) =xf(:);
      pff(:,:) =pf(:,:);
      call RTSsmoother2 (tfinal,m,IntegCov);

      xf(:) =xff(:);
      pf(:,:) =pff(:,:);
      call mBFsmoother (tfinal,m,IntegCov);

      xf(:) =xff(:);
      pf(:,:) =pff(:,:);
      call FPsmoother  (tFinal,m,IntegCov);

      open (10,file='smooth.plt',status='unknown');
      write (10,1000);
1000  format('variables ="t", "xf1-kf", "sigx1-kf", ',                  &
     &  '"xe-rts", "sigx-rts", "xe-mbf", "sigx-mbf", ',                 &
     &   '"xe-fp", "sigx-fp"')
      do i=0,m;
        write (10,'(13g13.5)') i*dt,(fout(1:2,i,j),j=1,nplt);
      enddo;

      stop;
      end;
!**********************************************
      subroutine kf (x1,p1,  yi,t,dt,r,phi2,qi);
      !** standard 2-state kalman filter
      use modelDat, only:pf,pa,xf,xa,phi,q,xtl,irecs,nestf,nstate,pad;
      implicit none;

      real(8),intent(in) :: yi,t,dt,r,phi2(nstate,nstate),qi;
      real(8),intent(out) :: x1,p1;

      integer(4) n;
!      real(8) pa(n,n),pf(n,n),xf(n),xa(n),phi(n,n)
      real(8) g(nestf),dy,a(nestf,nestf),hp(nestf),h(nestf),c;
      !__________________________________________

      if (nestf /= 2) stop 'kf error: nestf /= 2';
      n =nEstF;

      if (t == 0.d0) then;
        pf(:,:) =0.d0;
        pf(1,1) =pad(1)**2;
        pf(2,2) =pad(2)**2;
        xf(:) =0.d0;
        x1 =xf(1);
        p1 =sqrt(pf(1,1));
        return;
      endif;

      phi(:n,:n) =phi2(:,:);

      xa =matmul(phi,xf);
      !** p =phi*p*phi^T
      a(:,:) =matmul(phi,pf);
      pa(:,:) =matmul(a,transpose(phi));

      !** add Q
      q(1,1) =qi*dt**3/3.d0;
      q(1,2) =qi*dt**2/2.d0;
      q(2,1) =q(1,2);
      q(2,2) =qi*dt;
      pa(:,:) =pa(:,:) +q(:,:);

      !** meas update
      h(:) =(/ 1.d0, 0.d0 /);
      dy =yi -dot_product(h,xa);
      hp(:) =matmul(h,pa);
      c =dot_product(h,hp) +r;
      g(:) =hp(:)/c;

      !** write smoother data before state/cov update
      irecs =irecs+1;
      write (42,rec=irecs) t,t-dt,xtl,xa,xf,pf,phi,q,h,g,c,dy;

      xf(:) =xa(:) +g(:)*dy;
      pf(1,:) =pa(1,:) -g(1)*hp(:);
!      pf(2,:) =pa(2,:) -g(2)*hp(:)    !###
      pf(2,:) =(/ pf(1,2), pa(2,2)-g(2)*hp(2) /);

      x1 =xf(1);
      p1 =sqrt(pf(1,1));

      return;
      end subroutine kf;
!***********************************************************************
      subroutine RTSsmoother2(tFinal,nrec,IntegCov);
!***   RTS fixed-interval smoother.

      use modelDat, only:xt,xa,pa,xf,pf,phi,q,mmax,irecs,               &
     &              nState,ndynam,nestF,nF3,nDynamEst,fout;
      implicit none;

!  .... passed variables
      integer(4),intent(in) :: nrec;       !index of last record in unit 40
      logical(4),intent(in) :: IntegCov;   !if true, smoothed covariance is computed and
                                           ! printed in file sigout.dat
      real(8),intent(inout) :: tFinal;     !time of 1st ob (min from J2k)

!  ....  local
      integer(4) i,j,k,l,ier,nd,n;    !,irow,icol,nb,isat,i1,i2
      real(8),allocatable :: g(:,:);
      real(8),allocatable :: t1(:,:),t2(:,:),t3(:),dx(:);
      real(8),allocatable :: pb(:,:),xb(:),ps(:),sigx(:),xref(:);
      real(8),allocatable :: h(:),gf(:);
      real(8) tobsK,tKLst,td,c,dy;
      !__________________________________________________________

      nd =nDynamEst;
      n =nestf;

      allocate (g(nDynamEst,nestF),                                     &
     &         t1(nestF,nestF),t2(nestF,nestF),t3(nestF),dx(nestF),     &
     &         pb(nestf,nestf),xb(nestF),sigx(nestF),ps(nF3),           &
     &         xref(nState),h(nestf),gf(nestf),stat=ier);
      if (ier > 0) call errhand ('RTSsmoother allocate failed');

      !** initialize smoother state & cov to last filter state & cov
      xb(:) =xf(:);
      pb(:,:) =pf(:,:);

      !*** process filter data backward in time
      irecs =nrec+1;

      do;
        !** read filter variables
        irecs =irecs-1;
        if (irecs <= 0) exit;
        read (42,rec=irecs) tobsK,tKLst,xref,xa,xf,pf,phi,q,h,gf,c,dy;

        !**  form pa =pf(i+1/i) = phi*pf*phi^T +q
        t2(:n,:n) =matmul(phi(:n,:n),pf(:n,:n));
        pa(:n,:n) =matmul(t2(:n,:n),transpose(phi(:n,:n))) +q(:n,:n);
        t2(:,:) =pa(:n,:n);

        !** t2 =inv(pa)
        t3(:) =0.d0;
        call matInvGJ (ier,  t2,t3,  n,n,1);    !was gaussj (ier,t2,n,n,t3,1,n)
        if (ier > 0) stop 'RTSsmoother: inversion error on pa';

        !** g =I -q * inv(pa)
        g(:,:) =-matmul(q(:n,:n),t2(:,:));
        forall (i=1:n) g(i,i) =g(i,i) +1.d0;

        !** g =inv(phi) * (I -q * inv(pa))
        call matInvGJ (ier,  phi,g,  n,n,n);  !was gaussj (ier,phi,n,n,g,n,n)    !pb=q^(-1)
        if (ier > 0) stop 'RTSsmoother: inversion error on phi';

        dx(:) =xb(:)-xa(:);
        xb(1:nd) =xf(1:nd) +matmul(g,dx);

        t1(:,:) =pb(:,:) -pa(:n,:n);
        pb(:,:) =pf(:,:) +matmul(matmul(g,t1),transpose(g));

        forall (i=1:n) sigx(i) =sqrt(pb(i,i));
        write (6,'(f14.5,5(10g13.5/))') tKlst, xb(1:nd);
        write (6,'(40x,13g13.5)') sigx(1:nestF);

        fout(1,irecs-1,2) =xb(1)-xref(1);
        fout(2,irecs-1,2) =sqrt(pb(1,1));
      enddo;   !____ end backward time loop

!      close (42,status='delete')
      if (IntegCov) close (22);
      xf(:) =xb(:);
      tFinal =tKLst;

      deallocate (g,t1,t2,t3,dx,pb,xb,sigx,ps,                          &
     &            xref,h,gf);

      return;
      end subroutine RTSsmoother2;
!***********************************************************************
      subroutine mBFsmoother (tFinal, nrec,IntegCov);
!***   modified-Bryson-Fraser fixed-interval smoother.

      use modelDat, only:xt,xa,pa,xf,pf,phi,q,mmax,irecs,               &
     &              nState,ndynam,nestF,nF3,nDynamEst,fout;
      implicit none;

!  .... passed variables
      integer(4),intent(in) :: nrec;       !index of last record in unit 40
      logical(4),intent(in) :: IntegCov;   !if true, smoothed covariance is computed and
                                           ! printed in file sigout.dat
      real(8),intent(inout) :: tFinal;     !time of 1st ob (min from J2k)

!  ....  local
      integer(4) i,j,k,ier,nd,n;
      real(8),allocatable :: phid(:,:);
      real(8),allocatable :: t1(:,:),pb(:,:),t3(:),dx(:);
      real(8),allocatable :: pbs(:),xb(:),ps(:),sigx(:),xref(:);
      real(8),allocatable :: lambda(:),lcov(:,:),h(:),gf(:);
      real(8) tobsK,tKLst,c,dy,a;
      !__________________________________________________________

      nd =nDynamEst;
      n =nEstf;
      if (mmax > 1) stop 'mBFsmoother error: mmax /= 1';

      allocate (phid(nDynamEst,nDynamEst),                              &
     &         t1(nestF,nestF),pb(nestF,nestF),t3(nestF),dx(nestF),     &
     &         pbs(nF3),xb(nestF),sigx(nestF),ps(nF3),                  &
     &         xref(nState),lambda(nestf),lcov(nestf,nestf),            &
     &         h(nestf),gf(nestf),stat=ier);
      if (ier > 0) call errhand ('RTSsmoother allocate failed');

      !** initialize smoother state & cov to last filter state & cov
      xb(:) =xf(:);
      pb(:,:) =pf(:n,:n);
      lambda(:) =0.d0;
      lcov(:,:) =0.d0;

      !*** process filter data backward in time
      irecs =nrec+1;

      do;
        !** read filter variables
        irecs =irecs-1;
        if (irecs <= 0) exit;
        read (42,rec=irecs) tobsK,tKLst,xref,xa,xf,pf,phi,q,h,gf,c,dy;

        !** measurement update
        a =dot_product(gf(:n),lambda(:n));
        lambda(:) =lambda(:) +h(:n)*(dy/c -a);

        !** t1 =transpose(I-KH)
        forall (i=1:n) t1(:,i) = -h(:n)*gf(i);
        forall (i=1:n) t1(i,i) = t1(i,i) +1.d0;

        lcov(:,:) =matmul(t1,matmul(lcov,transpose(t1)));
        do i=1,n;
          do j=1,i;
            lcov(j,i) =lcov(j,i) +h(j)*h(i)/c;
            lcov(i,j) =lcov(j,i);
          enddo;
        enddo;

        !** backward time update
        lambda(:) =matmul(lambda,phi(:n,:n));    !=transpose(phi)*lambda
        t1(:,:) =matmul(lcov(:,:),phi(:n,:n));    !
        lcov(:,:) =matmul(transpose(phi(:n,:n)),t1(:,:));

        !** now form backward state and covariance
        xb(:) =xf(:n) +matmul(pf(:n,:n),lambda(:));
        t1(:,:) =matmul(pf(:n,:n),lcov(:,:));
        pb(:,:) =pf(:n,:n) -matmul(t1(:,:),pf(:n,:n));
        fout(1,irecs-1,3) =xb(1)-xref(1);
        fout(2,irecs-1,3) =sqrt(pb(1,1));
      enddo;

      xf(:) =xb(:);
      tFinal =tKLst;

      deallocate (phid,t1,pb,t3,dx,pbs,xb,sigx,ps,                      &
     &            xref,lambda,lcov,h,gf);

      return;
      end subroutine mBFsmoother;
!***********************************************************************
      subroutine FPsmoother (tFinal, nrec,IntegCov);
!***   Fraser-Potter forward-backward fixed-interval smoother.

      use modelDat, only:xt,xa,pa,xf,pf,phi,q,mmax,irecs,               &
     &              nState,ndynam,nestF,nF3,nDynamEst,fout,rnoise;
      implicit none;

!  .... passed variables
      integer(4),intent(in) :: nrec;       !index of last record in unit 40
      logical(4),intent(in) :: IntegCov;   !if true, smoothed covariance is computed and
                                           ! printed in file sigout.dat
      real(8),intent(inout) :: tFinal;     !time of 1st ob (min from J2k)

!  ....  local
      integer(4) i,j,k,ier,nd,n;
      real(8),allocatable :: u2(:,:);
      real(8),allocatable :: t1(:,:),t2(:,:),pb(:,:),t3(:),xfb(:);
      real(8),allocatable :: xb(:),ps(:),sigx(:),xref(:);
      real(8),allocatable :: w(:),u(:),h(:),gf(:),ju(:,:),ww(:,:);
      real(8) tobsK,tKLst,c,dy,y,epsm,eps,sigpf;
      logical(4) :: new =.false., plt_filt =.false.;
      !__________________________________________________________

      nd =nDynamEst;
      n =nEstf;
      if (mmax > 1) stop 'FPFsmoother error: mmax /= 1';

      allocate (u2(nestf,nestf),                                        &
     &         t1(nestF,nestF),t2(nestF,nestF),t3(nestF),xfb(nestF),    &
     &         xb(nestF),sigx(nestF),ps(nF3),                           &
     &         xref(nState),w(nestf),u(nestf*(nestf+1)/2),              &
     &         h(nestf),gf(nestf),ju(nestF,nestF),pb(nestF,nestF),      &
     &         ww(nestF,nestF),stat=ier);
      if (ier > 0) call errhand ('RTSsmoother allocate failed');

      !** initialize smoother state & cov to last filter state & cov
      xb(:) =xf(:);
      w(:) =0.d0;
      u(:) =0.d0;

      if (plt_filt) then;
        !** save backward filter output rather than smoother
        fout(1,nrec,4) =0.d0;
        fout(2,nrec,4) =100.;
      endif;

      !*** process filter data backward in time
      irecs =nrec+1;

      do;
        !** read filter variables
        irecs =irecs-1;
        if (irecs <= 0) exit;
        read (42,rec=irecs) tobsK,tKLst,xref,xa,xf,pf,phi,q,h,gf,c,dy;

        !** measurement update at tobsK
        y =dy +dot_product(h(:n),xa(:n));
        w(:) =w(:) +h(:n)*y/rnoise(1);
        k =0;
        do i=1,n;
          do j=1,i;
            k =k+1;
            u(k) =u(k) +h(i)*h(j)/rnoise(1);
          enddo;
        enddo;

        !** time update to tKlst.
        if (new) then;
          k =0;
          do i=1,n;
            do j=1,i;
              k =k+1;
              u2(i,j) =u(k);
              u2(j,i) =u2(i,j);
            enddo;
          enddo;

          !** ju =inv(I -Q*U)
          ju(:,:) =+matmul(q(:n,:n),u2(:,:));
          forall (i=1:n) ju(i,i) =ju(i,i) +1.d0;
          t3(:) =0.d0;
          call matInvGJ (ier,  ju,t3,  n,n,1);  !was gaussj (ier,ju,n,n,t3,1,1)    !ju =inv(I -Q*U)
          if (ier > 0) stop 'inversion error on ju';

          t1(:,:) =matmul(ju,phi(:n,:n));
          w(:) =matmul(w,t1);    !=transpose(t1) * w
          t2(:,:) =matmul(u2,t1);      != u * ju * phi

        else;
          !** backward filter eqns appearing in Fraser-Potter paper:
          !** they are wrong (doesn't match RTS solution while "new" eqns match)

          !** first form pb =inv(q)
          pb(:,:) =q(:n,:n);
          t3(:) =0.d0;
          call matInvGJ (ier,  pb,t3,  n,n,1);  !was gaussj (ier,pb,n,n,t3,1,1)    !pb=q^(-1)
          if (ier > 0) stop 'inversion error on q';

          !** t1 =inv(u +inv(q)),  t2 = u
          k =0;
          do i=1,n;
            do j=1,i;
              k =k+1;
              t1(i,j) =u(k) +pb(j,i);
              t1(j,i) =t1(i,j);
              t2(i,j) =u(k);
              t2(j,i) =t2(i,j);
            enddo;
          enddo;
          call matInvGJ (ier,  t1,t3,  n,n,1);  !was gaussj (ier,t1,n,n,t3,1,1)
          if (ier > 0) stop 'inversion error on t1)';

          !** ju =u*t1, ww =I -u*t1
          ju(:,:) =matmul(t2,t1);
          ww(:,:) =-ju(:,:);
          forall (i=1:n) ww(i,i) =ww(i,i) +1.d0;

          !**  t3 =ju * w
          !**  w = phi^T * t3  (w^T =t3^T * phi)
          t3(:) =matmul(ww,w);
          w(:) =matmul(t3,phi(:n,:n));

          !** pb = ju * u *ju^T +J*Q^(-1)*J
          pb(:,:) =matmul(ww,t2);
!          pb(:,:) =matmul(matmul(ww,t2),transpose(ww))     !this is correct but long
!     &            +matmul(matmul(ju,pb),transpose(ju))
!          pb(:,:) =matmul(matmul(ww,t2),transpose(ww)) +q(:n,:n)     !fraser-potter paper (wrong !!!)
          t2(:,:) =matmul(pb,phi(:n,:n));
        endif;

        !** u =phi^T * t2
        k =0;
        do i=1,n;
          do j=1,i;
            k =k+1;
            u(k) =dot_product(phi(:n,j),t2(:,i));
            t2(i,j) =u(k);
            t2(j,i) =u(k);
          enddo;
        enddo;

        !** if u is invertible, compute backward filter solution
        t1(:,:) =t2(:,:);
        call matInvGJ (ier,  t1,t3,  n,n,1);  !was gaussj (ier,t1,n,n,t3,1,1)
        if (ier > 0) then;
          xfb(:) =0.d0;
          t1(:,:) =10000.d0;
        else;
          xfb(:) =matmul(t1,w);
        endif;
        sigpf =sqrt(t1(1,1));

        !** now combine forward/backward estimates at tKlst
        t1(:,:) =matmul(pf(:n,:n),t2(:,:));
        forall (i=1:n) t1(i,i) =t1(i,i) +1.d0;   !t1 =I+pf*u

        !** ww =pf*(I+pf*u)^(-T)                 [transpose(w) =(I+pf*u)^(-1) * pf
        pb(:,:) =pf(:n,:n);
        call matInvGJ (ier,  t1,pb,  n,n,n);  !was gaussj (ier,t1,n,n,pb,n,n)    !pb=t1^(-1)*pf  =transpose(W)
        if (ier > 0) stop 'inversion error on (I+pf*u)';
        xb(:) =matmul(t1(:,:),xf(:n));          !=t1^(-1)*xf

        ww(:,:) =transpose(pb(:,:));            != pf * transpose(t1^(-1))
        t2(:,:) =matmul(ww(:n,:n),t2(:,:));     != ww * u
        t1(:,:) =-t2(:,:);
        forall (i=1:n) t1(i,i) =t1(i,i) +1.d0;  !t1 =I - ww*u
!        pb(:,:) =matmul(matmul(t1(:,:),pf(:n,:n)),transpose(t1(:,:)))
!     &          +matmul(t2(:,:),transpose(ww(:,:)))
!        call prtmx (ww,n,n,n,"ww")
        pb(:,:) =ww(:,:);
        call prtmx (pb,n,n,n,"pb");
        xb(:) =xb(:) +matmul(pb(:,:),w(:));     != t1^(-1)*xf + pb*w

        if (plt_filt) then;
          fout(1,irecs-1,4) =xfb(1)-xref(1);
          fout(2,irecs-1,4) =sigpf;
        else;
          fout(1,irecs-1,4) =xb(1)-xref(1);
          fout(2,irecs-1,4) =sqrt(pb(1,1));
        endif;
      enddo;

      xf(:) =xb(:);
      tFinal =tKLst;

      deallocate (u2,t1,pb,t3,xfb,xb,sigx,ps,                           &
     &            xref,w,u,h,gf,ju,ww);

      return;
      end subroutine FPsmoother;
!***********************************************************************
      subroutine RTSsmoother (tFinal,nrec,IntegCov);
!***   RTS fixed-interval smoother.

      use modelDat, only:xt,xa,pa,xf,pf,phi,q,mmax,irecs,               &
     &              nState,ndynam,nestF,nF3,nDynamEst,fout;
      implicit none;

!  .... passed variables
      integer(4),intent(in) :: nrec;       !index of last record in unit 40
      logical(4),intent(in) :: IntegCov;   !if true, smoothed covariance is computed and
                                           ! printed in file sigout.dat
      real(8),intent(inout) :: tFinal;     !time of 1st ob (min from J2k)

!  ....  local
!      logical(4) :: tests =.false.      !true if using tst_rts.f to test smoother  ###
      integer(4),parameter :: maxbias =100;     !max # of reconstructed biases allowed in this step
      integer(4) i,j,k,l,ier,nd;    !,irow,icol,nb,isat,i1,i2
!      integer(4) ibrow(maxbias),imax
      real(8),allocatable :: phid(:,:),g(:,:),pbias(:,:);
      real(8),allocatable :: t1(:,:),t2(:,:),t3(:),dx(:),qds(:);
      real(8),allocatable :: pbs(:),xb(:),ps(:),pas(:),sigx(:),xref(:);
      real(8),allocatable :: h(:),gf(:);
      real(8) tobsK,tKLst,td,c,dy;
      !__________________________________________________________

!      if (udFilt)
!     &     call errhand ("RTSsmoother does not yet work with UDfilt")
      nd =nDynamEst;

      allocate (phid(nDynamEst,nDynamEst),g(nDynamEst,nestF),           &
     &         t1(nestF,nestF),t2(nestF,nestF),t3(nestF),dx(nestF),     &
     &         pbs(nF3),xb(nestF),sigx(nestF),ps(nF3),pas(nF3),         &
     &         qds(nDynamEst*(nDynamEst+1)/2),pbias(nestF,maxbias),     &
     &         xref(nState),h(nestf),gf(nestf),stat=ier);
      if (ier > 0) call errhand ('RTSsmoother allocate failed');

      !** initialize smoother state & cov to last filter state & cov
      xb(:) =xf(:);
      k =0;
      do i=1,nestF;
        do j=1,i;
          k=k+1;
          pbs(k) =pf(j,i);
        enddo;
      enddo;
      fout(1,nrec,1) =xb(1)-xref(1);
      fout(2,nrec,1) =sqrt(pf(1,1));

!      write (6,'(/"smoother residuals"/
!     &           "  time       station   mtype     filter N/E",
!     &            "              smoother  N/E")')
      write (6,1070);
 1070 format ('TITLE ="smoother output"'/                               &
     &   'VARIABLES ="t", "xb-xt", "yb-yt", "zb-zt", "xdb-xtd", ',      &
     &   '"ydb-ytd", "zdb-ztd", "tanphib-tanphit", "accab-accat"')

!***   open plot file for saving state estimates
      open (24,file='xerr_smo.plt',status='unknown');
      write (24,1130);
 1130 format('VARIABLES ="t", "xb-xt", "yb-yt", "zb-zt", "xdb-xtd", ',  &
     &   '"ydb-ytd", "zdb-ztd", "tanphib-tanphit", "accab-accat"')

      if (IntegCov) then;
        open (22,file='Smo_sig.plt',status='unknown');
        write (22,1100);
 1100   format('TITLE="smoothed state 1sigma"')
          write (22,1120);
 1120   format('VARIABLES ="t","sig-xb", "sig-yb", "sig-zb", ',         &
     &   '"sig-xdb", "sig-ydb", "sig-zdb", "sig-tanphib", "sig-accab"')
      endif;

      !*** process filter data backward in time
      irecs =nrec+1;

      do;
        !** read filter variables
        irecs =irecs-1;
        if (irecs <= 0) exit;
        read (42,rec=irecs) tobsK,tKLst,xref,xa,xf,pf,phi,q,h,gf,c,dy;

        !**  form pa =pf(i+1/i) using sparse arrays
        pa(:nd,:nestF) =matmul(phi(:nd,:nestF),pf(:nestF,:nestF));
        t2(:nd,:nd) =matmul(pa(:nd,:nd),transpose(phi(:nd,:nd)));

        k=0;
        do i=1,nEstF;
          do j=1,i;
            k =k+1;
            if (i <= nd) then;
              qds(k) =q(j,i);
              pas(k) =t2(j,i) +qds(k);
            else if (j <= nd) then;
              pas(k) =pa(j,i);
            else;
              pas(k) =pf(j,i);
            endif;
          enddo;
        enddo;
!        call prtmx(phi,nstate,nEstF,nEstF,'phi')
!        call prtmx (pf,nstate,nEstF,nEstF,'pf')
!        call prtsmx (pas,nEstF,'pas')

        if (abs(tobsK-tKlst) <= 1.d-6) then;    !should not happen
          stop 'error: tobsK =tKlst';
          !** xa =xf, pa =pf
          l =0;
          do i=1,nestF;
            if (xa(i) /= xf(i)) stop 'xa(i) /= xf(i)';
            l =l+i;
            if (pas(l) /= pf(l,1)) stop 'pas(l) /= pf(l)';
          enddo;
          cycle;     !bypass xb & pb computations and write to files
        endif;

!        call prtsmx (pas,nEstF,'pas+')
!        call prtmx (pbias,nEstF,nEstF,nb,'pbias')
        pbs(:) =pbs(:) -pas(:); !compute before pas inversion

        !*** compute smoother gain g  (same results with RTSgain0 or RTSgain1,
        !*** but RTSgain1 run time is only 61% of RTSgain0)
        phid(:nd,:nd) =phi(:nd,:nd);
        call RTSgain1 (g,  pas,t1,t2,  phid,qds);

!        write (6,'("xa =",13g13.5)') xa(:)
!        write (6,'("xf =",13g13.5)') xf(:)
!        write (6,'("xb =",13g13.5)') xb(:)
        dx(:) =xb(:)-xa(:);
        xb(1:nd) =xf(1:nd) +matmul(g,dx);
        !** xb(nDynamEst+1:) is unchanged (not smoothable)
        write (6,'(f14.5,5(10g13.5/))') tKlst, xb(1:nd);

        !** write smoothed states: compute total state estimate
!        xref(:nestF) =xref(:nestF) +xb(:nestF)     !need to use mapx if missing states
!        write (43,rec=irecs) tKlst, xref(:nestF)
        write (24,'(9g13.5)') tKlst, xb(:nestF) -xref(:nestF);

        if (IntegCov) then;
          !** compute smoother covariance
          !** t2(:,:) =(pb-pa)*transpose(g)
          !** Note: pbs(:) =pbs(:) -pas(:) from above
          t1(1:nd,:) =g(:nd,:);
          t1(nd+1:,:) =0.d0;
          forall (i=nd+1:nEstF) t1(i,i) =1.d0;

          !** t2 =pbs*t1^T
          do i=1,nestF;
            t3(:) =t1(i,:);
            call sysmult (t2(:,i),  pbs,t3,nestF);
          enddo;

          !** pbs =pf +t1*t2
          l =0;
          do i=1,nestF;
            t3(:) =t1(i,:);
            call smult (dx,  t3,t2, 1,nEstF,i, 1,nestF,1);
            do j=1,i;
              l=l+1;
              pbs(l) =pf(i,j) +dx(j);
            enddo;
            sigx(i) =sqrt(pbs(l));
          enddo;
!          call prtsmx (pbs,nEstF,'pbs+')

          write (6,'(40x,13g13.5)') sigx(1:nestF);
          fout(1,irecs-1,2) =xb(1)-xref(1);
          fout(2,irecs-1,2) =sqrt(pbs(1));
        endif;

      enddo;   !____ end backward time loop

!      close (42,status='delete')
      if (IntegCov) close (22);

      !** now write ephemeris file in forward order
!      do irecs=1,nrec
!        read (43,rec=irecs) tKlst, xref(:)
!        write (44) tKlst, xref(:)
!      enddo
!      close (43)  !,status='delete')
!      close (44)
!      call cpu_time(tt1)
!      write (6,'(/" Time for RTSsmoother:   =",f10.3)') tt1-tt0

      xf(:) =xb(:);
      tFinal =tKLst;

!      call residual_stat ()
      deallocate (phid,g,t1,t2,t3,dx,pbs,xb,sigx,ps,pas,qds,            &
     &            pbias,xref,h,gf);

      return;
      end subroutine RTSsmoother;
!*****************************************************************
      subroutine RTSgain1 (g,  pas,t1,t2,  phid,qds);
!***   computes RTS smoother gain by computing
!***   G =inv(phi)*[I-Q*inv(P(i+1/i))].  This version gives the same results
!***   as the version using G =P(i/i)*trn(phi)*inv(P(i+1/i)

      use modelDat, only:nEstF,nF3,nDynamEst,q;
      implicit none;

!      real(8),intent(in) :: tlst,t
      real(8),intent(in) :: phid(nDynamEst,nDynamEst);      !state transition matrix
      real(8),intent(in) :: qds(nDynamEst*(nDynamEst+1)/2); !process noise matrix
      real(8),intent(inout) :: pas(nF3);                    !p(i+1/i) [inverse on output]
      real(8),intent(inout) :: t1(nDynamEst,nEstF);         !scratch array (actually dimensioned as nEstF,nEstF, but we don't need all)
      real(8),intent(inout) :: t2(nDynamEst,nDynamEst);     !scratch array (actually dimensioned as nEstF,nEstF, but we don't need all)
      real(8),intent(out) :: g(nDynamEst,nestF);            !RTS gain for dynamic states

      !** local
      logical(4) :: check =.false.;
      integer(4) i,j,k,ier;
      real(8) t5(nDynamEst,nDynamEst);   !explicitly allocate or pass  !!!
      real(8) eps,err,epsm;
      !_______________________________________________________

      !*** invert pas = filter P(i+1/i)
      eps =1.d-5;
      call sinv (ier,epsm,  pas,  nestF,eps,.false.);      !change to use Cholesky factor (20% faster) !!!
      if (ier < 0) stop 'pas inversion error';

      !** temporarily set g =inv(pas) and q =-qds
      k =0;
      do i=1,nestF;
        do j=1,i;
          k=k+1;
          if (j <= nDynamEst) g(j,i) =pas(k);
          if (i <= nDynamEst) then;
            g(i,j) =pas(k);
            q(i,j) =-qds(k);
            q(j,i) =-qds(k);
          endif;
        enddo;
      enddo;

      !** t1 =q*g =I-qds*inv(pas)
      call smult (t1,  q,g, nDynamEst,nDynamEst,nEstF,                  &
     &                    nDynamEst,nDynamEst,nDynamEst);
      forall (i=1:nDynamEst) t1(i,i) =t1(i,i) +1.d0;

      !** RTS gain gn =inv(phid)*(I-qd*inv(pp))
      !** first compute t2 =inv(phid(1,1))            !SLOW: should invert independent diagonal blocks separately !!!!
      t2(:,:) =phid(:,:);
      call matInvGJ (ier,  t2,t5,  nDynamEst,nDynamEst,1);
      if (ier > 0) stop 'inversion error on phid';

      if (check) then;
        !** check inversion
        t5(:,:) =matmul(t2,phid);
        do i=1,ndynamEst;
          do j=1,nDynamEst;
            err =t5(i,j);
            if (j == i) err =err-1.d0;
            if (abs(err) > 1.d-13)                                      &
     &        write(6,'("phid inversion error: i,j =",2i4,              &
     &              ", phid*inv(phid) =",g13.5)') i,j,err;
          enddo;
        enddo;
!        call prtmx (t2,nDynamEst,nDynamEst,nDynamEst,"inv(phid)")
      endif;

      !** finally g =t2*t1 =inv(phi)*(I-qd*inv(pp))
      call smult (g,  t2,t1, nDynamEst,nDynamEst,nEstF,                 &
     &                     nDynamEst,nDynamEst,nDynamEst);
      if (check) call prtmx (g,nDynamEst,nDynamEst,nEstF,"g");

      return;
      end subroutine RTSgain1;
!***************************************************************
      subroutine zeroRowCol (v,  p, j,n);
!**     sets the j-th row & column of symmetric, upper-triangular stored array p to 0
!**     and returns the original row/colum in vector v
      implicit none;
      integer(4),intent(in) :: n;       !dimension of p
      integer(4),intent(in) :: j;       !row/column to zero
      real(8),intent(inout) :: p(n*(n+1)/2);    !symmetric array p
      real(8),intent(out) :: v(n);      !original j-th row/column

      !** local
      integer(4) i,k;
      !___________________________________________

      k =((j-1)*j)/2;
      v(1:j) =p(k+1:k+j);
      p(k+1:k+j) =0.d0;
      k =k+j;
      do i=j+1,n;
        v(i) =p(k+j);
        p(k+j) =0.d0;
        k =k+i;
      enddo;

      return;
      end subroutine zeroRowCol;
