!** tankFil_Smo.f is a maneuvering tank-tracking EKF-smoother using reconstructed ATMT Scout tank data
!** Program tankFilSmo is a maneuvering tank-tracking 6-state EKF-filter/
!** RTS-smoother using the reconstructed ATMT Scout tank data as described
!** in Section 3.2.1.  The input data file is scoutSim4.plt.  Output plot
!** files Xerr_ekfsTnk.plt and Xerr_rtsTnk.plt list the estimate errors
!** for the filter and smoother respectively.  The output report file contains
!** state estimates at various locations in the code.  This is used in Example 9.2-2.

!**************************************
! The author grants the user a non-exclusive, worldwide, royalty-free copyright license to

! 1. reproduce and modify the software for your own purposes, and
! 2. to distribute the software provided that you give credit to the author,
!    do not make claims against the author or Wiley-Interscience,
!    and mark any modifications as your own.  If the software is incorporated in
!    a commercial product, you agree to defend and indemnify this author and
!    Wiley-Interscience against any losses, damages and costs arising from claims,
!    lawsuits and other legal actions brought by a third party.

! The software is provided on an as is basis, without warranties or conditions
! of any kind, either express or implied (including any warranties or conditions
! of title, non-infringement, merchantability or fitness for a particular purpose).
! The user is solely responsible for determining the appropriateness of using and
! distributing the program and assumes all risks associated with its exercise of
! rights under this agreement.
!**************************************

!** sample compile and execute using Intel compiler:

!**     Include in ifort.cfg:
!**   /check:bounds /traceback /include:\kf_book\distribSWm\library /link /nodefaultlib:msvcrt.lib
!**     \kf_book\distribSWm\library\libkf_bnd_ifort.lib

!**     To compile and execute:
!**   set ifortcfg=\kf_book\distribSWm\ifort.cfg
!**   ifort tank_FilSmo.f90     (libkf_bnd_ifort.lib should be linked by default for rnoruni, sinv, symult, prtmx, matInvGJ)
!**   tank_FilSmo >tankfs.lis
!**************************************

      module modelDat;
      !** includes model info for funct
      integer(4),parameter :: nd =6;   !x,y,xd,yd,aa,ac
      integer(4),parameter :: mtmax =6000, mmax=2;
      integer(4) irecs;    !direct access record number for smoothing
      integer(4) nState,ndynam,nestF,nF3,nDynamEst;

      real(8) xt(nd),xtl(nd),xo(2),xp(nd),pad(nd),dt;
      real(8) dyr(2,mtmax),rn(2);    !h(2,mtmax+nd,nd),res(2,mtmax+nd)
      real(8) xa(nd),pa(nd,nd),xf(nd),pf(nd,nd),phi(nd,nd),q(nd,nd);
      real(8) qs(nd),tau(nd),fout(13,0:mtmax,3);

      logical(4) :: firstIter =.false.;     !normal mode linearizing all iter about estimate ###
      logical(4) :: debugp =.true.;
      logical(4) :: markov =.false.;
      logical(4) :: IntegCov =.true.;
      character(60) datafile;
      end module modelDat;
!****************************************************
      program tankFilSmo;
      !** tests tank tracking
      use modelDat;
      use optlDat, only:fstat;
      implicit none;
      integer(4) i,n,m,ix,nmeas,lenrec,nrec;
      integer(4) :: method =1;  !1=EKF, 2=IEKF  ###
      real(8) xe(nd),pfs(nd*(nd+1)/2),cost,t,tfinal;
      real(8),external :: rnor;
      !____________________________________________________

      m =mtmax;
      n =nd;
      nState =nd;
      ndynam =nd;
      nestF =nd;
      nF3 =(nd*(nd+1))/2;
      nDynamEst =nd;

      ix =191919;
      cost =rnor(ix);

      rn(1) =0.0003;   !0.1d0/57.296       !0.1 deg 1-sigma
      rn(2) =2.;                !1 m range

      !** true state, prior and observer position
      datafile ='scoutSim4.plt';
      open (20,file=datafile,status='old');
      read (20,*);
      read (20,*) t,xt(:6);
      close (20);
      xo(:) =(/ 600.d0, -600.d0 /);
      xp(:4) =xt(:4);
      xp(5:6) =0.d0;
      pad(:) =(/ 5.d0, 5.d0, 5.d0, 5.d0, 2.d0, 2.d0 /); !prior 1-sigma
!      pad(:) =(/ 15.d0, 15.d0, 20.d0, 20.d0 /)  !prior 1-sigma

!      xp(:) =xt(:)      !###  test initialized at truth
      xe(:) =xp(:);     !initialize estimate with prior

      do i=1,m;
        dyr(1,i) =rn(1)*rnor(0);    !meas noise
        dyr(2,i) =rn(2)*rnor(0);    !meas noise
      enddo;
!      dyr(:,:) =0.d0    !### test

      debugp =.false.;
      write (*,'("meas#, meas, meas-resid, H(1:4)")');

      open (10,file='Xerr_ekfsTnk.plt',status='unknown');
      write (10,1100);
 1100 format('VARIABLES ="time (s)", "x-error (m)", ',                  &
     &       '"y-error (m)", "xdot-error (m/s)", ',                     &
     &      '"ydot-error (m/s)", "acc-along-error (m/s2)", '            &
     &      '"acc-cross-error (m/s2)"')

      !** smoothRTS
      lenrec =2*(n*(3+3*n)+2);  !record length in 4-byte words
      open (42,form='unformatted',access='direct',recl=lenrec);
      irecs =0;

      method =1;
      call ekf (cost,nmeas,tfinal,   xe,pfs,  n,debugp,method);
      close (10);

      nrec =irecs;
      xf(:) =xe(:);
      fout(1,nrec,2) =tfinal;
      fout(2:7,nrec,2) =xf(1:6)-xt(1:6);
      forall (i=1:6) fout(i+7,nrec,2) =sqrt(pf(i,i));

      call RTSsmoother2 (tFinal,nrec,IntegCov);
      open (10,file='Xerr_rtsTnk.plt',status='unknown');
      write (10,1100);
      do i=1,nrec;
        write (10,'(7g13.5)') fout(1:7,i,2);
      enddo;
      close(10);
!      stop

      open (10,file='Xerr_iekfsTnk.plt',status='unknown');
      write (10,1100);
      method =2;
      xe(:) =xp(:);     !initialize estimate with prior
!      call iefs (cost,nmeas,tfinal,   xe,pfs,  n,debugp,method)
      call ekf (cost,nmeas,tfinal,   xe,pfs,  n,debugp,method);

      close (42,disp='delete');

      stop;
      end program;
!******************************************************************
      subroutine ekf (cost,nmeas,tfinal,   xf,pfs,  nest,debugp,method);
      use modelDat, only:xt,xtl,xo,pad,dyr,mtmax,mmax,dt,rn,qs,irecs,   &
     &                  fout,datafile;
      use optlDat,only:fstat;
      implicit none;

      integer(4),intent(in) :: nest,method;
      logical(4),intent(in) :: debugp;
      real(8),intent(inout) :: xf(nest);
      real(8),intent(inout) :: pfs(nest*(nest+1)/2);
      real(8),intent(out) :: cost,tfinal;
      integer(4),intent(out) :: nmeas;

      integer(4) it,j,kk,i,k,n;
      real(8) t,tlst;
      real(8) xa(nest),xfi(nest),xfil(nest);
      real(8) pf(nest,nest),q(nest,nest),pa(nest,nest),phi(nest,nest);
      real(8) y(mmax,mtmax),ye(mmax,mtmax),dy(mmax);
      real(8) h(mmax,nest),hp(mmax,nest),g(nest,mmax);
!      logical(4),save :: first =.true.
      !_________________________________________

      if (nest /= 6) stop 'nest /= 6';
      cost =0.d0;
      n =nest;

      open (20,file=datafile,status='old');
      read (20,*);
      read (20,*) tlst,xt(:4);

      !** add prior
      qs(:2) =0.d0;
      qs(3:4) =0.00;
      qs(5:6) =(/ 1.0, 0.50 /); !### ??
      pf(:,:) =0.d0;
      do i=1,nest;
        pf(i,i) =pad(i)**2;
      enddo;

      it =0;
!      write (10,'(7g13.5)') tlst,xf(:)-xt(:)

      do;
        xtl(:) =xt(:);
        read (20,*,end=999) t,xt(:6);
        dt =t-tlst;
        tlst =t;
        it =it+1;

        call get_phix (xa,phi,q,   xf,dt,qs,n);
        write (6,'(/"t =",g23.15/6g15.7/6g15.7)') t,xt(:),xa(:);

        pa(:,:) =matmul(matmul(phi,pf),transpose(phi)) +q(:,:);
        do i=2,nest;
          do j=1,nest-1;
            pa(j,i) =0.5d0*(pa(j,i)+pa(i,j));
            pa(i,j) =pa(j,i);
          enddo;
        enddo;

        irecs =irecs+1;
        write (42,rec=irecs) t,t-dt,xtl,xa,xf,pf,phi,q;

        !** get meas and partials
        call get_yh (y(:,it),h,  xt,xo,n);    !h for truth model is ignored
        y(:,it) =y(:,it) +dyr(:,it);
        call get_yh (ye(:,it),h,  xa,xo,n);    !h is linearized about xf
        dy(:) =y(:,it) -ye(:,it);     !/rn

        !** compute Kalman gain matrix
        hp(:,:) =matmul(h,pa);
        call getG (g,  h,hp,rn,n);
        xf(:) =xa(:) +matmul(g(:,:),dy(:));

        if (method == 2) then;
          !** iterated EKF
          xfi(:) =xf(:);
          xfil(:) =1.d20;
          do k=1,100;
            call get_yh (ye(:,it),h,  xfi,xo,n);
            hp(:,:) =matmul(h,pa);
            call getG (g,  h,hp,rn,n);
            dy(:) =y(:,it) -ye(:,it) -matmul(h,xa(:)-xfi(:));     !/rn
            xfi(:) =xa(:) +matmul(g(:,:),dy(:));
            if (sqrt(sum((xfil -xfi)**2)) < 0.001) exit;
            xfil(:) =xfi(:);
            write (6,'("iekf ",i3,", x =",6g13.5)') k,xfi;
            if (k > 80) then;
              !** if more than 80 iterations, use original ekf estimate
              xfi(:) =xf(:);
              call get_yh (ye(:,it),h,  xa,xo,n);
              hp(:,:) =matmul(h,pa);
              call getG (g,  h,hp,rn,n);
              write (6,'(g13.5,", more than 80 iekf iterations")');
              exit;
            endif;
          enddo;
          xf(:) =xfi(:);
        endif;

!        forall (i=1:nest)
!     &         pf(:,i) =pa(:,i) -(g(:,1)*hp(1,i) +g(:,2)*hp(2,i))
        pf(:,:) =pa(:,:) -matmul(g(:,:),hp(:,:));
        do i=2,nest;
          do j=1,nest-1;
            pf(j,i) =0.5d0*(pf(j,i)+pf(i,j));
            pf(i,j) =pf(j,i);
          enddo;
        enddo;

        if (method == 2) then;
          j=3;
        else;
          j=1;
        endif;
        fout(1,it,j) =t;
        fout(2:7,it,j) =xf(1:6)-xt(1:6);
        forall (i=1:6) fout(i+7,it,j) =sqrt(pf(i,i));

        write (10,'(7g13.5)') t,xf(:)-xt(:);
        tfinal =t;
      enddo;    !end time loop

 999  nmeas =it;
      close (20);

      return;
      end subroutine ekf;
!******************************************************************
      subroutine iefs (cost,nmeas,tfinal,  xfi,pfs,  nest,debugp,method);
      !** iterated filter-smoother
      use modelDat, only:xt,xtl,xo,pad,dyr,mtmax,mmax,dt,rn,qs,irecs,   &
     &     fout,datafile,xa,pa,xf,pf,phi,q;
      use optlDat,only:fstat;
      implicit none;

      integer(4),intent(in) :: nest,method;
      logical(4),intent(in) :: debugp;
      real(8),intent(inout) :: xfi(nest);
      real(8),intent(inout) :: pfs(nest*(nest+1)/2);
      real(8),intent(out) :: cost,tfinal;
      integer(4),intent(out) :: nmeas;

      integer(4) it,j,kk,i,k,n,methodp;
      real(8) t,tlst;
      real(8) xfil(nest);    !xa(nest)
      real(8) y(mmax,mtmax),ye(mmax,mtmax),dy(mmax);
      real(8) h(mmax,nest),hp(mmax,nest),g(nest,mmax);
      !_________________________________________

      if (nest /= 6) stop 'nest /= 6';
      cost =0.d0;
      n =nest;
      xf(:) =xfi(:);

      open (20,file=datafile,status='old');
      read (20,*);
      read (20,*) tlst,xt(:4);

      !** add prior
      qs(:2) =0.d0;
      qs(3:4) =0.00;
      qs(5:6) =(/ 1.0, 0.50 /); !### ??
      pf(:,:) =0.d0;
      do i=1,nest;
        pf(i,i) =pad(i)**2;
      enddo;

      it =0;
!      write (10,'(7g13.5)') tlst,xf(:)-xt(:)

      do;
        xtl(:) =xt(:);
        read (20,*,end=999) t,xt(:6);
        dt =t-tlst;
        tlst =t;
        it =it+1;

        !** get meas and partials
        call get_yh (y(:,it),h,  xt,xo,n);    !h for truth model is ignored
        y(:,it) =y(:,it) +dyr(:,it);

        methodp =1;
        call iekfs (t,dt,y(:,it),rn,qs,methodp);

        j =3;
        fout(1,it,j) =t;
        fout(2:7,it,j) =xf(1:6)-xt(1:6);
        forall (i=1:6) fout(i+7,it,j) =sqrt(pf(i,i));

        write (10,'(7g13.5)') t,xf(:)-xt(:);
        tfinal =t;
      enddo;    !end time loop

 999  nmeas =it;
      xfi(:) =xf(:);
      close (20);

      return;
      end subroutine iefs;
!******************************************************************
      subroutine iekfs (t,dt,y,rnoise,qs,method);
      !*** iterated extended filter-smoother
      use modelDat, only:xt,xtl,xo,xa,pa,xf,pf,phi,q,mmax,              &
     &                    firstIter,irecs,nestF,debugP;
      implicit none;

      integer(4),intent(in) :: method;    !1=non-symmetric covariance update, !2=non-symmetric covariance update
      real(8),intent(in) :: t,dt,y(mmax),rnoise(mmax);
      real(8),intent(in) :: qs(nestF);

      integer(4) j,i,k,n,iter,ier,itmax;
      real(8) hp(mmax,nestF),g(nestF,mmax),eps,epsm;
      real(8) h3(mmax,nestF),yp(mmax),dy(mmax),pas(nEstF*(nEstF+1)/2);
      real(8) x0smo(nestF),x1mod(nestF),x1modl(nestF),pai(nestf,nestf);
      real(8) pz(mmax*(mmax+1)/2),s(nestF,nestF);
      real(8) t1(nestF,nestF),t2(nEstF),t3(nestF,mmax);
      !_________________________________________

      n =nestF;
      x0smo(:) =xf(:);
      x1modl(:) =1.d20;
      itmax =100;
      if (t < 5) itmax =1;    !###

      do iter=1,itmax;
        if (dt > 0.d0) then;
          call get_phix (xa,phi,q,   x0smo,dt,qs,n);
          t1(:n,:n) =matmul(phi(:n,:n),pf(:n,:n));
          pa(:n,:n) =matmul(t1(:n,:n),transpose(phi(:n,:n))) +q(:n,:n);
          xa(:) =xa(:) +matmul(phi(:n,:n),(xf(:n)-x0smo(:n)));
        else;
          xa(:) =xf(:);
          pa(:,:) =pf(:,:);
        endif;
        if (iter == 1) then;
          x1mod(:) =xa(:);
          write (6,'("iter ",i3," x1mod =",8g13.5)') iter,x1mod(:);
        endif;

        if (method == 1) then;
          !** t1 =pa^(-1)*phi*pf,  s =transpose(t1)
          pai(:n,:n) =pa(:n,:n);
          call matInvGJ (ier,  pai,t1,  n,n,n);  !was gaussj (ier,pai,n,n,t1,n,n)
          if (ier > 0) stop 'inversion error on pai';
          s(:,:) =transpose(t1(:,:));
          write (6,'("method 1")');

        else;
          !** s =transpose [pa^(-1)*phi*pf]
          !** create symmetric version of pa
          k =0;
          do i=1,n;
            do j=1,i;
              k=k+1;
              pas(k) =pa(j,i);
            enddo;
          enddo;

          !** invert pa
          eps =1.d-4;
          call sinv (ier,epsm,  pas,  n,eps,.true.);
          if (ier /= 0) stop 'iekfs error: sinv pa failure';

          !** s =transpose[pa^(-1)*t1]
          do i=1,n;
            t2(:) =t1(:,i);
            call symult (s(i,:),  pas,t2(:n),n);
          enddo;
          write (6,'("method 2")');
        endif;  !___ end calculation of s
        call prtmx (s,n,n,n,"s");

        call get_yh (yp,h3,  x1mod,xo,n);
        hp(:,:) =matmul(h3,pa);
        k =0;
        do i=1,mmax;
          do j=1,i;
            k=k+1;
            pz(k) =dot_product(hp(j,:n),h3(i,:n));
          enddo;
          pz(k) =pz(k) +rnoise(i)**2;
        enddo;

        eps =1.d-4;
        call sinv (ier,epsm,  pz,  mmax,eps,.true.);
        if (ier /= 0) stop 'ekfs error: sinv pz failure';
        do i=1,n;
          call symult (g(i,:mmax),  pz,hp(:mmax,i),mmax);
        enddo;

        dy(:) =y(:) -yp(:) -matmul(h3(:mmax,:n),(xa(:n)-x1mod(:n)));     !/rn
        x1mod(:n) =xa(:n) +matmul(g(:n,:mmax),dy(:mmax));
        x0smo(:n) =xf(:n) +matmul(s(:n,:n),x1mod(:n)-xa(:n));
        if (itmax > 1 .and. iter == itmax)                              &
     &         stop 'ekfs error: max iterations';
        if (sum((x1mod(:)-x1modl(:))**2) < 0.001**2) exit;
        x1modl(:) =x1mod(:);
        write (6,'("t =",g13.5,", iter ",i3," x1mod =",8g13.5)')        &
     &        t,iter,x1mod(:);
      enddo;   !___ end iteration loop

      irecs =irecs+1;
      write (42,rec=irecs) t,t-dt,xtl,xa,xf,pf,phi,q;

      xf(:) =x1mod(:);
      pa(:,:) =pa(:,:) -matmul(g,hp);
      t3(:n,:mmax) =matmul(pa,transpose(h3));
      forall (i=1:mmax) t3(:,i) =-t3(:,i) +g(:,i)*rnoise(i)**2;
      pf(:,:) =pa(:,:) +matmul(t3,transpose(g));
      do i=2,n;
        do j=1,n-1;
          pf(j,i) =0.5d0*(pf(j,i)+pf(i,j));
          pf(i,j) =pf(j,i);
        enddo;
      enddo;

!      write (20,'(9g13.5)') t,xf(:)-xt(:)
!      write (21,'(9g13.5)') t,(sqrt(pf(i,i)),i=1,n)

      return;
      end subroutine iekfs;
!*********************************************************
      subroutine get_phix (xa,phi,q,   xf,dt,qs,n);
!***    computes state transition and process noise matrix and integrates state
      use modelDat, only:nState,ndynam,debugP;
      implicit none;
      integer(4),intent(in) :: n;       !state dimension
      real(8),intent(in) :: xf(n);     !target state
      real(8),intent(in) :: dt;         !time step (s)
      real(8),intent(in) :: qs(n);      !state noise psd
      real(8),intent(out) :: xa(n);     !predicted state at t+dt
      real(8),intent(out) :: phi(nState,n); !state transition matrix
      real(8),intent(out) :: q(nState,n);   !state noise matrix

      integer(4) i;
      real(8) xdot(nState),f(n,n),phii(n,n),dti,t;
      real(8) :: dtmax =0.11d0;     !little difference between 0.1 and 0.2
      !_______________________________________

      if (n /= nState) stop 'get_phix error: n /= nState';

      t =0.d0;
      if (dt <= 0.) stop 'get_phix error: dt <= 0';
      xa(:) =xf(:);
      phi(:,:) =0.d0;
      forall (i=1:n) phi(i,i) =1.d0;
      q(:,:) =0.d0;

      do while (t < dt);
        dti =min(dtmax,dt-t);
        t =t+dti;
        call get_F (f,xdot,   xa,n);
        phii(:,:) =dti*f(:,:);
!        phii(:,:) =dti*(f(:,:) +0.5*dti*matmul(f,f))        !###
        forall (i=1:n) phii(i,i) =phii(i,i) +1.d0;
        phi(:,:) =matmul(phii(:,:),phi(:,:));
!        call prtmx (phi,n,n,n,"phii")
        q(:,:) =matmul(matmul(phii,q),transpose(phii));
        forall (i=1:n) q(i,i) =q(i,i) +qs(i)*dti;
        xa(:) =xa(:) +dti*xdot(:);
      enddo;

      if (debugP) then;
        write (6,'("get_phix t =",g13.5)') t;
        call prtmx (phi,n,n,n,"phi");
        call prtmx (q,n,n,n,"q");
        call prtmx (xa,1,1,n,"xa");
      endif;

      return;
      end subroutine get_phix;
!*********************************************************
      subroutine get_phix2 (xa,phi,q,   xf,dt,qs,n);
!***    computes state transition and process noise matrix and integrates state
!***    (little difference between get_phiq and get_phiq2)
      use modelDat, only:nState,ndynam,debugP;
      implicit none;
      integer(4),intent(in) :: n;       !state dimension
      real(8),intent(in) :: xf(n);     !target state
      real(8),intent(in) :: dt;         !time step (s)
      real(8),intent(in) :: qs(n);      !state noise psd
      real(8),intent(out) :: xa(n);     !predicted state at t+dt
      real(8),intent(out) :: phi(nState,n); !state transition matrix
      real(8),intent(out) :: q(nState,n);   !state noise matrix

      integer(4) i;
      real(8) xp1(n),xp2(n);
      real(8) :: xscale(n),epsx,delx;
      !_______________________________________

      if (n /= nState) stop 'get_phix2 error: n /= nState';
      xscale(:4) =10.;
      xscale(5:6) =(/ 0.1, 0.5/);
      epsx =1.d-5;

      if (dt <= 0.) stop 'get_phix2 error: dt <= 0';
      xa(:) =xf(:);

      do i=1,n;
        delx =epsx*xscale(i);
        xa(i) =xf(i) +delx;
        call integx (xp1,  xa,dt,n);
        xa(i) =xf(i) -delx;
        call integx (xp2,  xa,dt,n);
        xa(i) =xf(i);
        phi(:,i) =(xp1(:) -xp2(:))/(2.d0*delx);
      enddo;

      q(:,:) =0.d0;
      forall (i=1:n) q(i,i) =qs(i)*dt;
      q(:,:) =matmul(matmul(phi,q),transpose(phi));
      call integx (xa,  xf,dt,n);

      if (debugP) then;
        write (6,'("get_phix2 t =",8g13.5)') dt;
        call prtmx (phi,n,n,n,"phi");
        call prtmx (q,n,n,n,"q");
      endif;

      return;
      end subroutine get_phix2;
!*********************************************************
      subroutine integx (xa,  xf,dt,n);
!***    computes state transition and process noise matrix and integrates state
      use modelDat, only:nState;
      implicit none;
      integer(4),intent(in) :: n;       !state dimension
      real(8),intent(in) :: xf(n);     !target state
      real(8),intent(in) :: dt;         !time step (s)
      real(8),intent(out) :: xa(n);     !predicted state at t+dt

      real(8) xdot(nState),f(n,n),dti,t;
      real(8) :: dtmax =0.2;
      !_______________________________________

      if (n /= nState) stop 'integx error: n /= nState';

      t =0.d0;
      if (dt <= 0.) stop 'integx error: dt <= 0';
      xa(:) =xf(:);

      do while (t < dt);
        dti =min(dtmax,dt-t);
        t =t+dti;
        call get_F (f,xdot,   xa,n);
        xa(:) =xa(:) +dti*xdot(:);
      enddo;
!      write (6,'("integx t =",g13.5)') t

      return;
      end subroutine integx;
!**********************************************************************
      subroutine get_F (F,xdot,   xf,n);
!***    computes state transition and process noise matrix and integrates state
      use modelDat, only:nState,ndynam;   !,tau,markov
      implicit none;
      integer(4),intent(in) :: n;       !state dimension
      real(8),intent(in) :: xf(n);     !target state
      real(8),intent(out) :: F(n,n); !state transition matrix
      real(8),intent(out) :: xdot(n);     !steta derivative

      integer(4) i;
      real(8) ac,aa,vh,vh2;
      !_______________________________________

      if (n /= nState) stop 'get_f error: n /= nState';

      aa =xf(5);        !alongtrack accel
      ac =xf(6);        !crosstrack accel
      vh2 =xf(3)**2 +xf(4)**2;
      vh =sqrt(vh2);
      xdot(1) =xf(3);
      xdot(2) =xf(4);
      xdot(3) =(ac*xf(4) +aa*xf(3))/vh;
      xdot(4) =(-ac*xf(3) +aa*xf(4))/vh;
      xdot(5:6) =0.d0;
!      if (markov) forall (i=5:6) xdot(i) =-xf(i)/tau(i)

      f(:,:) =0.d0;
      forall (i=1:2) f(i,i+2) =1.d0;
      f(3,3:6) =(/  aa -xdot(3)*xf(3)/vh, ac -xdot(3)*xf(4)/vh,         &
     &            xf(3), xf(4) /) /vh;
      f(4,3:6) =(/ -ac -xdot(4)*xf(3)/vh, aa -xdot(4)*xf(4)/vh,         &
     &            xf(4), -xf(3) /) /vh;
!      if (markov) forall (i=6:8) f(i,i) =-1.d0/tau(i)

      return;
      end subroutine get_F;
!***************************************************************
      subroutine get_yh (y,h,  x,xoi,n);
      !** computes measurements of bearing and biased velocity and partials
      implicit none;

      integer(4),intent(in) :: n;
      real(8),intent(in) :: x(n),xoi(4);
      real(8),intent(out) :: y(2),h(2,n);

      real(8) dxe(2),r,r2;
      !_______________________________________________

      dxe(:) =x(:2) -xoi(:2);
      r2 =dxe(1)**2 +dxe(2)**2;
      r =sqrt(r2);
      y(1) =atan2(dxe(1),dxe(2));
      y(2) =r;

      h(:,:) =0.d0;
      h(1,:2) =(/ dxe(2), -dxe(1) /)/r2;
      h(2,:2) =(/ dxe(1), dxe(2) /)/r;

      return;
      end subroutine get_yh;
!*****************************************************
      subroutine getG (g,  h,hp,rn,n);
      !** computes Kalman gain matrix for 2 meas
      implicit none;
      integer(4),intent(in) :: n;
      real(8),intent(in) :: h(2,n),hp(2,n),rn(2);
      real(8),intent(out) :: g(n,2);

      integer(4) i;
      real(8) c(2,2),det,c11,d(2,2);
      !___________________________________________

      c(:,:) =matmul(hp,transpose(h));
      forall (i=1:2) c(i,i) =c(i,i) +rn(i)**2;
!      d(:,:) =c(:,:)
      !** invert c
      det =1.d0/(c(1,1)*c(2,2) -c(1,2)*c(2,1));
      c11 =c(1,1);
      c(1,:) =(/ c(2,2), -c(1,2) /)*det;
      c(2,:) =(/ -c(2,1), c11 /)*det;
!      write (6,'("C*invC =",4g13.5)') matmul(c,d)
      g(:,:) =matmul(transpose(hp(:,:)),c(:,:));

      return;
      end subroutine getG;
!***********************************************************************
      subroutine RTSsmoother2(tFinal,nrec,IntegCov);
!***   RTS fixed-interval smoother.

      use modelDat, only:xt,xa,pa,xf,pf,phi,q,irecs,                    &
     &              nState,ndynam,nestF,nF3,nDynamEst,fout;
      implicit none;

!  .... passed variables
      integer(4),intent(in) :: nrec;       !index of last record in unit 40
      logical(4),intent(in) :: IntegCov;   !if true, smoothed covariance is computed and
                                           ! printed in file sigout.dat
      real(8),intent(inout) :: tFinal;     !time of 1st ob (min from J2k)

!  ....  local
      integer(4) i,j,ier,nd,n;    !,irow,icol,nb,isat,i1,i2
      real(8),allocatable :: g(:,:);
      real(8),allocatable :: t1(:,:),t2(:,:),t3(:),dx(:);
      real(8),allocatable :: pb(:,:),xb(:),ps(:),sigx(:),xref(:);
      real(8),allocatable :: h(:),gf(:);
      real(8) tobsK,tKLst,c,dy;
      !__________________________________________________________

      nd =nDynamEst;
      n =nestf;

      allocate (g(nDynamEst,nestF),                                     &
     &         t1(nestF,nestF),t2(nestF,nestF),t3(nestF),dx(nestF),     &
     &         pb(nestf,nestf),xb(nestF),sigx(nestF),ps(nF3),           &
     &         xref(nState),h(nestf),gf(nestf),stat=ier);
      if (ier > 0) call errhand ('RTSsmoother allocate failed');

      !** initialize smoother state & cov to last filter state & cov
      xb(:) =xf(:);
      pb(:,:) =pf(:,:);

      !*** process filter data backward in time
      irecs =nrec+1;

      do;
        !** read filter variables
        irecs =irecs-1;
        if (irecs <= 0) exit;
        read (42,rec=irecs) tobsK,tKLst,xref,xa,xf,pf,phi,q;!,h,gf,c,dy

        !**  form pa =pf(i+1/i) = phi*pf*phi^T +q
        t2(:n,:n) =matmul(phi(:n,:n),pf(:n,:n));
        pa(:n,:n) =matmul(t2(:n,:n),transpose(phi(:n,:n))) +q(:n,:n);
        !** make symmetric
        do i=2,n;
          do j=1,i-1;
            pa(j,i) =0.5d0*(pa(j,i)+pa(i,j));
            pa(j,i) =pa(i,j);
          enddo;
        enddo;

        t2(:,:) =pa(:n,:n);

        !** t2 =inv(pa)
        t3(:) =0.d0;
        call matInvGJ (ier,  t2,t3,  n,n,1);  !was gaussj (ier,t2,n,n,t3,1,n)
        if (ier > 0) stop 'RTSsmoother: inversion error on pa';

        !** g =I -q * inv(pa)
        g(:,:) =-matmul(q(:n,:n),t2(:,:));
        forall (i=1:n) g(i,i) =g(i,i) +1.d0;

        !** g =inv(phi) * (I -q * inv(pa))
        call matInvGJ (ier,  phi,g,  n,n,n);  !was gaussj (ier,phi,n,n,g,n,n)    !pb=q^(-1)
        if (ier > 0) stop 'RTSsmoother: inversion error on phi';

        dx(:) =xb(:)-xa(:);
        xb(1:nd) =xf(1:nd) +matmul(g,dx);

        t1(:,:) =pb(:,:) -pa(:n,:n);
        pb(:,:) =pf(:,:) +matmul(matmul(g,t1),transpose(g));
        do i=2,n;
          do j=1,i-1;
            pb(j,i) =0.5d0*(pb(j,i)+pb(i,j));
            pb(j,i) =pb(i,j);
          enddo;
        enddo;

        forall (i=1:n) sigx(i) =sqrt(pb(i,i));
        write (6,'(f14.5,5(10g13.5/))') tKlst, xb(1:nd);
        write (6,'(40x,13g13.5)') sigx(1:nestF);

        fout(1,irecs-1,2) =tKlst;
        fout(2:7,irecs-1,2) =xb(1:6)-xref(1:6);
        forall (i=1:6) fout(i+7,irecs-1,2) =sqrt(pb(i,i));
      enddo;   !____ end backward time loop

!      close (42,status='delete')
      if (IntegCov) close (22);
      xf(:) =xb(:);
      tFinal =tKLst;

      deallocate (g,t1,t2,t3,dx,pb,xb,sigx,ps,                          &
     &            xref,h,gf);

      return;
      end subroutine RTSsmoother2;
