!**  tst_fft.f test Singleton's mixed-radix fft algorithm for a variety of data points

!**************************************
! The author grants the user a non-exclusive, worldwide, royalty-free copyright license to

! 1. reproduce and modify the software for your own purposes, and
! 2. to distribute the software provided that you give credit to the author,
!    do not make claims against the author or Wiley-Interscience,
!    and mark any modifications as your own.  If the software is incorporated in
!    a commercial product, you agree to defend and indemnify this author and
!    Wiley-Interscience against any losses, damages and costs arising from claims,
!    lawsuits and other legal actions brought by a third party.

! The software is provided on an as is basis, without warranties or conditions
! of any kind, either express or implied (including any warranties or conditions
! of title, non-infringement, merchantability or fitness for a particular purpose).
! The user is solely responsible for determining the appropriateness of using and
! distributing the program and assumes all risks associated with its exercise of
! rights under this agreement.
!**************************************

!** sample compile and execute using Intel compiler:

!**     Include in ifort.cfg:
!**   /check:bounds /traceback /include:\kf_book\distribSWm\library /link /nodefaultlib:msvcrt.lib
!**     \kf_book\distribSWm\library\libkf_bnd_ifort.lib

!**     To compile and execute:
!**   set ifortcfg=\kf_book\distribSWm\ifort.cfg
!**   ifort tst_fft.f90        !needs rnor and fft_mixedRadix from library
!**   tst_fft >tstfft.lis
!**************************************

      implicit none;
      integer(4),parameter :: nmax =1001;
      integer(4) i,ix,nc,nplt,iunit,junit,iplot,iWindow;
      real(4) c(nmax),d(nmax),dt;
      real(8),external :: rnor;
      !__________________________________________________

      ix =1919191;
      do i=1,nmax;
        c(i) =rnor(0);
      enddo;

      nplt =40;
      iunit =6;
      junit =6;
      dt =1.;
      iplot =0;
      iWindow =0;
      do nc =100,nmax,2;
!        call spectd (c,d,nc,nplt,iunit,junit,dt,iplot,iWindow)      !### uses discrete (not fast) Fourier transform
        call spect (c,d,nc,nplt,iunit,junit,dt,iplot,iWindow);       !uses FFT
      enddo;

      stop;
      end;
! **********************************************************************
      subroutine spect (c,d,nc,nplt,iunit,junit,dt,iplot,iWindow);
!***   compute power spectrum of d using FFT with optional cosine taper window
      implicit none;

      integer(4),intent(in) :: nc;      !number of data points in c
      integer(4),intent(in) :: nplt;    !number of points of d which are to be plotted
      integer(4),intent(in) :: iunit;   !plot unit #
      integer(4),intent(in) :: junit;   !unit # for printout
      integer(4),intent(in) :: iplot;   !0 = no plot, >0 =plot
      integer(4),intent(in) :: iWindow; !window function: 0=none, 1=10% cosine taper
      real(4),intent(in) :: c(nc);      !input data vector of length n
      real(4),intent(in) :: dt;         !time step between points (for sec
      real(4),intent(out) :: d(nc/2+1); !power spectrum of c (length n/2)

      integer(4) n,n2,n3,n5,nd,k,i,j,i1,i2,isn,imin,ix,np,ierr;
      real(4) fn,sumc,freq,pwr,f5,dum,tspan,dmin,dmax,range;
      real(4) a(nc/2+1),b(nc/2+1),f,fr,scalep,scalew;
      integer(4),save :: icall =0, ifail =0;
      !________________________________________

      n =nc;
      isn =0;
      icall =icall +1;

      do j=1,20;
        write(6,'(/"start execution of subroutine spect: n=",i5)') n;
        n2 =n/2;
        n3 =n2+1;
        fn =float(n);
        n5 =n/10+1;
        f5 =3.141593/(0.1*fn);
        sumc =sum(c(1:n))/fn;
!       write(0,'(/"mean from spect :",g13.4)') sumc;
        k=1;
        do i=1,n2;
          a(i) =(c(k)-sumc)/fn;
          b(i) =(c(k+1)-sumc)/fn;
          k =k+2;
        enddo;
        nd =n+1-n5;

!.......multiply by 10% cosine taper window function
        if (iwindow == 1) then;
          do i=1,n5;
            dum =0.5*(1.-cos(f5*float(i-1)));
            i2 =i/2;
            if ( i /= 2*i2 ) then;
              a(i2+1)=dum*a(i2+1);
            else;
              b(i2)=dum*b(i2);
            endif;
            i1=n-i+1;
            i2=i1/2;
            if ( i1 /= 2*i2 ) then;
              a(i2+1)=dum*a(i2+1);
            else;
              b(i2)=dum*b(i2);
            endif;
          enddo;
          scalew =1./0.8625;
        else;
          scalew =1.;
        endif;

!       write(0,'(/"calling fft from spect")');
        isn =1;      !forward transform
        call fft (ierr,  a,b,n2,n2,n2,isn);
        if (ierr == 0 ) exit;
        n =n-2;
        write (6,'("fft cannot handle n: try n=",i5)') n;
        if (j > 19) stop 'spect error: cannot factor n';
      enddo;      !___ end j loop (for reducing n)

      if (n < nc) ifail =ifail+1;
      write (0,'("icall, ifail =",2i4)') icall,ifail;

!     write(0,'(/" calling realtr from spect")');
      isn =1;
      call realtr(a,b,n2,isn);
!     write(0,'(/" realtr completed execution")');

      tspan =fn*dt;
      dum =scalew*tspan;

      dmax =-1.e28;
      pwr =max(a(1)**2*dum,1.e-36);
!      write (0,'("pwr0 =",g13.5)') pwr
      d(1) =10.*log10(pwr);

      do i=2,n3;
        freq=(i-1)/tspan;
        pwr =(a(i)**2+b(i)**2)*dum;
        d(i) =10.*log10(pwr);
        if (d(i) > dmax ) dmax=d(i);
        if ( i > nplt ) cycle;
        write(6,300) i,freq,pwr,d(i);
  300   format(1x,i5,2g13.6,f14.6)
      enddo;

      if ( iplot > 0 ) then;
        range =50.;
        i1 =0.1*dmax+0.999+1.e5;
        dmax =10*(i1-100000);
        dmin =dmax-range;
        scalep =100./range;
        write(iunit,200) dmax;
  200   format('0  decibel plot : max db=',f12.6)
        imin =dmin;
!        call plot (0,iunit,0,imin,10,0,0.,0.)
        np =min0(nplt,n2);
        do i=1,np;
          f =i/tspan;
          ix =floor(scalep*(d(i+1)-dmin)+0.5);
!          call plot (1,iunit,i,ix,0,0,f,d(i+1))
        enddo;

      else;
        write(junit,310) 0.,d(1);
  310   format(2g13.5)
        do i=2,n2;
          fr=(i-1)/(fn*dt);
          if (i <= nplt) write(junit,310) fr,d(i);
        enddo;
        if (nplt >= n3) write(junit,310) n2/(fn*dt),d(n3);
      endif;

   88 continue;
!     write(0,'(/" exiting spect")');
      return;
      end subroutine spect;
! **********************************************************************
      subroutine spectd (c,d,nc,nplt,iunit,junit,dt,iplot,iWindow);
!***   compute power spectrum of d using DFT (not fast) with optional cosine taper window
      implicit none;

      integer(4),intent(in) :: nc;      !number of data points in c
      integer(4),intent(in) :: nplt;    !number of points of d which are to be plotted
      integer(4),intent(in) :: iunit;   !plot unit #
      integer(4),intent(in) :: junit;   !unit # for printout
      integer(4),intent(in) :: iplot;   !0 = no plot, >0 =plot
      integer(4),intent(in) :: iWindow; !window function: 0=none, 1=10% cosine taper
      real(4),intent(in) :: c(nc);      !input data vector of length n
      real(4),intent(in) :: dt;         !time step between points (for sec
      real(4),intent(out) :: d(nc/2+1); !power spectrum of c (length n/2)

      integer(4) n,n2,n3,n5,nd,k,i,j,i1,i2,isn,imin,ix,np,ierr;
      real(4) fn,sumc,freq,pwr,f5,dum,tspan,dmin,dmax,range;
      real(4) a(nc),b(nc),cc(nc),dd(nc),f,fr,scalep,scalew;
      !________________________________________

      n =nc;
      isn =1;

      !** the following loop is retained only to allow beyondCompare to match FFT output
      !** lines between spect and spectd since dft is called below. ###

      do j=1,10;
        write(6,'(/"start execution of subroutine spect: n=",i5)') n;
        n2 =n/2;
        n3 =n2+1;
        fn =float(n);
        n5 =n/10+1;
        f5 =3.141593/(0.1*fn);
        sumc =sum(c(1:n))/fn;
!       write(0,'(/"mean from spect :",g13.4)') sumc;
        k=1;
        do i=1,n2;
          a(i) =(c(k)-sumc)/fn;
          b(i) =(c(k+1)-sumc)/fn;
          k =k+2;
        enddo;
        nd =n+1-n5;

        !** multiply by 10% cosine taper window function
        if (iwindow == 1) then;
          do i=1,n5;
            dum =0.5*(1.-cos(f5*float(i-1)));
            i2 =i/2;
            if ( i /= 2*i2 ) then;
              a(i2+1)=dum*a(i2+1);
            else;
              b(i2)=dum*b(i2);
            endif;
            i1=n-i+1;
            i2=i1/2;
            if ( i1 /= 2*i2 ) then;
              a(i2+1)=dum*a(i2+1);
            else;
              b(i2)=dum*b(i2);
            endif;
          enddo;
          scalew =1./0.8625;
        else;
          scalew =1.;
        endif;

!       write(0,'(/"calling fft from spect")');
        isn =1;
        call fft (ierr,  a,b,n2,n2,n2,isn);     !fft is only called to find n values for DFT that can be compared FFT output
        if (ierr == 0 ) exit;
        n =n-2;
        write (6,'("fft cannot handle n: try n=",i5)') n;
        if (j > 9) stop 'spect error: cannot factor n';
      enddo;

      write(0,'(/" calling dft from spect")');
!      call realtr(a,b,n2,1)
      a(:n) =(c(:n)-sumc);
      b(:n) =0.;
      isn =1;        !forward transform
      call dft (cc,dd,  isn,n,a,b);
      write(0,'(/"dft completed execution")');

      tspan =fn*dt;
      dum =scalew*tspan;

      dmax =-1.e28;
      pwr =cc(1)**2*dum;
      d(1) =10.*log10(pwr);
      do i=2,n3;
        freq=(i-1)/tspan;
        pwr =(cc(i)**2 +dd(i)**2)*dum*4.;  !### 2 or 4 ?
        d(i) =10.*log10(pwr);
        if (d(i) > dmax ) dmax=d(i);
        if ( i > nplt ) cycle;
        write(6,300) i,freq,pwr,d(i);
  300   format(1x,i5,2g13.6,f14.6)
      enddo;

      if ( iplot > 0 ) then;
        range =50.;
        i1 =0.1*dmax+0.999+1.e5;
        dmax =10*(i1-100000);
        dmin =dmax-range;
        scalep =100./range;
        write(iunit,200) dmax;
  200   format('0  decibel plot : max db=',f12.6)
        imin =dmin;
!        call plot (0,iunit,0,imin,10,0,0.,0.)
        np =min0(nplt,n2);
        do i=1,np;
          f =i/tspan;
          ix =floor(scalep*(d(i+1)-dmin)+0.5);
!          call plot (1,iunit,i,ix,0,0,f,d(i+1))
        enddo;

      else;
        write(junit,310) 0.,d(1);
  310   format(2g13.5)
        do i=2,n2;
          fr =(i-1)/(fn*dt);
          if (i <= nplt) write(junit,310) fr,d(i);
        enddo;
        if (nplt >= n3) write(junit,310) n2/(fn*dt),d(n3);
      endif;

   88 continue;
!     write(0,'(/" exiting spect")');
      return;
      end subroutine spectd;
! **********************************************************************
      subroutine dft (c,d,   isn,n,a,b);
!**   discrete Fourier transform (not fast)
      implicit none;
      integer(4),intent(in) :: n;         !number of points in a,b
      integer(4),intent(in) :: isn;       !+1 =forward transform, -1=inverse
      real(4),intent(in) :: a(n),b(n);    !input real and imaginary data
      real(4),intent(out) :: c(n),d(n);   !transform output real and imaginary parts

      integer(4) i,j;
      real(4) fn,dw,er,ei,cd,sd,dcd,del,csum,ssum;
      real(4) dtx;
      !___________________________________________

      fn =1.0/float(n);
      dw =6.2831853*fn;
      if (isn > 0) dw=-dw;
      dtx =0.;

      do i=1,n;
        er =-2.*sin(dtx/2.)**2;
        ei =sin(dtx);
        cd =1.;
        sd =0.;
        csum =a(1);
        ssum =b(1);
        do j=2,n;
          dcd =cd;
          !** compute sin and cosine by recursion
          cd =cd +er*cd -ei*sd;
          sd =sd +ei*dcd+er*sd;
          !** the following three statements compensate for truncation
          !** arithmetic. if rounded arithmetic is used, they may be removed
          del =0.5+0.5/(cd**2+sd**2);
          cd =del*cd;
          sd =del*sd;
          !** test (both are close
!          cd =cos((j-1)*dtx)
!          sd =sin((j-1)*dtx)
          csum =csum +cd*a(j)-sd*b(j);
          ssum =ssum +cd*b(j)+sd*a(j);
        enddo;
        c(i) =csum;
        d(i) =ssum;
        dtx =dtx+dw;
      enddo;

      if (isn < 0) return;
      c(:n) =c(:n)*fn;
      d(:n) =d(:n)*fn;

      return;
      end subroutine dft;
